/* digiconf.h
 * Define Digi adapter families here
 *
 *
 * :set ts=4		// Formatted for tabstops every 4 spaces
 */

// BUS type for this adapter
#define ISA		1
#define EISA	2
#define PCI		3

// WINSIZE: size of shared memory windows in Kbytes
#define	WINSIZE_NONE	0		// Non-windowed (e.g: PCI) adapter
#define WINSIZE_8		8
#define WINSIZE_32		32
#define WINSIZE_64		64


#define MAX_IO_CHOICES 10		// Max number of possible IO address settings
#define MAX_MEM_CHOICES  20		// Max number of possible RAM address settings
#define MAX_PORT_CHOICES 30		// Max number of port count settings

// Keep track of the boards we've configured so far
struct board_info epcaconf[MAXBOARDS];

// Here's the list of adapters that digiconf knows about
struct adapter_struct {
	unsigned int bus;
	unsigned int winsize;
	unsigned int boardid;
	unsigned int io_array[MAX_IO_CHOICES];		// Null-terminated list
	unsigned int mem_array[MAX_MEM_CHOICES];	// Null-terminated list
	unsigned int nports_array[MAX_PORT_CHOICES];	// Null-terminated list
}  adapters[] = {
	
// PC/Xe
	{
		ISA,			// BUS
		WINSIZE_8,		// WINSIZE
		PCXE,			// BOARDID
		{				// IO_ARRAY
				0x100, 0x110, 0x120, 0x200, 0x220, 0x300, 0x320, 0
		},

		{				// MEM_ARRAY
				0xC0000, 0xC2000, 0xC4000, 0xC6000,
				0xC8000, 0xCA000, 0xCC000, 0xCE000,
				0xD0000, 0xD2000, 0xD4000, 0xD6000,
				0xD8000, 0xDA000, 0xDC000, 0xDE000,
				0xE0000, 0
		},
		
		{ 				// NPORTS_ARRAY
				2, 4, 8, 0
		},
	},

// PC/Xr
	{
		ISA,			// BUS
		WINSIZE_32,		// WINSIZE
		PCXR,			// BOARDID
		{				// IO_ARRAY
				0x104, 0x114, 0x124, 0x204, 0x224, 0x304, 0x324, 0
		},

		{				// MEM_ARRAY
				0xa0000, 0xa8000, 0xb0000, 0xb8000,
				0xc0000, 0xc8000, 0xd0000, 0xd8000,
				0
		},
						// NPORTS_ARRAY
		{ 		4, 8, 0
		},
	},

// PC/Xem
	{
		ISA,			// BUS
		WINSIZE_32,		// WINSIZE
		PCXEM,			// BOARDID
		{				// IO_ARRAY
				0x104, 0x114, 0x124, 0x204, 0x224, 0x304, 0x324, 0
		},
				
		{				// MEMARRAY
				0xa0000, 0xa8000, 0xb0000, 0xb8000,
				0xc0000, 0xc8000, 0xd0000, 0xd8000,
				0
		},
				
		{				// NUMPORTS

	                	4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
        	        	44, 48, 52, 56, 60, 64, 0
                                                                
		},
	},

// PC/XI
	{
		ISA,			// BUS
		WINSIZE_64,		// WINSIZE
		PCXI,			// BOARDID
		{				// IO_ARRAY
				0x100, 0x110, 0x120, 0x200, 0x220, 0x300, 0x320, 0
		},

		{				// MEM_ARRAY
				0xC0000, 0xC2000, 0xC4000, 0xC6000,
				0xC8000, 0xCA000, 0xCC000, 0xCE000,
				0xD0000, 0xD2000, 0xD4000, 0xD6000,
				0xD8000, 0xDA000, 0xDC000, 0xDE000,
				0xE0000, 0
		},
		
		{ 				// NPORTS_ARRAY
				4, 8, 16, 0
		},
	},

// PC/CX
	{
		ISA,			// BUS
		WINSIZE_32,		// WINSIZE
		ISACX,			// BOARDID
		{				// IO_ARRAY
				0x108, 0x118, 0x128, 0x208, 0x228, 0x308, 0x328, 0
		},

		{				// MEM_ARRAY
				0xa0000, 0xa8000, 0xb0000, 0xb8000,
				0xc0000, 0xc8000, 0xd0000, 0xd8000,
				0
		},
		
		{ 				// NPORTS_ARRAY

                                8, 16, 24, 32, 40, 48, 56, 64, 72,
                                80, 88, 96, 104, 112, 120, 128, 0
		},
	},

// PCI/Xem
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCIXEM,			// BOARDID
		{				// IO_ARRAY
				0
		},
				
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS

                		4, 8, 12, 16, 20, 24, 28, 32, 36, 40,
                		44, 48, 52, 56, 60, 64, 0
                                                                

		},
	},

// PCI/Xr
	{
		PCI,			// BUS
		WINSIZE_NONE,	// WINSIZE
		PCIXR,			// BOARDID
		{				// IO_ARRAY
				0
		},
				
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				4, 8, 0
		},
	},

// PCI/CX
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCICX,			// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
		                                8, 16, 24, 32, 40, 48, 56, 64, 72,
		                                80, 88, 96, 104, 112, 120, 128, 0
		
		},
	},

// PCI/XrJ
	{
		PCI,				// BUS
		WINSIZE_NONE,			// WINSIZE
		PCIXRJ,				// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				4, 8, 0
		},
	},

// PCI/Xr-422
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCIXR_422,		// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				4, 8, 0
		},
	},

// PCI/Xr920_2
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCI920_2,		// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				2, 0
		},
	},

// PCI/Xr920_4
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCI920_4,		// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				4, 0
		},
	},

// PCI/Xr920_8
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCI920_8,		// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				8, 0
		},
	},

// PCI/EPCJ
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCIEPCJ,			// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
		8, 16, 24, 32, 40, 48, 56, 64, 72,
		80, 88, 96, 104, 112, 120, 128, 136,
		144, 152, 160, 168, 176, 184, 192, 
		200, 208, 216, 224, 0
		
		},
	},

// IBM 8-Port Asynchronous (PCI) Adapter
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCIXR_IBM,		// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
				8, 0
		},
	},

// IBM 128-Port Asynchronous (PCI) Adapter
	{
		PCI,			// BUS
		WINSIZE_NONE,		// WINSIZE
		PCICX_IBM,			// BOARDID
		{				// IO_ARRAY
				0
		},
		{				// MEMARRAY
				0
		},
				
		{				// NUMPORTS
		                                8, 16, 24, 32, 40, 48, 56, 64, 72,
		                                80, 88, 96, 104, 112, 120, 128, 0
		
		},
	},



	{0}
};
