/* ------------------------------------------------------------------------

	digiDload.c 
	Copyright (C) 1996 Digi International.
 
	For technical support please email digiLnux@dgii.com or
	call Digi tech support at (952) 912-3456

	This code represents our first attempt at providing a universal
	code downloader for our products.  As we decide to support upcoming
	products, this code will be modified to provide that support.  It is
	our hope that much of the card specifics can be abstracted out of
	the kernel and into this application.  This will allow for easier
	development as well as debug, and enable longer release schedules
	on the driver itself (Since it will have to be changed less often).  

 	This program is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
 	the Free Software Foundation; either version 2 of the License, or
 	(at your option) any later version.

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 	GNU General Public License for more details.

 	You should have received a copy of the GNU General Public License
 	along with this program; if not, write to the Free Software
 	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Change history:

V1.3.6 October 1998: PEB (Peter E. Berger, pberger@brimson.com)
       -- Added code (ifdef'd to PCIBUG) to circumvent a problem with some
          PCI chipsets.
       -- Fixed a bug that kept multiple ISA cards from initializing.
       -- Previous versions would quit on any error.  In a multiple
          adapter setup a failure to initialize a prior board would also
          keep subsequent boards from being initialized.  This version
          displays error messages as appropriate, but then continues trying
          to initialize any subsequent boards.
       -- Changed makefile to link digiDload using the static option so
          it will run even on platforms with incompatible libraries.
       -- Fixed a include file references that prevented compiling digiDload
          against newer Linux source trees.
       -- Added messages to display the adapter type and number of ports
          as they are initialized.  This is especially useful in the event
          that some adapters succeed while others fail to initialize.

V1.3.7 January 1999: PEB
       -- Removed -D__KERNEL__ from Makefile -- unnecessary and it caused
          compile failures with 2.1.x kernels.
       -- Added CX initialization.  Much of this actually belongs in the
          driver, but it was easier to develop it here first.
       -- Added PCI 2R-290 support.
       -- Added init() time call to memoff() to turn off card memory on
          non-PCI cards to avoid conflicts with shared memory addresses.

V1.3.8  February 2, 1999: PEB
       -- Removed CX concentrator upload code since that's now taken care
          of in the driver itself (epca.c)
       -- Add call to DIGI_DISABLE before initializing cards.  Otherwise,
          subsequent calls to digiDload fail with CX cards since download
          requests from the card are misinterpreted.
V1.3.9  February 3, 1999: PEB
       -- Add support for multiple CX configurations via "cxconf" and
          "/etc/digi/cxconf.dat"
V1.3.10  February 4, 1999: PEB
       -- Removed debugging printfs
V1.3.11  February 5, 1999: PEB
       -- Added report_numports() and call to it following DIGI_INIT calls
          to report "numports" for each card.  This is necessary because
          these values cannot be read from PCI cards until then.
V1.3.12  February 25, 1999: PEB
       -- Added calls to new ioctls: DIGI_CXCON_INIT and DIGI_CXCON_NEXT
          to upload C/X concentrator binaries from digiDload's user space
          to the driver (the driver then takes care of getting this to the
          card as needed).  Uploading the concentrator image to the card
          cannot be digiDload's responsibility since the concentrators
          may call for image refreshes long after digiDload exits (e.g: if
          the concentrators are power cycled).
V1.3.13  March 9, 1999: PEB
       -- Added PCI XR_422 support
V1.3.14  April 16, 1999: PEB
       -- Removed the now-extraneous references to PCIXR_REVD
V1.3.15  May 24, 1999: CWS
       -- Made digiDload aware of the EPC/Jupiter board (type 17, PCIEPCJ)
       -- Added calls to new ioctls: DIGI_EPCXCON_INIT and DIGI_EPCXCON_NEXT
          to upload EPC/X concentrator binary from digiDload's user space to
          the driver (the driver then takes care of getting this to the card
          as needed). 
       -- Added new Concentrator image /etc/digi/fxcon.bin, to deal with 
          EPC/X concentrators.  EPC/X concentrators only work on EPC/X cards,
          but C/X concentrators *AND* EPC/X concentrators are supported on
          the PCI EPC/X Jupiter card.
       -- Added debug code ifdef'd to CWSDEBUG
V1.3.16  May 27, 1999: CWS/PEB
       -- In cxmagic(), needed to increment cxindex after each magic string
          load otherwise all CX's get the same config string.
       -- Added EPC/X concentrator configuration support to getcxconf()
          and cxmagic().  CX and EPC/X concentrator topology information
          now comes from "/etc/digi/cxconf.dat" and "/etc/digi/epcxconf.dat"
          respectively.
V1.3.17  June 9, 1999: CWS/PEB
       -- In 1.3.16, CWSDEBUG was left on by default.  Fixed that.
V1.3.18  June 15, 1999: PEB
       -- Only load C/X and/or EPC/X concentrator images if needed (i.e:
          if C/X and/or EPC/X adapters have been found).
V1.3.19  July 30, 1999: PEB
       -- Added support for SmartPort adapter family
       -- Only display *C/X concentrator "using defaults" messages if
          a *C/X card is selected.
V1.3.20  September 19, 1999: PEB
       -- Added support unlinking old and creating current device nodes
          on the fly, and for querying the driver for the major numbers
          in order to do so.  In an attempt to keep from breaking any
          existing software, the first 256 devices are handled as per
          historical precedent: they are associated with the standard major
          device numbers (DIGI_MAJOR and DIGICU_MAJOR, defined in <major.h>),
          and so their device nodes can be managed by digiConfig.  When
          configured to support more than 256 devices, the driver dynamically
          allocates new major device numbers as needed.  Each major
          number supports another bank of 256 ports, so 2 major numbers
          are needed to support up to 512 ports, 3 for up to 768, etc...

          Since these new major numbers are allocated at driver load
          time, the nodes for those devices can only reliably be made
          after that time.  By default, digiDload will query the driver
          for the list of major numbers currently in use, and -- for any
          devices after the first 256 --  it deletes any existing device
          nodes associated with those nodes, then mknods new nodes for
          any ports that the driver certifies as operational.

          Note: the device nodes are created with permissions of 666
          (readable and writable by everyone) but masked against the
          caller's current umask.  To really get 666 permissions the
          caller's umask should be 0 (or something that doesn't have bits
          in common with 0666).  E.g:
                 x=`umask`
                 umask 0
                 digiDload v
                 umask $x

       -- Added "-nomknod" command line option.  This is only needed
          when the driver is configured to work with more than 256 ports.
          By default -- for these device -- digiDload removes and creates
          the device nodes each time it is run.  But there may be some
          reasons to suppress this default behavior.  Notably: if DEVFS is
          enabled, it handles creating and removing the device nodes in a
          much more elegant way, so digiDload's efforts would be misguided.
          Specifying the "-nomknod" argument causes digiDload to suppress
          this behavior -- digiDload will neither unlink old nor mknod
          new device nodes.

       -- Add code to mess() so it only displays non-error messages if
          verbose is set.  Error messages are always displayed.
V1.3.21  January 19, 2000: CWS
       -- Added code to count errors in the mess() code.
       -- We now track if there's been an error - and if there has,
          we scream loudly at the user, instead of just saying 
          "digiDload complete."
       -- In addition to the error detection, when an error *HAS* been
       	  detected, when we exit, we exit with a negative value, to 
       	  indicate to a calling shellscript that there has been a 
       	  problem, and to fail appropriately.
V1.3.22  February 25, 2000: CWS
       -- Made changes to digiDload that relate to the structural changes
       	  required for DPA compatiblity
       -- Removed support for Xeve/Xe64k/Xem-EISA/CX-EISA boards, as 
          these boards are either A) Don't apply anymore, or B) are 
          obsoleted.
V1.3.23	 February 26, 2000: CWS
       -- we were looking at /dev/digi_ctl for our download device. This
	  has been changed.  We're looking at /dev/digiCtl now.
       -- Started looking for all supporting files in /etc/digiepca rather
          than /etc/digi
       -- Enabled the "device creation" routines, for the first 256 ports.  
          In previous, this functionality was in digiDload, however it had
          been disabled for the first 256 ports, so as not to interfere
          with current operation.
V1.3.24  March 24, 2000: CWS
       -- Readded EPC/X Jupiter (PLX) PCI support to digiDload. EPC/X PCI
          is now officially supported by Digi International.
V1.3.25  September 6, 2000: CWS
	-- Made digiDload look in /etc/digiepca/firmware - for its firwmare
	   binaries.  This is so that we can pull in the proper feps from
	   the CVS tree, upon source checkout.
V1.3.26  April 24, 2001: JAR
	-- Made digiDload look in /usr/lib/dg/epca/firmware - for its
           firwmare images.  This is to better match the other linux drivers.
        -- Look for the configuration information in /etc/digi again.
        -- Some support for DEVFS -- not yet complete.
        -- Make device nodes in /dev and in /dev/dg/epca like the dgdm driver
           unless we're using DEVFS -- in which case, just in /dev/dg/epca
        -- general comment cleanup (This code deserves a general cleanup
           but I don't have the time at the moment -- FIXME)
V1.3.27  June 22, 2001: JAP
	-- Added support for the IBM 8-Port Async PCI Adapter, which
	   is similar to the AccelePort Xr
V1.3.28  September 7, 2001: JAP
	-- Added support for the IBM 128-Port Async PCI Adapter, which
	   is similar to the AccelePort C/X
V1.3.29  September 23, 2002: PEB
    -- Added transparent print support

--------------------------------------------------------------------------- */

/*
 * Some PCI Chipsets fail on some multi-byte (e.g: 32-bit) transfers.  If
 * PCIBUG is defined here, we will use bytewise transfers (via slow_memcpy())
 * rather than using calls to memcpy().
 */
#define PCIBUG

#undef CWSDEBUG

typedef unsigned char unchar;
extern int iopl( int level );

#include <linux/config.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/param.h>

#include "config.h"


/******************************************************************
 * NOTE: In asm/io.h outb seems to have its aruments reversed
 * from the norm.
 ******************************************************************/

#include <asm/io.h>
#include <termio.h>
#include <sys/ioctl.h>

 
#include "rascon.h"
#include "epca.h"

#include <sys/utsname.h>

#include <linux/major.h>


#define TRUE  1
#define FALSE 0
#define BIOS 1
#define FEP  2

#define MAXCARDS 7

/* This is used to calculate the path to digiDload, for temp_name
 *
 * NOTE: LIBDIR and ETCDIR are defined in the Makefile and inherited from
 *       Makefile.inc   Changes should be made in Makefile.inc
 *       The default is now /usr/lib/dg/epca for LIBDIR and
 *       /etc/digi for ETCDIR
 */

#define FWPATH	LIBDIR"/firmware"

/* concentrator download images */
char *con_fnames[] = {
	FWPATH"/cxcon.bin",
	FWPATH"/fxcon.bin",
	FWPATH"/ibmcxcon.bin",
	FWPATH"/ibmencon.bin",
	0,
};

/* C/X concentrator configuration */
#define CXCONFIG ETCDIR"/cxconf.dat"
#define EPCXCONFIG ETCDIR"/epcxconf.dat"

/* CX Concentrator configuration strings can be a maximum of 21 bytes long */
#define CXCFGLEN 21
unsigned char cx_cfg[MAXCARDS][CXCFGLEN];
unsigned char epcx_cfg[MAXCARDS][CXCFGLEN];

/* How many CX configuration images have we found? */
unsigned int no_cximages = 0;
unsigned int no_epcximages = 0;

/*  Which CX configuration image should we use next? */
unsigned int cxindex = 0;
unsigned int epcxindex = 0;

/* The driver can handle 256 devices per major number.  For historical
 * reasons (i.e: in order not to break existing programs that may refer
 * expect Digi devices to be associated with DIGI_MAJOR) the first major
 * number is fixed at DIGI_MAJOR (defined in <major.h>).  If more than
 * 256 devices are specified, the driver allocates new major numbers
 * dynamically.  digiDload keeps track of them here. */
unsigned int majors[sizeof(TTY_BANKNAMES)-1];	/* tty devices */
unsigned int prmajors[sizeof(TTY_BANKNAMES)-1];	/* transparent print devices */

char dev_path[MAXPATHLEN];  /* pathnames for the mknod invocations */
char *epca_banknames=TTY_BANKNAMES;


int do_mknod = 1;	/* Manage device nodes in digiDload */

/* There should be a way to interrogate board for this */
#define RAM_SIZE 32768

int globerr = 0;	/* If there's an error, lets tell the user that. CWS */

/* ---------------- Begin function prototypes ------------------ */

void mess(int err, int indent, const char * format, ...);
void nap(int);
int init(int digiFD);
void memoff(int card);
/*void memon(int card);*/
int reset_cards(int);
int binary_is_running(int card, int);
int wait_for_binary(int card, int);
int load_binary(int card, int);
unsigned long int readRAM(int, int, int, int);
void writeRAM(int, long, int, int, int);
void update_devs();


/* ---------------- End function prototypes ------------------ */


struct file_map
{
    int board_type;      /* What card does this go on ?      */
    int binary_type;     /* Is it a BIOS, a FEP, or a CONC ? */
    char *fname;         /* Name of binary in filesystem.    */
    unsigned long addr;  /* Where the binary goes.           */
};

struct board_t 
{
	volatile char * ram_addr; 
	volatile char * control_addr; 
	struct digi_info digi_info;
};

struct board_t board[MAXCARDS] ;

/* ----------------------------------------------------
	Note: PCXE and PCXEVE share the same binaries.
	Null filenames mean we are at end of list.
------------------------------------------------------- */

struct file_map file_map[] = 
{
    {PCXEM, BIOS, "sxbios.bin", 0x1000}, 
    {PCXEM, FEP,  "sxfep.bin",  0x1000},
    {PCXR, BIOS, "xrbios.bin", 0x1000}, 
    {PCXR, FEP,  "xrfep.bin",  0x1000},
    {PCIXR,  BIOS, "xrbios.bin", 0x1000}, 
    {PCIXR,  FEP,  "xrfep.bin",  0x1000},
    {PCIXRJ, BIOS, "xrbios.bin", 0x1000}, 
    {PCIXRJ, FEP,  "xrfep.bin",  0x1000},
    {PCIXEM, BIOS, "sxbios.bin", 0x1000}, 
    {PCIXEM, FEP,  "sxfep.bin",  0x1000},
    {PCXE,   BIOS, "xxbios.bin", 0x1800}, 
    {PCXE,   FEP,  "xxfep.bin",  0x0},
    {PCXI,   BIOS,  "xxbios.bin",  0xFF800},
    {PCXI,   FEP,   "xxfep.bin",  0x2000},
    {PCI920_4,  BIOS, "xrbios.bin", 0x1000}, 
    {PCI920_4,  FEP,  "xrfep.bin",  0x1000},
    {PCI920_8,  BIOS, "xrbios.bin", 0x1000}, 
    {PCI920_8,  FEP,  "xrfep.bin",  0x1000},
    {PCI920_2,  BIOS, "xrbios.bin", 0x1000}, 
    {PCI920_2,  FEP,  "xrfep.bin",  0x1000},
    {PCIXR_422,  BIOS, "xrbios.bin", 0x1000}, 
    {PCIXR_422,  FEP,  "xrfep.bin",  0x1000},
    {PCIXR_IBM,  BIOS, "xrbios.bin", 0x1000},
    {PCIXR_IBM,  FEP,  "xrfep.bin",  0x1000},

    {PCICX,     BIOS, "cxpbios.bin",  0x1000},
    {PCICX,     FEP,  "cxpfep.bin",  0x1000},
    {PCICX_IBM, BIOS, "cxpbios.bin",  0x1000},
    {PCICX_IBM, FEP,  "cxpfep.bin",  0x1000},

    {ISACX,  BIOS, "cxbios.bin",  0x7800},
    {ISACX,  FEP,  "cxfep.bin",  0x5000},

    {PCIEPCJ, BIOS, "cxpbios.bin", 0x1000},
    {PCIEPCJ, FEP,  "cxpfep.bin", 0x1000},
    
    {-1, -1, NULL, -1}		/* End of array */
};

/*  ---------------------------------------------------------------------
	The last argument of each file_map above indicates the address  
	at which the binary is to be loaded.  In case of the PCXE & PCXEVE
	BIOS this should be calculated because it is dependant on the size of
	the binary.	
------------------------------------------------------------------------- */

char *version = "1.3.29";

int verbose = FALSE;	/* verbose defaults to true */
int board_count = 0;	/* Number of boards found by driver */
int port_count = 0;	/* Number of ports found by driver */
int do_mkdev=1;		/* Manage (unlink and mknod) device lines */
int forcecuds = 0;	/* Force making cud devices? */
int do_cuds = 0;	/* Do we do cud devices? */

struct kv {
        int major;
        int minor;
};
                

struct kv kv;		/* Kernel version */

#ifdef PCIBUG
/* From linux/lib/string.c */
void * slow_memcpy(void * dest,const void *src,size_t count)
{
    char *tmp = (char *) dest, *s = (char *) src;
    while (count--)
        *tmp++ = *s++;

    return dest;
}
#endif


void buildkernelver(void) {
	struct utsname *uts;
	int err;
	uts = (struct utsname*)malloc(sizeof(struct utsname));
	err = uname(uts);
	/* set up kernel structure buffer only ONCE during the program. */
	if(err) {
		perror("uname()");
		return;
	}

	err = sscanf(uts->release, "%d.%d", &kv.major, &kv.minor);
	
	if(err != 2) {
		printf("checkkernelver() - Parse error on kernel version\n");
		return;
	}
}



/* -------------------- Begin usage  --------------------- */
void usage() {
	printf( "Usage: digiDload [-][v] [-nomkdev]\n" );
}

/* -------------------- Begin mess --------------------- */
void mess(int err, int indent, const char * format, ...)
{ /* Begin mess */

	va_list args ; 

	/* Always display error messages */
	if( verbose || err ) {
		va_start (args, format);

		/* Indent so messages appear grouped under the card id header*/
		if( indent )
			fprintf(stderr, "....." );
		if( err == 1) {
			fprintf(stderr, "<Error> - " );
			globerr++;
		}
		
		if( err == 2) {
			fprintf(stderr, "<Warning> - " );
		}
		
		if( err == 3) {
			fprintf(stderr, "<Suggestion> - " );
		}
		
		vfprintf(stderr, format, args);
		va_end(args);
	}

} /* End mess */


/* -------------------- Begin nap ----------------------- */
void nap(int ms)
{ /* Begin nap */

	struct timeval tv ;

	/* Shouldn't this be in the library somewhere? */

	tv.tv_sec = 0 ; 
	tv.tv_usec = ms * 1000 ; 

	select(0,0,0,0,&tv) ;

} /* End nap */

/* -------------------- Begin writeRAM  ---------------------- */
void writeRAM(int target_addr, long value, int size, int card, int bus)
{ /* Begin writeRAM */

	if (bus == PCI_CTL)
	{ /* Begin PCI_CTL */

		if (size == SIZE8)
			*((int_8 *)&board[card].control_addr[target_addr]) = value;
		else 
		if (size == SIZE16)
			*((int_16 *)&board[card].control_addr[target_addr]) = value;
		else
		if (size == SIZE32)
			*((int_32 *)&board[card].control_addr[target_addr]) = value;
		else
			mess(1, 1, "writeRAM unknown size\n");

	} /* End PCI_CTL */
	else
	{ /* Begin non-PCI */

		if (size == SIZE8)
			*((int_8 *)&board[card].ram_addr[target_addr]) = value;
		else 
		if (size == SIZE16)
			*((int_16 *)&board[card].ram_addr[target_addr]) = value;
		else
		if (size == SIZE32)
			*((int_32 *)&board[card].ram_addr[target_addr]) = value;
		else
			mess(1, 1, "writeRAM unknown size\n");

	} /* End non-PCI */

} /* End writeRAM */  
/* -------------------- Begin readRAM  ---------------------- */
unsigned long int readRAM(int target_addr, int size, int card, int bus)
{ /* Begin readRAM */

	unsigned long int retVal;
	int_8  val8 = 0;
	int_16 val16 = 0;
	int_32 val32 = 0;

	if (bus == PCI_CTL)
	{ /* Begin PCI_CTL */

		if (size == SIZE8)
			val8 = *((int_8 *)&board[card].control_addr[target_addr]);
		else 
		if (size == SIZE16)
			val16 = *((int_16 *)&board[card].control_addr[target_addr]);
		else
		if (size == SIZE32)
			val32 = *((int_32 *)&board[card].control_addr[target_addr]);
		else
			mess(1, 1, "writeRAM unknown size\n");

	} /* End PCI_CTL */
	else
	{ /* Begin normal */

		if (size == SIZE8)
			val8 = *((int_8 *)&board[card].ram_addr[target_addr]);
		else 
		if (size == SIZE16)
			val16 = *((int_16 *)&board[card].ram_addr[target_addr]);
		else
		if (size == SIZE32)
			val32 = *((int_32 *)&board[card].ram_addr[target_addr]);
		else
			mess(1, 1, "readRAM unknown size\n");
	} /* End normal */

	retVal = val8 | val16 | val32;
	return(retVal);

} /* End readRAM */

/* -------------------- Begin init ---------------------- */
int init(int digiFD)
{ /* Begin init */

	int x; 
	int kmem_fd; 
	unsigned int ram_size;
	short iopl_active = FALSE;

	/*
	 * This function initializes our internals.  It should *NOT* touch
	 * any hardware itself.
	 */

	/* NOTE: MMAP_DEVICE is picked up from config.h, which was autoconf'ed */
	kmem_fd = open(MMAP_DEVICE, O_RDWR);

	if (kmem_fd < 0 )
	{
		mess(0, 1, "Open of /dev/kmem failed, errno %d\n", errno);
		return( 1 );
	}

	for (x = 0; x < MAXCARDS; x++) { /* Begin for each board */
		struct digi_info *di = &board[x].digi_info ;
		di->info_bdnum = x ; 
		if (ioctl(digiFD, DIGI_GETINFO, di) < 0) {
			/* Should we fall completely out here (Not just continue) ? */
			continue;
		}
		board_count++;
		port_count += di->info_nports;

		/*
		 * The membase returned is the physical address of the device
		 * in memory.  Example 0xd0000, 0xd8000 ...
		 */

		/* Make ram size temporarily = 64K.  This is abritrary */
#if 0
		/* CWS removed these */
		if (di->info_bdtype < PC64XE)
		  ram_size = RAM_SIZE;
		else
#endif
		
		ram_size = (RAM_SIZE * 2);
			
		board[x].ram_addr = mmap (NULL, ram_size, PROT_READ|PROT_WRITE, 
		      	                   MAP_SHARED|MAP_FILE , kmem_fd, (off_t)di->info_physaddr) ;

		/*
		 * SHK - 11/13/2003
		 * The old "less than 0" test here was simply wrong.
		 * Per the man page of mmap:
		 * On error, MAP_FAILED  (-1)  is returned, and errno is set appropriately.
		 */
		if (board[x].ram_addr == MAP_FAILED) {
			perror("Mmap failed");
			mess(0, 1, "Mmap failed on ram addr , errno %d\n", errno);
			return(1);
			
		}

		/* 
		 * We check the type identifier to see if the board is PCI.
		 * The identifiers for non-PCI boards are all smaller than
		 * the identifier for the PCIXEM.  If the board is non-PCI
		 * we map the boards i/o ports.  Since most Digi PCI boards
		 * do not have a seperate address space (None that this
		 * program supports), we must insure that the program
		 * does not attempt to map i/o for these boards.  The
		 * below check does that.
		 */ 
		if (di->info_bdtype < PCIXEM) {
		/* Begin the board detected is NOT a PCI board */

			/* Crank open I/O addresses */

			/* 
			 * iopl gives us access to all possible 65536 ports.
			 * It would be better to just open access to the
			 * ports we needed, but the closest routine that
			 * does that (ioperm) does not handle ports
			 * larger than 0x3ff.  Since we also support
			 * EISA this will not
			 * work.
			 */

			if (!iopl_active)
			{
				if (iopl(3))
					perror("iopl failed ");
				iopl_active = TRUE;
			}

			/* Turn off this card's memory so it doesn't conflict with
			 * initialization of other cards */
			 memoff( x );

		 	/* End the board detected is NOT a PCI board */
		} else {
			/* Begin the board detected IS a PCI board */
			
			/* 
			 * Most Digi PCI cards have their control registers
			 * mapped into memory as opposed to i/o space.  We
			 * map those registers in here.
			 */

			board[x].control_addr = mmap (NULL, RAM_SIZE, PROT_READ|PROT_WRITE, 
			                       MAP_SHARED|MAP_FILE , kmem_fd, (off_t)di->info_ioport) ;


			if (board[x].control_addr < 0 ) {
				mess(0, 1, "Mmap failed on PCI control space, errno %d\n", errno);
				return(1);
			}

		} /* End the board detected IS a PCI board */


	} /* End for each board */

	if (board_count == 0) {
		mess(0, 1, "No boards detected.  Exiting now\n");
		return(1);
	}

	/* Turn off the poller.  Bad things happen to those who are squirting 
	 * in code when the poller comes along. */
	if (ioctl(digiFD, DIGI_POLLER, 1) < 0  ) {
		mess(0, 1, "Can't ioctl DIGI_POLLER %d %d\n", digiFD, errno); 
		return( 1 );
	}
	return( 0 );

} /* End init */

/* -------------------- Begin reset_cards ------------------- */
int reset_cards(int card) { /* Begin reset_cards */

	int index; 
	unsigned long temp;
	struct digi_info *di = &board[card].digi_info ;
	int_16 shiftedAddr;
	int_8  lowAddr, highAddr;
	
	/* Do whatever it takes to reset the card, and make it available.
	 * This includes asserting reset, mapping the memory address, and 
	 * interrupts and possibly setting the windowing mode (Dependant on 
	 * the card type. */
 
	/*
	 *
	 * For most Digi cards (Including the PCXEM family) the formula for
	 * calculating the address is as follows :
	 * 
	 * Set bitfields in t to select interrupt.
	 *
	 * disabled ->  0000h	3 ->  0001h
	 *       5 ->  0002h	7 ->  0003h
	 *      10 ->  0004h  11 ->  0005h
	 *      12 ->  0006h  15 ->  0007h
	 *
	 * Set bitfields 0xFF80 ot according to the bit fields
	 * 0xFF8000h of the base address.
	 *
	 * devb[2] = t and 00FFh
	 * devb[3] = t / 100h 
	 *
	 * For example, a XEM board with interrupt 5, and base address of 
	 * 0x00D8000h would have the below settings:
	 *
	 * t = 0xD80h + 0002h
	 *
	 * devb[2] = 82h = t and 00FF
	 * devb[3] = 0Dh = t / 100h
	 * 
	 * The PC/XE and PC/XEVE are a little different because you have 
	 * the option of setting a window.  If no window is set the result
	 * would be the same (exactly) as the formula above.  However, if
	 * windows are to be set bit position 0010h should be set in 't'
	 * above.
	 *
	 * For example, a XE board with interrupt 5, and base address of 
	 * 0x00D8000h with windows on would have the below settings:
	 *
	 * t = 0xD80h + 0002h + 0010h <- Set windows on
	 * 
	 * devb[2] = 92h = t and 00FF
	 * devb[3] = 0Dh = t / 100h
	 *
	 * Look at the example of the cx configuration on page 17 of the
	 * fep writers guide for another example.  
	 *
	 */


	switch (di->info_bdtype)
	{ /* Begin switch di->type */

		case ISACX:/*dbu*/
		case PCXR:
		case PCXEM:

			outb(0x04, (int)di->info_ioport + 0); /* Hold card in a reset */

			for (index = 0; (inb((int)di->info_ioport) & 0x0e) != 0x4; index ++) 
			{
				if (index  > 5) 
				{
/*dbu*/
					if (((di->info_bdtype) == ISACX))
						mess(1, 1, "reset_card : CX not resetting\n");	
/*dbu*/
					else if ((di->info_bdtype) == PCXR )
						mess(1, 1, "reset_card : PCXR not resetting\n");	

					else
						mess(1, 1, "reset_card : XEM/CX not resetting\n");	

					return(1);
				}
				nap(200);  /* 1/5th of a second */
			}

			shiftedAddr = ((long)(di->info_physaddr)) >> 8;

			lowAddr = shiftedAddr & 0xFF;
			outb(lowAddr, (int)di->info_ioport + 2); 

			highAddr = shiftedAddr >> 8;
			outb(highAddr, (int)di->info_ioport + 3);

			break;

	  		outb (0x02, (int)di->info_ioport + 4 ) ; /* You put your reset in */
	  		nap(200);  /* Dictated by capacitor on reset line */
	  		outb (0x00, (int)di->info_ioport + 4 ) ; /* You take your reset out */

	  		nap(200);  /* Wait for B.O.X. to start reset */

	 		/* Wait for board to come out of a reset */
			for (index = 0; inb((int)di->info_ioport) == 0; index ++) {
				if (index  > 20 ) {
					mess(1, 1, "reset_card : SmartPort not resetting\n");	
					return(1);
				}
				nap(200);  /* 1/5th of a second */
			}
	  		break ; 

		case PCIXR:
		case PCICX:/*dbu*/
		case PCICX_IBM:
		case PCI920_2:
		case PCI920_4:
		case PCI920_8:
		case PCIXR_422:
		case PCIXR_IBM:
		case PCIXRJ:
		case PCIXEM:
		case PCIEPCJ:
			writeRAM(0, 0x04, SIZE8, card, PCI_CTL); /* Hold card in reset */
			for (index = 0; ((temp=readRAM(0x0, SIZE8, card, PCI_CTL)) & 0x0e) != 0x4; index ++) 
			{
				if (index  > 5) 
				{
					mess(1, 1, "reset_card : PCI not resetting\n");	
					return(1);
				}
				nap(200);  /* 1/5th of a second */
			}

			break;

		case PCXE:
			/* 
			 * The format of the below is a little different
			 * than the PCXEM case.  The values are calculated
			 * exactly the same as above, BUT if 8K windows are
			 * to be used the value 0x10 has to be added to 
			 * the low byte.  Hence 0x00 (For 0xd0000) becomes
			 * 0x10.  This only occurs if 8K windowing is desired.
			 * 
			 * The only options for the new XE (glasgo) is 8*8K
			 * windows or 2*32K windows. The driver knows nothing
			 * about 32K windows. 
			 * 
			 * Below we automaticly choose windowing.
			 * We always choose windowing.
			 */

			shiftedAddr = ((long)(di->info_physaddr)) >> 8;

			/* We set bit 0x10 to enable windows */
			lowAddr = (shiftedAddr + 0x10) & 0xFF;
			outb(lowAddr, (int)di->info_ioport + 2); 

			highAddr = shiftedAddr >> 8;
			outb(highAddr, (int)di->info_ioport + 3);

			/* ---------------- Start code to reset board */
			outb(0x06, (int)di->info_ioport + 0); /* Hold card in a reset */

			/*
			 * NOTE : The manual talks about enabling the XE
			 *        memory.  They would use an 0x6 above
			 *        and below.  We have been using a 0x4
			 */

			for (index = 0; (inb((int)di->info_ioport) & 0x0e) != 0x6; index ++) 
			{
				if (index  > 5) 
				{
					mess(1, 1, "reset_card : PCXE not resetting\n");	
					return(1);
				}
				nap(200);  /* 1/5th of a second */
			}
			break;

		case PCXI:

			/*
			 * Note, the PCXI has its memory options set via.
			 * dip switch.  This cannot be changed dynamically
			 * hence (unlike above) you will not see code here
			 * programming the memory address or the windowing
			 * disposition.
			 */
 
			/* --------- Start code to reset board --------- */
			outb(0x06, (int)di->info_ioport + 0); /* Hold card in a reset */


			for (index = 0; (inb((int)di->info_ioport) & 0x0e) != 0x6; index ++) 
			{
				if (index  > 5) 
				{
					mess(1, 1, "reset_card : PCXI not resetting\n");	
					return(1);
				}
				nap(200);  /* 1/5th of a second */
			}
			break;

	} /* End switch di->type */
	return (0) ;    

} /* End reset_cards */

/* -------------------- Begin binary_is_running ----------------- */
int binary_is_running(int card, int binary_type)
{ /* Begin binary_is_running */

	struct digi_info *di = &board[card].digi_info ;
	int ret_val = 0 ; 

	/* When necessary, turn on card RAM */

	switch(di->info_bdtype)
	{
		case ISACX:/*dbu*/
		case PCXR:
		case PCXEM:
			outb(0x80 , (int)di->info_ioport + 1);
			break;
		case PCXE:
			outb(0x02 , (int)di->info_ioport + 0 );
			break;
	}

	switch (binary_type)   
	{ /* Begin switch binary type */
		case FEP:
		{
			char fep_sig0;
			char fep_sig1;

			if(di->info_bdtype == ISACX) {
				fep_sig0 = readRAM(0xc00, SIZE8, card, NORMAL);
				fep_sig1 = readRAM(0xc01, SIZE8, card, NORMAL);
			} else {
				fep_sig0 = readRAM(0xd20, SIZE8, card, NORMAL);
				fep_sig1 = readRAM(0xd21, SIZE8, card, NORMAL);
			}

			/* If the FEP boots correctly, it should return the ascii
			 * string "OS" in locations 0xD20 and 0xD21. */
			if ((fep_sig0 == 'O') && (fep_sig1 == 'S') )
				ret_val=1;
			
			break ; 
		}

		case BIOS:
		{
			char bios_sig0;
			char bios_sig1;

			bios_sig0 = readRAM(0xc00, SIZE8, card, NORMAL);
			bios_sig1 = readRAM(0xc01, SIZE8, card, NORMAL);

			/* 
			 * If the BIOS boots correctly, it should return
			 * the ascii string "GD" in locations 0xC00 and 0xC01.
			 */
			if (( bios_sig0 == 'G') && (bios_sig1 == 'D') ) 
				ret_val=1;
                    	
		}
		
	} /* End switch binary type */
	return ret_val ;

} /* End binary_is_running */

/* -------------------- Begin wait_for_binary ----------------- */
int wait_for_binary(int card, int binary_type)
{ /* Begin wait_for_binary */

	int interations ; 

	/* 
	 * Note when I set this for 30 interations (6 seconds) 
	 * it will only boot if there is one concentrator.  Doubling the
	 * time allows 2 concentrators.  240 is a lot, but it should
	 * handle 8 concentrators.
	 */
	for (interations = 0; interations < 480; interations++)
	{
		if (binary_is_running(card, binary_type))
		{
			return(0);
		}
		nap(200); /* Expressed in milliseconds == 1/5 second */
	}

	mess(0, 1, "Could not start %s on card %d\n", 
	binary_type == FEP ? "FEPOS" : "BIOS", card);
	return(1);

} /* End wait_for_binary */

#if 0
void memon( int card )
{
	struct digi_info *di = &board[card].digi_info ;


	/* Select the card memory */
	switch( di->info_bdtype ) {
		case PCXE:
			outb(0x02, (int)di->info_ioport+0 );
			break;
		case PCXEM:
			outb(0x80, (int)di->info_ioport + 1);
			break;
		default:
			break;


	}

}
#endif

void memoff( int card )
{
	struct digi_info *di = &board[card].digi_info ;


	/* Deselect the card (to make way for the next one) */
	switch( di->info_bdtype ) {
		case PCXE:
		case PCXI:
			outb(0x00, (int)di->info_ioport );
			break;
		case ISACX:
		case PCXR:
		case PCXEM:
			outb(0x00, (int)di->info_ioport + 1);
			break;
		default:
			break;

	}

}

void cxmagic( int card, struct digi_info *di, int epc ) {
	unsigned char *pt;
	unsigned int i;
	unsigned int *index;
	unsigned int *no_images;

#ifdef CWSDEBUG
	printf("Writing %sC/X card magic string\n", epc? "EP" : "" );
#endif
	if( epc ) {
		index = &epcxindex;
		no_images = &no_epcximages;
	} else {
		index = &cxindex;
		no_images = &no_cximages;
	}
	if( *index < *no_images ) {
	  	/* Enable memory on non-PCI cards */
		if ( di->info_bdtype != PCICX && di->info_bdtype != PCICX_IBM &&
		     di->info_bdtype != PCIEPCJ){
			outb(0x80, (int)di->info_ioport + 1);
		}

		/* 0xCD0 is magic EPCX concentrator config location */
		i=0;
		pt = (unsigned char *)&(board[card].ram_addr[0xcd0]);
		do {
			if( epc )
				*pt = epcx_cfg[*index][i];
			else
				*pt = cx_cfg[*index][i];
			++i;
		} while( *pt++ != 0xff );

		/* Increment the global so we use the next magic string
		 *  on the next call to cxmagic (for the next card)
		 */
		(*index)++;
	} else {
		mess(1, 1,
	"Cannot find a configuration string for this %sCX concentrator\n",
		epc? "EP" : "" );
	}
}

/* -------------------- Begin load_binary -------------------- */
int load_binary(int card, int binary_type)
{ /* Begin load_binary */

	struct file_map *fm = file_map; 
	unsigned int cmdInPtr, target_window;
	unsigned long memory_seg=0, target_addr, target_offset = 0;
	int binaryFD, index, nbytes, ret_val = 0; 
	struct digi_info *di = &board[card].digi_info ;

	/* Note: buf size needs to be precisely 1024 for CONC files */
	char *temp_name, buf[1024];


	for (fm = file_map ; fm->binary_type != -1 ; fm++ )
	{ /* Begin loop until we run out of binary files. */

		if (( fm->binary_type == binary_type)  &&
		    ( fm->board_type == di->info_bdtype ) )
		{
			/* If this board doesn't get this type of binary, bail */
			if ( fm->fname == NULL )
			{
				/* Note: this is not an error condition */
				return(0) ;
			}
			break;
		} 

	} /* End loop until we run out of binary files. */
    
	
	/*
	 * The PCXI and PCXE can have varying amounts of memory onboard.
	 * Because of this we need to calculate where that memory ends,
	 * and therefore where we can load bios info.
	 */
	if (di->info_bdtype == PCXI)
	{
		/* 
		 * This memory_seg assumes the XI or XE is a 
		 * 64K board.  We need to add code to query
		 * this to be sure.  See internal release 110c. 
		 */
		
		memory_seg = 0xf000;  
	}

	target_addr = fm->addr ;
	temp_name = (char *)malloc((size_t)((strlen(fm->fname)) +
					    strlen(FWPATH) + 1)); 

	sprintf(temp_name, "%s/%s", FWPATH, fm->fname);
	if (verbose) {
		mess(0, 1, "Downloading %s to %lx on %s\n", fm->fname, (long) fm->addr, btypes[di->info_bdtype].desc );
	}

	/* 
	 * Note: programmer can do something stupid here and pass
	 * a non-buf size alignable value that will make this copy 
	 * fail.  Don't do that.
	 */
 

	if ((binaryFD = open( temp_name, O_RDONLY )) < 0 ) 
	{
		mess(0, 1, "Open of %s failed, errno %d\n", temp_name, errno);
		return( 1 );
	}

	free((void *)temp_name);

	/* DBU: special for CX: before loading FEP code, put the CX config string
	 * onboard
	 * PEB: special for SPXX: before loading FEP, setup some pointers
	 */
	if( binary_type == FEP ) {
		switch (di->info_bdtype) {
			case PCICX:
			case PCICX_IBM:
			case ISACX:
				cxmagic( card, di, 0 );
				break;
			
			case PCIEPCJ:
				cxmagic( card, di, 1 );	
				break;
		}
	}

	/* Now, slurp the whole binary in, jam it on the card */
	while ((nbytes = read (binaryFD, buf, sizeof(buf))))
	{ /* Begin while read loop */

		/* 
		 * This block does whatever is necesary to
		 * prepare the card to receive this chunk.
		 */

		/* First get all the data on the card; switch windows as necessary */
		switch (di->info_bdtype)
		{ /* Begin switch di->info_bdtype */

			case PCICX:
			case PCICX_IBM:
			case PCIXR:
			case PCI920_2:
			case PCI920_4:
			case PCI920_8:
			case PCIXR_422:
			case PCIXR_IBM:
			case PCIXRJ:
			case PCIXEM:
			case PCIEPCJ:
				switch (binary_type)
				{ /* Begin switch binary_type */

					case BIOS:
					case FEP:
						target_offset = target_addr;
						break;
	
				} /* End switch binary_type */
				break;

			case ISACX:
				switch (binary_type)
				{ /* Begin switch binary_type */

					case BIOS:
						target_window = 0xf;
						outb(0x80 | target_window , (int)di->info_ioport + 1);
						target_offset = target_addr;
						break;
					case FEP:
						target_window = (target_addr >> 15) + 1; 
						outb(0x80 | target_window , (int)di->info_ioport + 1);
						target_offset = target_addr;
						break;

				} /* End switch binary_type */
				break;

			case PCXR:
			case PCXEM:
				switch (binary_type)
				{ /* Begin switch binary_type */

					case BIOS:
					case FEP:

						/* 
						 * The below will cause the a
						 * window switch if window
						 * is about to be exceeded.
						 */

						target_offset = target_addr & (RAM_SIZE - 1);
						target_window = target_addr >> 15; 
						outb(0x80 | target_window , (int)di->info_ioport + 1);
						break;
				} /* End switch binary_type */
				break;

			case PCXE:
				switch (binary_type)
				{ /* Begin  switch binary_type */

					case BIOS:
			/* The below board[card].ram_addr already points
                         * at the board mapped in at physical memory.
                         * If target_addr was set to 0x1800 (Assuming
                         * 2k bios) the same code should NEARLY work.
                         * Watch out for the shift 15
			 */

					        /* We mask against 0x2000 because it is an 8K window */
						target_offset = target_addr & (0x2000 - 1);
						/* Later develop code just in case bios > 2K */
						/* FIXME : Do not depend on addr = 0x1800 and window = 0xf */

						outb(0x80 | 0xf , (int)di->info_ioport + 1);
						break;

					case FEP:
						/* We mask against 0x2000 because it is an 8K window */
						target_offset = target_addr & (0x2000 - 1);
						/* Select window 1 always */
						outb(0x80 | 0x1 , (int)di->info_ioport + 1);

						break;
			} /* End switch binary_type */
			break;

		case PCXI:
			switch (binary_type)
			{ /* Begin  switch binary_type */

				case BIOS:
					outb(0x6 , (int)di->info_ioport);
					target_offset = target_addr - (memory_seg << 4);

				break;

				case FEP:
					target_offset = target_addr;
				break;

			} /* End switch binary_type */
			break;

		} /* End switch di->info_bdtype */


		/* Now, we write the chunk to the card */

/* Some PCI Chipsets fail on some multi-byte (e.g: 32-bit) transfers.  With
 * these, try doing bytewise transfers instead. */
#ifdef PCIBUG
		slow_memcpy((void *)&board[card].ram_addr[target_offset], (const void *)buf, nbytes);

#else
		memcpy((void *)&board[card].ram_addr[target_offset], (const void *)buf, nbytes);
#endif
		target_addr += nbytes ; 

	} /* End while read loop */

	/*
	 * Note : When we arrive here the memory window we are currently
	 * in is NOT guaranteed to be 0.
	 */


	/* Consider puting a check to make sure the code got there okay */
	/* Now, all the code is on the card.  We just have to execute it. */
	switch (di->info_bdtype)
	{ /* Begin switch di->info_bdtype */

		case PCICX:
		case PCICX_IBM:
		case PCIXR:
		case PCI920_2:
		case PCI920_4:
		case PCI920_8:
		case PCIXR_422:
		case PCIXR_IBM:
		case PCIXRJ:
		case PCIXEM:
		case PCIEPCJ:
		{ /* Begin PCI XR, XRJ, XRM case */

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:
					/* Set boot vector */
					/* Note - these are long words */

					writeRAM(0, 0x0bf00401, SIZE32, card, NORMAL);
					writeRAM(4, 0x0, SIZE32, card, NORMAL);
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					break;

				case FEP:

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);
					/* Set initial address and BIOS command */
					writeRAM(0xc34, 0xbfc01004L, SIZE32, card, NORMAL);
					writeRAM(0xc30, 0x3L, SIZE32, card, NORMAL);

					break;

			} /* End switch binary_type */

			/* Release reset now */
			writeRAM(0, 0x0, SIZE8, card, PCI_CTL);

			/* Wait for binary (BIOS or FEP) to boot */
			if(wait_for_binary(card, binary_type))
				return(1);

			break;
		} /* End PCI XR, XRJ, XRM case */


		case ISACX:
		{ /* Begin case ISA/EISA CX */

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:

					/* Put us back in window 0 */
					outb(0x80, (int)di->info_ioport + 1);

					/* Clear confirm word */
					writeRAM(0, 0x0c00, SIZE16, card, NORMAL);

					/* Release reset now */
					outb(0x00, (int)di->info_ioport);

					/* Wait for binary to boot */
					wait_for_binary(card, binary_type);
					break;

				case FEP:
					/* Select page 0 and enable memory */
					outb(0x80, (int)di->info_ioport + 1);

					/* Form BIOS execution request */
					writeRAM(0xc40, 0x1, SIZE16, card, NORMAL);
					writeRAM(0xc42, 0xd00, SIZE16, card, NORMAL);
					writeRAM(0xc44, 0x4, SIZE16, card, NORMAL);

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					outb((inb((int)di->info_ioport) & ~0x8), (int)di->info_ioport);
					outb((inb((int)di->info_ioport) | 0x8), (int)di->info_ioport);
					/* Wait for binary to boot */
					wait_for_binary(card, binary_type);

					break;

			} /* End switch binary_type */

			break;
		} /* End case ISA/EISA CX */



		case PCXR:
		case PCXEM:
		{ /* Begin case PCXEM */

			/* Put us back in window 0 */
			outb(0x80, (int)di->info_ioport + 1);

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:
					/* Set boot vector */
					/* Note - these are long words */

					writeRAM(0, 0x0bf00401, SIZE32, card, NORMAL);
					writeRAM(4, 0x0, SIZE32, card, NORMAL);
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					break;

				case FEP:

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					/* Set initial address and BIOS command */
					writeRAM(0xc34, 0xbfc01004L, SIZE32, card, NORMAL);
					writeRAM(0xc30, 0x3L, SIZE32, card, NORMAL);

					break;

			} /* End switch binary_type */

			/* Release reset now */
			outb(0x00, (int)di->info_ioport + 0);

			/* Wait for binary (BIOS or FEP) to boot */
			if(wait_for_binary(card, binary_type))
				return(1);

			break;
		} /* End case PCXEM */

		case PCXE:
		{ /* Begin case PCXE */
			/* Put us back in window 0 */
			outb(0x80, (int)di->info_ioport + 1);

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:

					/* Clear confirm word */
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					/* Release reset now */
					outb(0x00, (int)di->info_ioport + 0);
	
					if(wait_for_binary(card, binary_type))
						return(1);;
					break;

				case FEP:

			/*
			 * Do not forget to byte swap these for little endian.
			 * 0c40 <- 0002h     BIOS request to move program
			 * 0c42 <- f200      seg + 200h
			 * 
			 *                  = f2000002h
			 * c44  <- 0000h
			 * c46  <- 0200h
			 * 
			 *                  = 02000000h
			 * c48  <- 0000h
			 * c4a  <- 2000h	
			 *                  = 20000000h
			 */

			/* 
			 * Interestingly enough the 8E glasgo does not
			 * need a command to move the fep program since
			 * dual port ram is shadowed over local memory.
			 *
			 * Additionally while the Glasgo offers the standard
			 * 8 * 8K windows, it also offers 2 * 32K windows.
			 * It does NOT support the older 64K window.
			 *
			 * The below command is not needed, but we
			 * will keep it for XEVE compatability.  As long
			 * as we use the 8 * 8K windows the XEVE and XE will
			 * look identical to the driver.  They only differ
			 * when using the 2 * 32K Glasgo window mode or
			 * the 1 * 64K XEVE window mode.
			 */
 

			/* Form BIOS request to move program to 0x2000 */
					writeRAM(0xc40, 0xf2000002L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x02000000L, SIZE32, card, NORMAL);
					writeRAM(0xc48, 0x20000000L, SIZE32, card, NORMAL);

					/* 
					 * NOTE : The int below is NOT
					 * asserted on the host bus unless
					 * the function code in 0xc40 = 0.
					 * In this case ints are generated
					 * on the card AND host.  In all
					 * other situations the int is 
					 * generated just on the card.
					 */
 
					/*
					 * Toggle board interrupt.  This is
					 * done to make the BIOS run the
					 * command.
					 */
					outb(0x8, (int)di->info_ioport);
					outb(0x00, (int)di->info_ioport);


					for (index = 0; index < 4; index++)
					{
						if ((readRAM(0xc40, SIZE16, card, NORMAL)) == 0)
							break; /* We got a 0, we can leave this loop now ! */
						sleep(1);
					}

					if (index == 4) {
						mess(0, 1, "Attempt to toggle interrupt failed.\n");
						return(1);
					}
						

					/* Form BIOS request to execute FEPOS */
					writeRAM(0xc40, 0x02000001L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x4, SIZE16, card, NORMAL);

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					/*
					 * Toggle board interrupt.  This is
					 * done to make the BIOS run the
					 * command.
					 */
					outb(0x08, (int)di->info_ioport);
					outb(0x0, (int)di->info_ioport);

					/* Wait to see if the FEP comes up (BOOTS) */
					if(wait_for_binary(card, binary_type))
						return(1);;



					/* Get command in pointer */
					cmdInPtr = readRAM(0xd10, SIZE16, card, NORMAL);

					/* Set buffer space command */
					writeRAM((0x400 + cmdInPtr), 0x92, SIZE8, card, NORMAL);
					writeRAM((0x401 + cmdInPtr), 0x0, SIZE8, card, NORMAL);

					/*
					 * Set up for 32L buffers allow each
					 * receive, transmit buffer to have
					 * its on 8K window.
					 */

					writeRAM((0x402 + cmdInPtr), 0x8000, SIZE16, card, NORMAL);

					/* Advance the command in pointer */
					writeRAM(0xd10, (cmdInPtr + 4), SIZE16, card, NORMAL);
					break;

			} /* End switch binary_type */

			break;
		} /* End case for PCXE and PCXEVE */

		case PCXI:
		{ /* Begin case PCXI */
			switch (binary_type)
			{ /* Begin  switch binary_type */

				case BIOS:

					/* Clear confirm word */
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					/* Release reset now, but leave memory enabled */
					outb(0x02, (int)di->info_ioport);
	
					if(wait_for_binary(card, binary_type))
						return(1);;

				break;

				case FEP:
			/*
			 * Do not forget to byte swap these for little endian.
			 *
			 * 0c40 <- 0002h     BIOS request to move program
			 * 0c42 <- f200      seg + 200h  (f000 for a 64K board)
			 *
			 *			= f2000002h
			 *
			 * c44  <- 0000h
			 * c46  <- 0200h
			 *
			 *	                = 02000000h
			 * c48  <- 0000h
			 * c4a  <- 2000h	
			 *			= 20000000h
			 */

					outb(0x02, (int)di->info_ioport); /* Leave memory enabled */


					/* Form BIOS request to move program to 0x2000 */

					writeRAM(0xc40, (0x02000002L + (memory_seg << 16)),
					         SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x02000000L, SIZE32, card, NORMAL);
					writeRAM(0xc48, 0x20000000L, SIZE32, card, NORMAL);
					/* 
					 * NOTE : The int below is NOT
					 * asserted on the host bus unless
					 * the function code in 0xc40 = 0.
					 * In this case ints are generated
					 * on the card AND host.  In all
					 * other situations the int is 
					 * generated just on the card.
					 */
 
					/*
					 * Toggle board interrupt.  This is
					 * done to make the BIOS run the
					 * command.
					 */
					outb(0x0A, (int)di->info_ioport);
					outb(0x02, (int)di->info_ioport); /* Leave memory enabled */

					for (index = 0; index < 4; index++)
					{
						if ((readRAM(0xc40, SIZE16, card, NORMAL)) == 0)
							break; /* We got a 0, we can leave this loop now ! */
						sleep(1);
					}

					if (index == 4) {
						mess(0, 1, "Attempt to toggle interrupt failed.\n");
						return(1);
					}
						

					/* Form BIOS request to execute FEPOS */
					writeRAM(0xc40, 0x02000001L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x0004, SIZE16, card, NORMAL);

					/* Clear confirm word */
					writeRAM(0xd20, 0x0000, SIZE16, card, NORMAL);

					/*
					 * Toggle board interrupt.  This is
					 * done to make the 
					 * BIOS run the command.
					 */
					outb(0x0A, (int)di->info_ioport);
					outb(0x02, (int)di->info_ioport);
					/* Wait to see if the FEP comes up (BOOTS) */
					if(wait_for_binary(card, binary_type))
						return(1);
					break;

			} /* End switch binary_type */

			break;
		} /* End case PCXI and PC64XE */

	} /* End switch di->info_bdtype */

	return(ret_val);	      

} /* End load_binary */


int getcxconf( unsigned char cfg[][CXCFGLEN], char *fname, int epc )
{
	int fd;
	unsigned char c;
	unsigned int cx, offset, nbytes; 

	if ((fd = open( fname, O_RDONLY )) < 0 ) 
	{
#if 0
		mess(0, 1, "No %s, using defaults for %sCX concentrators\n",
		fname, epc? "EP" : "" );
#endif
		return( 1 );
	}

	cx = 0;
	offset = 0;
	while ( (nbytes = read (fd, &c, 1 )) ) {
		cfg[cx][offset] = c;
		if( c == 0xff ) {
			  /* We've read a full config string */
#ifdef CWSDEBUG
		  printf("End of config string #%d! (%sC/X card!)\n",
			 cx, epc? "EP" : "" );
#endif                                                                   
		  	/* Update the global */ 
			if( epc )
				++no_epcximages;
			else
				++no_cximages;
			/* Next card */
			++cx;
			offset = 0;
		} else
			++offset;
	}
	close(fd);
	return( 0 );
}

void report_numports( int digiFD ) {
	int card;
	struct digi_info *di;
	int fepnports = 0;

	mess( 0, 0, "Number of ports:\n" );
	for (card = 0; card < board_count; card++) {
		di = &board[card].digi_info ;
		/* Repeat this, because "numports" is not available
		 * on PCI cards until after the card is booted */
		if (ioctl(digiFD, DIGI_GETINFO, di) < 0)
			mess( 0, 0, "(second DIGI_GETINFO failed!)..." );
		else {
			switch(di->info_bdtype) {
				case PCXE:
				case PCXI:
				case PCXEM:
				case PCXR:
				case ISACX:
				
					outb(0x80, di->info_ioport + 1);
			 		break;
			 	
			 	default:
			 		break;
			 }
			
			switch(di->info_bdtype) {
				case PCXE:
				case PCXI:
					fepnports = *(char*)(board[card].ram_addr + XEPORTS);
					break;
				
				case ISACX:
				case PCXEM:
				case PCXR:
				case PCICX:
				case PCICX_IBM:
				case PCIXEM:
				case PCIXR:
				case PCIXRJ:
				case PCI920_4:
				case PCI920_8:
				case PCI920_2:
				case PCIXR_422:
				case PCIXR_IBM:
				case PCIEPCJ:
					fepnports = (unsigned short)*(unsigned char *)(board[card].ram_addr + XEMPORTS);
					break;
					
				default:
					break;
			}
			
		
			switch(di->info_bdtype) {
				case PCXE:
				case PCXI:
				case PCXEM:
				case PCXR:
				case ISACX:
					outb(0x0, di->info_ioport + 1);
				
				default:
					break;
			}
		
		        mess( 0, 1, "Card #%d(%s): card reports %d ports found\n",
		        card, btypes[di->info_bdtype].desc, fepnports );
		                                                
			if(fepnports == di->info_nports) continue;
		
			/* Error condition! Lets tell the user... */

			printf("\n");
			mess(2, 1, "User specified %d ports, but FEP/OS reports %d ports!\n",
				di->info_nports,		
				fepnports);
		
		}

		
		if(di->info_bdtype == PCIXEM || di->info_bdtype == PCXEM ) {
			if(fepnports == 0) {
				printf("\n");
				mess(3, 1, "Ensure that all Ports/EM modules are properly connected.\n");
				printf("\n");
			}
			
			if(fepnports == 64) {
				printf("\n");
				mess(3, 1, "1> Ensure that the proper number of modules are connected.\n");
				mess(3, 1, "2> Make sure that the cable attached to your host adaptor\n");
				mess(3, 1, "   plugs into the jack labeled 'EBI IN' on Ports/EM module\n"); 
				mess(3, 1, "   #1.\n");
				printf("\n");
			}
		}
	
	
		if(di->info_bdtype == PCXI ||
		   di->info_bdtype == PCXE ||
		   di->info_bdtype == PCXR || 
		   di->info_bdtype == PCIXR_422 ||
			di->info_bdtype == PCIXR_IBM) {
		   	printf("\n");
		   	mess(3, 1, "Make sure you've specified the correct number of ports for\n");
		   	mess(3, 1, "your digiBoard.\n");
			printf("\n");
		}
	
		if(di->info_bdtype == ISACX ||
		   di->info_bdtype == PCICX ||
		   di->info_bdtype == PCICX_IBM ||
		   di->info_bdtype == PCIEPCJ) {
		   	printf("\n");
		   	mess(3, 1, "1> Make sure you specify the correct number of ports for\n");
		   	mess(3, 1, "   your digiBoard.\n");
		   	mess(3, 1, "2> Make sure you specify the correct concentrator\n");
		   	mess(3, 1, "   configuration information for your digiBoard.\n");
		   	printf("\n");
		 }	
	}
}

/* Copy concentrator images to the driver */

int con_init( int digiFD ) {
	struct stat stat_buf;
	char buf[CONC_CHUNKSIZE];
	CONC_UPLOAD conc_chunk;
	int con_fd, nbytes;
	extern int errno;

	unsigned int zero = 0;

	char *con_fname;
	int cur_fname = 0;

	/* Reset the driver load tables (clearing any preloaded images) */
	if (ioctl(digiFD, DIGI_CON_INIT, &zero) < 0  ) {
		mess(0, 0, "Can't ioctl DIGI_CON_INIT to reset: %d\n",  errno); 
		return(1);
	}

	for( ; con_fnames[cur_fname]; cur_fname++ ) {
		con_fname = con_fnames[cur_fname];

		if( stat( con_fname, &stat_buf ) ) {
			printf( "Cannot stat '%s'\n", con_fname );
			exit(1);
		}
		if ((con_fd = open( con_fname, O_RDONLY )) < 0 )  {
			printf("Open of %s failed, errno %d\n", con_fname, errno);
			return( 1 );
		}
		if (ioctl(digiFD, DIGI_CON_INIT, &stat_buf.st_size) < 0  ) {
			mess(0, 0, "Can't ioctl DIGI_CON_INIT %d\n",  errno); 
			return(1);
		}
		mess(0,1,"Downloading %s to EPCA Memory Space\n",con_fname);
		while ((nbytes = read (con_fd, buf, sizeof(buf)))) {
			conc_chunk.size = nbytes;
			memcpy( conc_chunk.data, buf, nbytes );
			if (ioctl(digiFD, DIGI_CON_NEXT, &conc_chunk) < 0  ) {
				mess(0, 0, "Can't ioctl DIGI_CON_NEXT %d\n",  errno); 
				return(1);
			}
		}
		close( con_fd );
	}

	return( 0 );
}


/* get_majors
 * Query driver for the currently active major numbers
 */
void get_majors( int digiFD ) {
	int bank, major, nbanks;

	nbanks = sizeof(majors)/sizeof(majors[0]);
	bank = 0;
	do {
		major = bank;
		/* Called with a bank number, ioctl returns major
		 * number for that bank.  Invoking the ioctls from
		 * bank 0 in ascending order, we can stop when we have
		 * either a full array or get a 0 return value. */
		if( ioctl(digiFD, DIGI_GETMAJOR, &major ) < 0 ) {
			mess( 1, 0,
			"DIGI_GETMAJOR failed for bank %d!...", bank );
			majors[bank] = -1;
		} else {
			majors[bank] = major;
		}

		major = bank;
		if( ioctl(digiFD, DIGI_GETPRMAJOR, &major ) < 0 ) {
			mess( 1, 0,
			"DIGI_GETPRMAJOR failed for bank %d!...", bank );
			prmajors[bank] = -1;
		} else {
			prmajors[bank] = major;
		}
		++bank;
	} while( bank < nbanks && major != 0 );
}


/* chan_to_majmin
 * Given a channel number, map into the corresponding major and minor numbers
 */ 
int chan_to_majmin( int chan, int *major, int *minor, int printer )
{
	if( printer )
		*major = prmajors[chan/NDEV_MAJOR];
	else
		*major = majors[chan/NDEV_MAJOR];
	*minor = chan % NDEV_MAJOR;

	return( *major );

#if 0
	printf( "chan_to_majmin: chan:%d, major:%d, minor:%d\n",
		chan, *major, *minor );
#endif
	
}


/* chan_to_path
 * Given an channel number, return the corresponding device's pathname
 */
char *chan_to_path( int channo, int printer ) {

#ifdef CONFIG_DEVFS_FS
  	sprintf(dev_path,"%s/%d", "dg/epca", channo);
#else
	char bankname;
	int offset;

	/* First bank is ttyD, next are ttyE, ttyF... */
	bankname = epca_banknames[ channo/NDEV_MAJOR ];
	offset = channo % NDEV_MAJOR;
	sprintf( dev_path, "%s%c%03d", printer? TPR_BASENAME: TTY_BASENAME,
		 bankname, offset );
#endif
	return( dev_path );

}

/* make_nodes
 * Loop through list of registered adapters and unlink any existing
 * device nodes corresponding to these channel numbers, then -- if
 * "make" is set -- creating new nodes with the current major/minor
 * numbers.
 */
void make_nodes( int first, int num, int make ) {

#ifndef CONFIG_DEVFS_FS
	/* Only do this if we're not using DEVFS. */

	int major = 0, prmajor = 0, minor = 0;
	int mode = 0666 | S_IFCHR;
	int channo, ret, printer;
	char path[1024];

	for( channo=first; channo < first+num; channo++ ) {
#if 0
		/* For now, the nodes for the first 256 devices are managed
		 * (unlink and mknod) by digiConfig, not by digiDload. */
		if( channo < NDEV_MAJOR )
			continue;
#endif

		/* TTY device */
		if( chan_to_majmin( channo, &major, &minor, 0 ) < 0 ) {
			mess( 1, 1, "Bad major number (%d) for chan #:%d\n",
			major, channo );
			continue;
		}

		/* TPR (transparent print) device */
		if( chan_to_majmin( channo, &prmajor, &minor, 1 ) < 0 ) {
			mess( 1, 1, "Bad printer major number (%d) for chan #:%d\n",
			major, channo );
			continue;
		}


		for( printer=0; printer<=1; ++printer ) {


			sprintf(path, "/dev/%s",chan_to_path( channo, printer ));
			/* Ignore "No such file..." errors */
			if( (ret = unlink( path )) < 0 && errno != ENOENT )
				mess( 1, 1, "unlink failed: %s\n", strerror(errno), errno );

			/*
		 	* the tty's need to be made in /dev for getty to work unless
		 	* we're using DEVFS..
		 	*/
			if( make && mknod(path, mode,
			makedev(printer? prmajor : major, minor)) )
					mess( 1, 1, "mknod failed: %s\n", strerror(errno) );

#if !defined(CONFIG_DEVFS_FS)
			/* 
			 * only need to make two sets of nodes, in /dev and in
			 * BASENAME, if we're not using DEVFS
			 */
			sprintf(path, "%s/%s", BASENAME, chan_to_path( channo, printer ));
			/* Ignore "No such file..." errors */
			if( (ret = unlink( path )) < 0 && errno != ENOENT )
				mess( 1, 1, "unlink failed: %s\n", strerror(errno), errno );
	
			if( make && mknod(path, mode, makedev(printer? prmajor :  major, minor)) )
					mess( 1, 1, "mknod failed: %s\n", strerror(errno) );
#endif
		}

		/* The "cu" devices are being dropped from the kernel.  In
		 * order not to break old programs we'll create the links
		 * that might have been used in the past (/dev/cud0...cud255)
		 * with fixed DIGICU_MAJOR major number, but no more */

		if(do_cuds) {
			if( channo < NDEV_MAJOR ) {
				sprintf( path, "/dev/cud%d", channo );
				if( (ret = unlink( path )) < 0 && errno != ENOENT )
					mess( 1, 1, "unlink failed: %s\n", strerror(errno) );
				if( make && mknod(path, mode, makedev(DIGICU_MAJOR, channo)) )
					mess( 1, 1, "mknod failed: %s\n", strerror(errno) );
			}
		}
		
	}
#endif
}

/* update_devs
 * unlink old and mkdev current device links
 */
void update_devs() {

#ifndef CONFIG_DEVFS_FS
	/* Only do this if we're not using DEVFS. */

	int crd, make;

	mess( 0, 0, "Creating device nodes:\n" );
	for( crd=0; crd<board_count; crd++ ) {

		if( board[crd].digi_info.info_bdstate == ENABLED ) {
			mess( 0, 1, "Creating device nodes for card #%d\n", crd );
			make = 1;
		} else {
			mess( 1, 1, "Card #%d is disabled.  Device nodes not created.\n",
			crd );
			make = 0;
		}
		make_nodes( board[crd].digi_info.firstchan,
		board[crd].digi_info.info_nports, make );
	}
#endif
}

void doargs( int ac, char *av[] ) {
	int err=0;

	while( --ac ) {
		if( !strcmp( *++av, "v" ) || !strcmp( *av, "-v" ) )
			verbose = TRUE;
		else if( !strcmp( *av, "-nomkdev" ) )
			do_mkdev = 0;
		else if( !strcmp( *av, "--createcuds" ) ) {
			printf("\n\ndoargs() - Forced enabling of cud device creation\n\n");
			forcecuds = 1;
		}
		
		else {
			mess( 1, 1, "Invalid option:'%s'\n", *av );
			err = 1;
		}
	}
	if( err )
		usage();

}


/* ------------------- Begin main ------------------------- */
int main(int argc, char *argv[])
{ /* Begin main */

	int digiFD, card, someone_there, cx_there, epcx_there, no_cxconf, no_epcxconf;
	struct digi_info *di;

	doargs( argc, argv );

	if ((digiFD = open(BASENAME"/digiCtl", O_RDWR)) < 0 ) {
		perror(" cannot open digi download device ");
	}

	if( init(digiFD) )
		return( 1 );
	if( verbose )
		mess(0, 0, "digiDload version %s\n",version);
	someone_there = 0;	/* Any Digi cards found? */
	cx_there = 0;		/* Any C/X cards found? */
	epcx_there = 0;		/* Any EPC/X cards found? */


	no_cxconf=getcxconf( cx_cfg, CXCONFIG, 0 );
	no_epcxconf=getcxconf( epcx_cfg, EPCXCONFIG, 1 );
	
	for (card = 0; card < board_count; card++)
	{
		di = &board[card].digi_info ;
		
		if(di->info_ioport == 0) {
			mess (1, 1, "Error Initializing card #%d(%s): Board not found\n",
				card, btypes[di->info_bdtype].desc );
			continue;
		}
		
		if( verbose )
			mess( 0, 0, "Initializing card #%d(%s):\n",
 			card, btypes[di->info_bdtype].desc );
        	if (ioctl(digiFD, DIGI_DISABLE, di) < 0)
            		mess( 0, 0, "Cannot disable card!..." );
		if( !reset_cards(card) && !load_binary(card, BIOS) ) {
			sleep(1); 
			if( !load_binary(card, FEP) ) {
				sleep( 2 );
				someone_there = 1;
				switch( di->info_bdtype ) {
					case ISACX:
					case PCICX:
					case PCICX_IBM:
						cx_there = 1;
						break;
				
					case PCIEPCJ:
						epcx_there = 1;
						break;
				}
			}
		}

		memoff( card );	/* Disable card memory so other cards can be initialized */
	}


	if( someone_there ) {

		if( cx_there && no_cxconf ) {
			mess(0, 1, "No %s, using defaults for CX concentrators\n",
			CXCONFIG );
		}
			
		if( epcx_there && no_epcxconf ) {
			mess(0, 1, "No %s, using defaults for EPCX concentrators\n",
			EPCXCONFIG );
		}
			
		if( cx_there || epcx_there ) {
			mess(0, 0, "Downloading concentrator image%s:\n",
			     (cx_there && epcx_there)? "s": "" );
			if( con_init( digiFD ) )
				mess(0, 0, "Can't load concentrator images\n");
		}
		
		if (ioctl(digiFD, DIGI_INIT, 0) < 0  ) {
			mess(0, 0, "Can't ioctl DIGI_INIT %d\n",  errno); 
			return(1);
		}

		if( verbose )
			report_numports( digiFD );

		/* unlink old and mkdev current device links */
		get_majors( digiFD );
#if 0
		umask( 0 );
#endif
		if( do_mknod ) {
			if(!forcecuds) {
				buildkernelver();
				if(kv.major == 2 && kv.minor >= 1) {
					do_cuds = 0;
				} else {
					do_cuds = 1;
				}
			} else {
				do_cuds = 1;
			}
			
			update_devs();
		}

		if (ioctl(digiFD, DIGI_POLLER, 0) < 0  ) {
			mess(0, 0, "Can't ioctl DIGI_POLLER %d\n", errno); 
			return(1);
		}

	}

	if(!globerr) {
		mess(0, 0, "digiDload complete.\n");
		exit(0);
	} else {
		mess(0, 0, "digiDload complete - %d CRITICAL ERRORS\n",globerr);
		exit(-1);
	}
	
	return( 0 );

} /* End main */
