/* -*- linux-c -*- (Autoset linux indent style when using emacs)
 *
 * Copyright (C) 1996 Digi International.
 *
 * For technical support please email digiLinux@dgii.com or
 * call Digi tech support at (952) 912-3456
 *
 * Much of this design and code came from pcxx.c which was 
 * copyright (C) 1994, 1995 Troy De Jongh, and subsquently 
 * modified by David Nugent, Christoph Lameter, Mike McLagan. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * October 1997: PEB : Modified original version 1.1.0 to fix two problems
 *      with sensing DCD modem signal sensing.  Bumped version number
 *      to 1.3.0.  See changes.doc.
 *
 * January 1998: PEB : Added support for Xr-920 family.  Fixed timer problem
 *      with KERNEL2_x.  Fixed extended baud rate communication problem
 *      with ditty.  Changed version number to 1.3.2.
 *
 * March  1998:  PEB : Fixed problems with dropped characters; break
 *     handling; special baud rate remapping when "fastbaud" is set;
 *     Xr family notification, removed refs to DIGI_TIMER2 and DIGI_BH2.
 *     fixed a DCD handling problem; fixed init_PCI to recognize multiple
 *     PCI boards of same type.  Version number is now 1.3.3(Beta).
 *
 * October 1998: PEB: Fixed a problem in pc_open() where the port was being
 *     reinitialized with every open.  The problem with this is that data
 *     could be lost if one process had a port open for reading/writing
 *     and another process came along and opened the port, for example:
 *     to check its status.  Note: a number of changes were also made
 *     to digiDload (current version of digiDload is 1.3.6) and the
 *     FEP and BIOS files were updated (to v2.3.11) as part of this
 *     release.  Changed epca.c version to 1.3.4(Beta)
 *
 * November 12, 1998: PEB: Fixed a problem in post_fep_init() that caused
 *     subsequent accesses to the board (e.g: after enabling polling)
 *     to crash the driver if no ports were found at boot time.  This
 *     situation could arise, for example: on a PCI Xem if the PORTS
 *     module is disconnected.  Now we just mark the board as DISABLED
 *     and continue.  Changed epca.c version to 1.3.4(Beta-2)
 *
 * December 21, 1998: PEB Fixed kernel module implementation.  Now the
 *     driver can be managed via modutils (e.g: modprobe, rmmod, lsmod).
 *
 * January 17 - February 4, 1999: PEB
 *      -- Converted from old to new style (introduced in later
 *         2.1 kernels) PCI recognition and configuration routines.
 *      -- Added PCI 2r-920 support.
 *      -- Put minimal card existence check code in post_fep_init().
 *      -- Add CX adapter support.
 *      -- Setting the open count to 0 on init to allow digiDload to
 *         clear unclosed ports.
 *      -- Also added a new ioctl: DIGI_DISABLE
 *      -- Adding back the old PCI code, #ifdef'd for pre-2.1 kernels.
 *      -- Patches from Dan Taylor to:
 *         1. Fix schedule_timeout for 2.2 kernels
 *         2. Add "digiepca=Override" to allow boottime "digiepca=2"
 *            overriding of epcaconfig.h values
 *      -- Changed epca.c version to 1.3.6(Beta-6)
 * February 23 - March 16, 1999: PEB
 *      -- Removed fepcmd() calls to RXPAUSE in pc_throttle() and RXRESUME
 *         in pc_unthrottle() since these should be redundant and the
 *         RXRESUME can cause the CX concentrator to reboot under some
 *         circumstances.
 *      -- Added new ioctls (DIGI_CXCON_INIT and DIGI_CXCON_NEXT)
 *         to get C/X concentrator image (/etc/cxcon.bin), if needed, from
 *         digiDload at runtime.  (Previous versions had this file compiled
 *         into the driver.)
 *      -- Added Xr-422 support
 *      -- Patches from John Lombardo (john@deltanet.com):
 *          - Allowed module load-time parameters:
 *             mod_type       Module type 2=ISA PC/Xem, etc
 *             mod_altpin     Altpin?
 *             mod_numports   How many serial ports do we support?
 *             mod_port       The base of the io ports window
 *             mod_membase    The base of the mem window
 *
 *            So, the following insmod line would install a ISA PC/Xem card:
 *                 insmod epca.o mod_type=2 mod_numports=16 mod_port=0x224
 *                 mod_membase=0xd0000
 *          - Added /proc/epca to display the driver's concept of the installed
 *            cards
 *          - Fixed types for mod_*.  Seems only ulongs are arrayable.
 *      -- Include /proc code only if CONFIG_PROC_FS is enabled (as per
 *         autoconf.h)
 *      -- Changed version to 1.4.0
 *
 * March 31, 1999: PEB
 *      -- Changed version to 1.4.1
 *      NOTE: The version number change from 1.4.0 to 1.4.1 merely
 *      indicates that this file came from the 1.4.1 distribution,
 *      with utility programs that have been updated since the 1.4.0
 *      release.  Other than the version number change (and this note
 *      describing it, of course), epca.c itself is unchanged since version
 *      1.4.0.
 *
 * April 16, 1999: PEB/CWS
 *      -- Added a call to setup_empty_event() in the pc_ioctl() calls
 *         for TCSETSW and TCSETAW since otherwise the corresponding
 *         tty_wait_until_sent() calls won't wake up until another event is
 *         triggered for that port.  This problem resulted in port hangups,
 *         notably visible under Pick D3 Linux.
 *      -- Fixed a problem with the module use counter.
 *      -- Folded in Chad Schwartz' patches:
 *         1. /proc code enhancements:
 *            A) We state the version number in /proc/epca
 *            B) We state when the driver has been loaded as a module
 *            C) We don't just return an empty /proc/epca file when we
 *               haven't initialized any cards.
 *         2. Added Engineering DIGI_MODDEC ioctl() so that if a
 *            MOD_INC_USE_COUNT is placed incorrectly, we can still recover
 *            from it, and do an rmmod on the module, and fix the problem,
 *            instead of having to reboot. :)
 *      -- Added printk() calls in pc_ioctl(), ifdef'd to IOCTL_DEBUG.
 *      -- Changed version to 1.4.2(Beta-4)      
 *
 *
 * April 28 - May 5, 1999: CWS/PEB
 *      -- Added 'total concentrator inits' to /proc/epca.  This could be
 *         useful if we are trying to debug a user issue.  It would 
 *         possibly tell us if a C/Con was rebooting itself, or that
 *         a particular C/Con on a chain is not asking to be initialized, 
 *         for some particular reason.  If the board, however, is _NOT_ a 
 *         C/X board, we make sure and put an "N/A" in the concinits field
 *         so that we KNOW we're not counting concentrator inits on boards
 *         without concentrators. :)
 *      -- Attempt to fix reported problem where setting baud rate to 0
 *         does not drop the modem lines.  It appears that the lines are
 *         now dropped but not for long enough.
 *      -- Changed version to 1.4.2(Beta-5)      
 *
 * June 1, 1999: CWS/PEB
 *      -- Added PCI/EPCX Jupiter board support
 *      -- Added ioctl's DIGI_EPCXCON_INIT and DIGI_EPCXCON_NEXT.  These 
 *         ioctl's are used to copy over an EPC/X concentrator image from the
 *         userspace program digiDload, into the kernel driver itself.  
 *      -- When a PCI/EPCX Jupiter board is installed, the driver can 
 *         now autodetect what type of concentrator it is initializing, and 
 *         send the correct concentrator image.  If the concentrator is a
 *         C/CON, the driver knows to send the 80186 based C/CON image.
 *         (/etc/digi/cxcon.bin) If the concentrator is an EPC/CON, the driver
 *         knows to send the  MIPS-based EPC/Con fep. (/etc/digi/fxcon.bin). 
 *      -- In post_fep_init(), if kmalloc fails on any port, disable the whole
 *         card.  Otherwise kernel panics will surely follow.
 *      -- In post_fep_init(), call kfree before kmalloc if tmp_buf is
 *         non-NULL.  Otherwise running digiDload multiple times will leak
 *         memory.
 *      -- Changed version to 1.4.2(Beta-7)
 * June 13, 1999: PEB/CWS
 *      -- Added special override to cxconc since CXs can only support
 *         CX/CON concentrators.
 *      -- Fixed baudrate mapping problem where requests by "stty" for
 *         230400 and 460800 were incorrectly mapped to 460800 and 921600.
 *         Requests made using "ditty fastbaud" are unaffected (they worked
 *         properly before and continue to do so now.)
 *      -- Fixed problem where setting baudrate to 0 didn't reset DTR and
 *         RTS lines.
 *      -- Replace DIGI_MODDEC ioctl -- dropped by mistake from Beta-7.
 *      -- Added conditional code to support new 2.3.X kernel wait_queue
 *         structures. 
 *      -- Changed version to 1.4.2(Beta-9)
 * June 14, 1999: PEB
 *      -- Fixed several problems with modem line handler ioctls
 *         (TIOCMSET/TIOCMBIS/TIOCMBIC)
 *      -- Changed version to 1.4.2(Beta-10)
 * June 19, 1999: PEB
 *      -- Fixed baudrate mapping problem with requests for 57600 and
 *         added 76800.
 *      -- Added "digiext" flag to epcaparam() and termios2digi_h() so
 *         the latter can determine how to interpret a cleared CRTSCTS bit.
 *         Previously, after once setting RTS/CTS handshaking, attempts
 *         to unset it (e.g: "ditty -crtscts") would seem to work (i.e:
 *         ditty would report -crtscts), but in fact RTS/CTS handshaking
 *         would still be active in the driver.
 *      -- Changed version to 1.4.2(Beta-11)
 * June 24, 1999: CWS/PEB
 *      -- Changed epcassert value from 64 to 224 in post_fep_init.
 *      -- In pc_open() exit with error if the board containing this
 *         port has not been successfully initialized by digiDload.
 *      -- Changed version to 1.4.2(Beta-12)
 * July 7, 1999: PEB
 *      -- Added "pc_infotable".  Without this, the digi_ctl device
 *         inappropriately shared memory space with /dev/ttyD0.  If
 *         both happened to be open at the same time, unexpected problem
 *         behavior was very likely.
 *      -- Changed version to 1.4.2(Beta-13)
 * July 30, 1999: PEB
 *      -- Added support for Smartport family
 *      -- Changed version to 1.4.2(Beta-14)
 * September 5, 1999: PEB/CWS
 *      -- Added dynamic major numbering.
 *      -- Added support for >256 ports.  Currently Linux minor numbers
 *         are limited to 8 bits, hence the driver could normally support
 *         only 256 devices per major number.  To support more than 256
 *         devices we need more major numbers.  In order to not break
 *         existing programs that may refer to the major numbers previously
 *         allocated to Digi, we use the standard major numbers (from
 *         <major.h>) for the first bank of 256 ports, then dynamically
 *         allocate subsequent ones.
 *      -- Changed version to 1.4.2(Beta-15)
 *
 * September 8, 1999: PEB
 *      -- Added dynamic, runtime allocation of the channel arrays.
 *         By default the driver allocates space for MAXPORTS (defined
 *         in epca.h) ports.  You can redefine MAXPORTS (and recompile)
 *         to change this default.  The compiled-in value of MAXPORTS
 *         can be overridden at driver load time by specifying a value
 *         for "totalports" on the driver invocation line (see below).
 *
 *         In another attempt to keep from breaking existing programs,
 *         the device names look like they always have (e.g: /dev/ttyD0
 *         ... /dev/ttyD255), but devices after the first 256 are given a
 *         different letter designator (e.g: /dev/ttyE0... /dev/ttyE255,
 *         /dev/ttyF0.../dev/ttyF255...) Since some letters may be
 *         unavailable on your system (e.g: /dev/ttyS0... are reserved for
 *         the built-in serial ports) the actual list of letters used is
 *         specified, in the order in which you want them appliedi, in the
 *         TTY_BANKNAMES string (defined in epca.h).  The first bank is
 *         named using the first letter in TTY_BANKNAMES (typically 'D'),
 *         the second bank using the second letter (typically 'E'), etc.
 *         
 *         Since "callout" (e.g: "/dev/cud*") support is slated to be
 *         dropped from future kernels, I have expanded only the first
 *         bank of device arrays, and their names ignore the value
 *         of TTY_BANKNAMES.  Only the first 256 ports have corresponding
 *         "cu" and their names are fixed as /dev/cud0, /dev/cud1 ...
 *         /dev/cud255.
 *      -- Add "totalports" global for the internal size and offset
 *         calculations.  "totalports" is initialized at compile time to
 *         MAXPORTS but can be reset at driver load time by specifying
 *         an alternate initial value via a commandline argument to LILO,
 *         insmod or modprobe.  E.g:
 *               modprobe epca totalports=700
 *
 *         "n_banks" is calculated in pc_init() (from "totalports")
 *         at driver load time.
 *      -- Added proper error return support to init_module() and pc_init()
 *         so, for example: if a malloc fails, everything gets cleaned
 *         up and unloaded.  Previously any initialization failures 
 *         unabashedly called panic().
 *      -- Added epca_mess() -- the individual printk calls were out
 *         of control.
 *      -- Added calls to kfree "cxconc" and "epcxconc" in cleanup_module(),
 *         otherwise there was a potential memory leak when unloading the
 *         driver module.
 *      -- Added firstchan element to board_info struct so we don't have
 *         to calculate it in firstchan(), and to digi_info struct so
 *         we can communicate it to digiDload.
 *      -- Added DIGI_GETMAJOR ioctl so digiDload can query for the list of
 *         major numbers allocated (statically or dynamically) by the driver.
 *      -- Changed version to 1.4.2(Beta-17)
 *
 * October 4-November 11, 1999: PEB
 *      -- Fixed a problem in epcaparam() with setting/resetting DTR/RTS
 *         modem signals.
 *      -- Allow fepcmd calls to be processed even on DISABLED boards.
 *         This is how, for example, we set hardware flow control from
 *         post_fep_init().
 *      -- Fixed a problem with pc_info device under 2.2/2.3 kernels.
 *      -- Unless otherwise set (e.g: via ditty), set RTS and DTR high
 *         by default (though the signals are only manifested electrically
 *         when the port is opened).
 *      -- Fixed a problem with allocating and freeing termios, termios_locked
 *         and table elements of pc_callout devices.
 *      -- Added call to setup_empty_event() at end of pc_write() to fix
 *         a problem that shows up e.g: when running "screen"
 *      -- Added code to call fepcmd when setting/resetting ALTPIN so
 *         the lights work properly on C/X concentrators.
 *      -- Changed version to 1.4.2
 * January 19, 2000: CWS
 *	-- This version change merely signifies the fact that we changed
 *	   digi specific include files to be read from the local directory,
 *	   rather than in /usr/src/linux/include/linux.  There were no 
 *	   functional changes to this release.
 *	-- Changed version to 1.4.3b1
 * January 20, 2000: CWS
 *      -- OOPS! Forgot digiPCI.h - we were still including it from 
 *         /usr/src/linux/include/linux - This caused us not to compile
 *	   at all, seeing as though the file wasn't there.
 *      -- Changed version to 1.4.3b2
 * January 24, 2000: CWS
 *  	-- Changed descriptions to long "Product Catalog" format, rather than
 *         "Shortform definition" format.  I.E. Instead of 'PC/Xe', we now
 *         state "Acceleport Xe ISA".  This should clear up any differences
 *         between the release notes, and this driver.
 *	-- Due to the lengthened descriptions, we needed to make room for said
 *   	   board descriptions in the /proc/epca file. Changed max description
 *         length from 10 characters, to 25 characters.
 *      -- Changed version number to 1.4.3
 * February 22, 2000: CWS
 *	-- Version change only to fix problem in another part of the digi
 *	   EPCA rpm package.
 *	-- Changed version number to 1.4.3.2
 * March 8, 2000: CWS
 *      -- We weren't properly toggling DTR off and back on on a baudrate
 *         0 request. thus, we weren't hanging up modems properly in several
 *         apps that still used the baudrate 0 interface.
 *      -- Changed version to 1.4.3.3
 * January 28, 2000: CWS
 *	-- Added DIGI_PRINTER digi extended flag handling support.
 *	   Traditionally, this flag has been used to keep termios settings
 *	   "sticky".  However, "sticky" termios settings are a default
 *	   setting in the linux device driver. (We never actually free
 *	   termios structures, except when unloading the driver.)
 *	   Thus, the only thing we needed to add, was that when the flag
 *	   was set, the module usage count for this driver goes up by 1,
 *	   as well as keeping modem output signals HIGH at all times (even
 *	   when the last process closes the port.) Of course, they do have
 *	   the ability to go low, if we're doing DTR/RTS based input flow
 *	   control.
 *	-- Added native TCFLSH (TCIFLUSH/TCIOFLUSH/TCOFLUSH) support to the
 *	   driver.  Now, when a user executes a 'ditty flush' command on us,
 *	   we don't just flush the TTY buffers out (as the n_tty_ioctl() code
 *	   does) but now, in addition,  we actually flush out the CARD's
 *	   appropriate channel buffers, as well. (Thus, for example, when
 *	   TCIFLUSH ioctl is called, we completely flush all tty line
 *	   discipline buffers, as WELL as the receive buffer on the card
 *	   itself.
 *	-- Changed version number to 1.4.4-Beta1
 * January 28, 2000: CWS
 *	-- Added TCXONC ioctl call (ditty startin/startout/stopin/stopout)
 *	-- All the TCXONC options work now..  startout/stopout are handled
 *	   by us... startin/stopin handled correctly via n_tty_ioctl()
 *	   already.
 *	-- Fixed a problem where if the user was in a SOFTWARE OUTPUT flow
 *	   control state at portclose, when the user REOPENED the port, the
 *	   port would STILL be unable to transmit a character until a
 *	   RESUMETX fepcmd was sent.
 *	-- Changed version number to 1.4.4-Beta2
 * February 25, 2000: CWS
 *	-- Added DIGI_GETDD ioctl. This IOCTL passes the number of boards
 *	   installed back to userland, along with the version string of
 *	   the driver.
 *	-- Added DIGI_GETBD ioctl. Similar to DIGI_GETINFO, with several
 *	   changes.  This has been put in place for compatibility with
 *	   'DPA', a diagnostic utility written by Digi International.
 *	-- Added DIGI_KME ioctl.  This ioctl's function is to grab a 
 *	   specified "chunk" of a Digiboard's memory, and pass it back
 *	   to userspace.  It can also WRITE to Digi's memory. It can also
 *	   READ/WRITE from REMOTE CONCENTRATOR memory.  This function adheres
 *	   to Digi International's DIGI_KME ioctl from other drivers.
 *	-- Added the kme_wait waitqueue, to ensure that when we are in
 *	   the middle of a remote concentrator WRITE call, that we have 
 *	   sole access to the queue.
 *	-- Changed the way we handle the digi_info structure. Several data
 *	   types were added and/or changed, to handle the "DPA Way" of 
 *	   passing data back and forth to and from the driver.
 *	-- Added 'dpatype' flag to the digi_info structure. this allows us
 *	   to pass the DPA type flags to dpa directly in its own format, 
 *	   rather than making DPA work "our way".  
 *	-- Added 'dpastatus' flag to the digi_info structure. this allows
 *	   us to pass the DPA type status flags to dpa directly in its own
 *	   format, rather than making DPA work "our way".
 *	-- The change to the digi_info structure also requires us to use
 *	   a different format in the epcaconfig.h headerfile.  We now are
 *	   required to pass in the 'dpatype' of board we're configuring.
 *	   the 'digiConf' userspace program knows about these types.
 *	-- Changed version to 1.4.4Beta3
 * March 6, 2000: CWS
 *	-- Remove support for several obsolete digiboards.  These boards
 *	   are as follows: Acceleport Xe/64k ISA, Acceleport C/X EISA, 
 *	   Acceleport Xem EISA, and Acceleport XeVe ISA. The last version of 
 *	   the driver to support these boards officially is version 1.4.3.2.
 *	-- Removed ISA/EISA board limit of 7 - Made the limit MAXBOARDS,
 *	   currently set to 12.
 *	-- Changed some of John Lombardo's code, to allow it to work 
 *	   properly in > 2.0 kernels, as well as now being able to support
 *	   the new PCI card types.  (We now look at mod_numports[0] to know
 *	   if any options were specified on the cmdline)
 *	-- Added a compatibility version of schedule_timeout() which is
 *	   conditionally compiled in for 2.0.x kernels.
 *	-- Added type checking code.  Now, if we're told that we have a
 *	   digiboard, but it doesn't match a type we know, we IMMEDIATELY
 *	   disable the board, AND zero out its memory address and I/O
 *	   address from the driver so digiDload skips it. We also make
 *	   it known to the user by exporting an error message in place of
 *	   the card description, to /proc/epca.
 *	-- PCI boards are now specified in epcaconfig.h (or insmod cmdline)
 *	   rather than getting autoprobed directly.  Many changes to
 *	   the driver were done to comply with this enhancement. 
 *	-- Added pci_boards structure pointer. Now, when the driver 
 *	   does its PCI probe it can allocate memory dynamically to
 *	   this pointer, rather than having the whole structure in RAM.
 *	   Once its done with the memory it can free it all, thus only
 *	   using the memory space of 1 pointer.
 *	-- Now, as a result of the PCI enhancement, a user can specify
 *	   ALTPIN by default on a PCI digiboard, when previously they
 *	   couldn't.
 *	-- If a PCI board is specified but not detected in the PCI 
 *	   probe at loadtime, the driver keeps the board in its detected list, 
 *	   however marks it disabled, with I/O and memory zeroed out so 
 *	   digiDload knows not to initialize it. (It also keeps port 
 *	   locations fixed, rather than renumbering.)
 *	-- Fixed problem where if user were to load our driver on
 *	   a system that didn't have PCI support, we OOPS'ed at loadtime.
 *	   this was caused by an improperly placed declaration.
 *	-- Added epca_removepci(), which conditionally gets compiled in
 *	   when the user doesn't have PCI support enabled in their kernel.
 *	   Its job is to STRIP PCI boards that have been specified, from
 *	   the boards structure. (And leave only ISA boards.)
 *	-- Got rid of the totalports modular cmdline option, seeing as 
 *	   though now we already know EXACTLY how many ports we're going
 *	   to support in total.
 *	-- Added the epca_paranoia() function, which we can now use to 
 *	   place repetitive sanity checking code, rather than implementing
 *	   it elsewhere.
 *	-- in post_fep_init(), we were overwriting part of FEP/OS memory, 
 *	   when we were attempting to check if card memory was present.  
 *	   This has been fixed.  We now have a command within 
 *	   epca_paranoia(), that can check that board memory is present, 
 *	   and that FEP/OS is running properly.
 *	-- Added error checking that detects when a board is configured for a 
 *	   certain number of ports, but FEP/OS reports a different number
 *	   of ports detected.  We now epca_mess() the user to tell them
 *	   what happened. (And still only support the configured number of
 *	   ports)
 *	-- Added DIGI_SPOLL ioctl, which allows the ability to dynamically 
 *	   change the driver's polling rate.  By default, the driver polls 
 *	   the event queue every timer tick. However, this value can be 
 *	   slowed down all the way up to 1 second poll times.
 *	-- In previous, we were using an array of strings called 
 *	   board_desc, to supply us with textual descriptions of each board
 *	   type. Now, we use a predefined array of structs, called 'btypes',
 *	   that holds both the textual description of the board, as well as
 *	   the DPA specific type for the board.
 *	-- Re-added EPC/X Jupiter (PLX) PCI support to the driver. EPC/X PCI
 *	   is now officially supported by Digi International.
 *	-- If an app executed TIOCMBIC, but didn't reverse its behavior,
 *	   the driver wouldn't ever allow the signals that had been shut
 *	   off, to come high again - unless someone executed a TIOCMSET on
 *	   the port.  Fixed.
 *	-- Balanced out globalwinon()'s to memoff()'s so that we don't
 *	   inadvertently execute an operation that requires the board's
 *	   memory window to be turned on, on a board that has its memory
 *	   turned off.
 *	-- Changed version number to 1.4.4
 * June 15, 2000: CWS
 *	-- We were attempting to translate PCI remapped address into 
 *	   virtual addresses.  (We shouldn't be doing that. remapped
 *	   addresses already ARE virtual addresses.)  Now, we only attempt
 *	   to translate addresses for boards that are NOT PCI. (I.E. ISA
 *	   boards.)
 *	-- We were incorrectly doubling up on some physical --> virtual
 *	   memory-mapping translations, which caused OOPSes in several
 *	   situations, that would only occur in certain circumstances.
 *	   This has been fixed.
 *	-- Fixed a few CONFIG_MODVERSIONS anomalies in the driver.  We
 *	   were using it wrong in a few instances, and it was causing 
 *	   problems in newer redhat versions.
 *	-- Changed version number to 1.4.5-1
 * July 24, 2000: CWS
 *	-- PCI code for 2.0.x was completely broken, and wouldn't even
 *	   compile.  This has been fixed.
 *	-- For 2.2.x kernels, We now compile using BIOS32 PCI instructions
 *	   rather than using the higher level PCI layer.  This maintains 
 *	   compatibility between 2.0.x PCI code and 2.2.x PCI code.
 *	-- Changed version to 1.4.5eb1-1
 * July 31, 2000: CWS
 *	-- This version changes nothing in epca.c, other than the version
 *	   string.  The release indicates a change in the package the driver
 *	   is distributed in.
 *	-- Changed version to 1.4.5eb2-1
 * August 4, 2000: CWS
 *	-- This version changes nothing in epca.c, other than the version
 *	   string.  The release indicates a change in the package the driver
 *	   is distributed in.
 *	-- Changed version to 1.4.5eb3-1
 * September 13, 2000: CWS
 *	-- Fixed a bug that would only affect the user if they had no boards
 *	   installed via epcaconfig.h, or via module commandline config
 *	   methods.  The driver would error out on not being able to 
 *	   malloc termios structures. However, the error occured because
 *	   the total number of boards count was 0.  This has been fixed.
 *	   The driver will not allow you to load it without any boards
 *	   configured.
 *	-- Fixed a few PCI issues - which prevented us from compiling
 *	   under 2.0.x kernels.
 *	-- Integrated this driver in a Source RPM package, to comply
 *	   with Redhat's certification requirements.
 *	-- Thanks goes to Bill Parker (dogbert@netnevada.net) for pointing out
 *	   a change in kernel version v2.2.18-pre12 and up that rendered the
 *	   driver uncompilable, under that kernel version. Previous versions
 *	   of the kernel use the variable 'loops_per_sec' to properly
 *	   calculate timed delay loops, among other things.  Now, the kernel
 *	   has been converted to use 'loops_per_jiffy', which is a smaller
 *	   unit of calculation.  Included support for both cases.  
 *	-- Fixed CRTSCTS brokenness.  Now, CRTSCTS is entirely independant
 *	   from CTSPACE and RTSPACE.  Those values are now "Override" 
 *	   values, which - if CRTSCTS is NOT set, become active.
 *	-- Fixed DTR/RTS anomalies that would prevent getty_ps from
 *	   initializing a port correctly, under some very rare circumstances.
 *	-- Changed version to 1.4.6
 * March 11, 2001: Jeff Randall (Jeff_Randall@digi.com)
 *	-- General cleanup of code formatting and comments to Linux
 *         coding style
 *      -- Support for 2.4 and devfs added (from patches supplied by
 *         Jim Paris <jim@jtan.com>).
 *      -- removed KERNEL2_x define and replaced with direct KERNEL_VERSION
 *         defines.
 *      -- Changed the means of reporting the version.
 * June 22, 2001: James Puzzo (jamesp@digi.com)
 *      -- Added minimal support for the IBM 8-Port Async PCI Adapter
 *         which is similar to the Digi AccelePort Xr
 * August 21, 2001: David Fries <dfries@digi.com>
 *      -- Added a wake up for the pool_wait in 2.2.15+ kernels.
 * September 2002: Peter E. Berger (pberger@brimson.com),
 *                 Al Borchers (alborchers@steinerpoint.com)
 *      -- Added transparent print support.
 *         (Note: No attempt yet to make transparent print work with DEVFS.)
 *      -- Removed dead-code: do_softint. do_pc_bh, pc_sched_event,
 *         and some globals and channel structure fields used only
 *         by these functions.
 *      -- Close of transparent printer device sends a final print
 *         off string, if needed.
 *      -- Added support for maxcps and bufsize ditty parameters to
 *         control printer output speed.
 *	-- Added transparent printer related default (via digi_init)
 */


#ifdef MODULE
#include <linux/config.h>
#endif

#include <linux/version.h>
#include <linux/autoconf.h>

#if !defined(KERNEL_VERSION) /* defined in version.h in later kernels */
# define KERNEL_VERSION(a,b,c)  (((a) << 16) + ((b) << 8) + (c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(1,3,0)
# define NEW_MODULES
#endif

#ifdef NEW_MODULES
# ifdef CONFIG_MODVERSIONS
#  include <linux/modversions.h>
# endif
#endif

#ifdef MODULE
# include <linux/module.h>
#endif


#include <linux/errno.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/tty_driver.h>
#include <linux/kernel.h>
#include <linux/signal.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    #include <linux/slab.h>
#else
    #include <linux/malloc.h>
#endif

#include <linux/sched.h>

#include <linux/mm.h>
#include <linux/ctype.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>


#include <asm/bitops.h>

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty_flip.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/major.h>
#include <linux/ioport.h>

#if defined(MODULE) && !defined(NEW_MODULES)
char kernel_version[]=UTS_RELEASE;
#endif

/* Is it an older 1.2.x kernel ? */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
/* 
 * Technically 1.2.x kernels have PCI support but it does not have the ability
 * to map PCI high addresses such as what might be returned from a base address
 * register.  This pretty much eliminates PCI support for this kernel.
 */
#ifdef CONFIG_PCI
#define ENABLE_PCI
#endif /* CONFIG_PCI */

#else
/* Define the below as nothing so the compiler won't complain */
#define bus_to_virt 
/* B460800 is defined in include/asm/termbits.h in 2.X kernels */
#define  B460800 0010004
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0) */


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
/* ------------------------------------------------------------------------
The below represent changes in the naming convention since 2.0.0.  I alias
them here so we won't have two separate code bases.
---------------------------------------------------------------------------- */
#ifdef ENABLE_PCI
# define ioremap vremap
# define iounmap vfree
#endif /* ENABLE_PCI */

#define copy_from_user memcpy_fromfs
#define copy_to_user memcpy_tofs
#define putUser(arg1, arg2) put_fs_long(arg1, (unsigned long *)arg2)
#define getUser(arg1, arg2) arg1 = get_fs_long((unsigned long *)arg2)
#else
#include <asm/uaccess.h>
#define putUser(arg1, arg2) put_user(arg1, (unsigned long *)arg2)
#define getUser(arg1, arg2) get_user(arg1, (unsigned int *)arg2)
#endif /* Version is 2.0.0 */

/* ------------------------------------------------------------
 * The 1.2.x version does not provide a mapping facility for
 * those addresses that exist above RAM.  For this reason PCI
 * support is not supported under 1.2.x
 ---------------------------------------------------------------*/


#ifdef ENABLE_PCI
# if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#  include <linux/bios32.h>
# endif
# include <linux/pci.h>
# include "digiPCI.h"
#endif /* ENABLE_PCI */

#include "rascon.h"
#include "digiFep1.h"
#include "epca.h"
#include "epcaconfig.h"


/* ---------------------- Begin defines ------------------------ */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,0,0)
#define DIGI_MAJOR   22
#define DIGICU_MAJOR 23
#endif

#define DIGIINFOMAJOR       35  /* For Digi specific ioctl */ 

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define MAXCARDS 7
#define epcaassert(x, msg)  if (!(x)) epca_error(__LINE__, msg)

/* ----------------- Begin global definitions ------------------- */
char *epca_version =
"$Id: epca.c,v 1.54 2004/04/29 21:42:29 scottk Exp $";

#if defined(MODULE_LICENSE)
    MODULE_LICENSE("GPL");
#endif

static char mesg[100];
static int pc_refcount, nbdevs = 0, num_cards = 0, liloconfig = 0;
static int digi_poller_inhibited = 1 ;

static int setup_error_code = 0;
static int invalid_lilo_config = 0;

/*
 *  This array will hold any concentrator images which are
 *  passed to the driver via the "DIGI_CON_*" ioctls.  At present,
 *  a total of 16 concentrator images can be handled.
 */
#define MAX_CONC_IMAGES 16
static char *conc_images[MAX_CONC_IMAGES] = {0};
static int   num_conc_images = 0;

/* static int conctype = 0; */

/* This holds the number of times each C/X has been initialized */
static int concinits[MAXBOARDS];

static int pollrate = 0;
				
/* This holds the status of the KME buffer */
static int kmebusy = 0;

static struct board_info boards[MAXBOARDS];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
static int pc_timeron = 0;
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
# define ALLOW_ALT_FAIL_OPEN 1
# define ALT_FAIL_OPEN_DEFAULT 0

static struct timer_list pc_timer = { {NULL, NULL}, 0, 0, 0};
# else
#define ALLOW_ALT_FAIL_OPEN 0
static struct timer_list pc_timer = {NULL, NULL, 0, 0, 0};
# endif
#endif


/* Local storage for epca_mess printk messages */
#define MESBUFSIZE 1024
static char epca_mesbuf[MESBUFSIZE];
static char *epca_mesbufpt=epca_mesbuf;

static struct board_info **pci_boards;	/* Utilized in PCI probing. */
					/* Freed afterwords.        */

#ifdef MODULE
/* ------------- Module Specific Globals ------------------ */
static ulong    mod_type                [MAXCARDS] = {0,};
static ulong    mod_altpin              [MAXCARDS] = {0,};
static ulong    mod_numports    [MAXCARDS] = {0,};
static ulong    mod_port                [MAXCARDS] = {0,};
static ulong    mod_membase             [MAXCARDS] = {0,};
#if ALLOW_ALT_FAIL_OPEN
static ulong    mod_alt_fail_open = ALT_FAIL_OPEN_DEFAULT;
#endif
/* --------- End of  Module Specific Globals -------------- */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(mod_type, 		"1-" __MODULE_STRING(MAXCARDS) "i");
MODULE_PARM(mod_altpin, 	"1-" __MODULE_STRING(MAXCARDS) "i");
MODULE_PARM(mod_numports, 	"1-" __MODULE_STRING(MAXCARDS) "i");
MODULE_PARM(mod_port, 		"1-" __MODULE_STRING(MAXCARDS) "i");
MODULE_PARM(mod_membase, 	"1-" __MODULE_STRING(MAXCARDS) "i");
#if ALLOW_ALT_FAIL_OPEN
MODULE_PARM(mod_alt_fail_open, "i");
#endif
#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,15)\
	&& LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define SERIAL_HAVE_POLL_WAIT 1
/* This allows the driver to wake up select or poll processes when write
 * space becomes available.  Only later 2.2.x kernels require this 8-21-2001
 */
#endif

extern struct module __this_module;

static int n_banks;
/* Keep track of the major numbers allocated for each bank of minor numbers */
static int majors[sizeof(TTY_BANKNAMES)-1];
static int majorspr[sizeof(TTY_BANKNAMES)-1];

/*
 * Keep track of the letter designator (e.g: 'D' in "ttyD")
 * for each bank of minor numbers
 */
static char *epca_banknames=TTY_BANKNAMES;


struct tty_driver **pc_driver;
struct tty_driver **pr_driver;
struct tty_driver pc_callout;
struct tty_driver pc_info;


/* There's only one single "info" device */
static struct tty_struct *pc_infotable[1];
static struct termios *pc_infotermios[1];
static struct termios *pc_infotermios_locked[1];

/* ------------------------------------------------------------------------
 * digi_channels: array of pointers to channel structures to keep
 * track of local information (including stuff not known by the
 * tty subsystem) such as transmit and receive pointers, termio data,
 * and signal definitions (DTR, CTS, etc ...)   Note: These structures
 * are for driver housekeeping and are NOT used to overlay the cards' 
 * physical channel structure.
 -------------------------------------------------------------------------- */
  
static struct channel **digi_channels;


/* ------------------------------------------------------------------------
 * card_ptr is an array used to hold the address of the
 * first channel structure of each card.  This array will hold
 * the addresses of various channels located in digi_channels.
 -------------------------------------------------------------------------- */
static struct channel *card_ptr[MAXCARDS];

/* ---------------------- Begin function prototypes --------------------- */

/* ----------------------------------------------------------------------
 * Begin generic memory functions.  These functions will be alias
 * (point at) more specific functions dependent on the board being
 * configured.
 ----------------------------------------------------------------------- */
	

#ifdef MODULE
int                init_module(void);
void               cleanup_module(void);
#endif /* MODULE */

static inline void memwinon(struct board_info *b, unsigned int win);
static inline void memwinoff(struct board_info *b, unsigned int win);
static inline void globalwinon(struct channel *ch);
static inline void rxwinon(struct channel *ch);
static inline void txwinon(struct channel *ch);
static inline void memoff(struct channel *ch);
static inline void assertgwinon(struct channel *ch);
static inline void assertmemoff(struct channel *ch);

/* ---- Begin more 'specific' memory functions for cx_like products --- */

static inline void pcxem_memwinon(struct board_info *b, unsigned int win);
static inline void pcxem_memwinoff(struct board_info *b, unsigned int win);
static inline void pcxem_globalwinon(struct channel *ch);
static inline void pcxem_rxwinon(struct channel *ch);
static inline void pcxem_txwinon(struct channel *ch);
static inline void pcxem_memoff(struct channel *ch);

/* ------ Begin more 'specific' memory functions for the pcxe ------- */

static inline void pcxe_memwinon(struct board_info *b, unsigned int win);
static inline void pcxe_memwinoff(struct board_info *b, unsigned int win);
static inline void pcxe_globalwinon(struct channel *ch);
static inline void pcxe_rxwinon(struct channel *ch);
static inline void pcxe_txwinon(struct channel *ch);
static inline void pcxe_memoff(struct channel *ch);

/* ---- Begin more 'specific' memory functions for the pc64xe and pcxi ---- */
/* Note : pc64xe and pcxi share the same windowing routines */

static inline void pcxi_memwinon(struct board_info *b, unsigned int win);
static inline void pcxi_memwinoff(struct board_info *b, unsigned int win);
static inline void pcxi_globalwinon(struct channel *ch);
static inline void pcxi_rxwinon(struct channel *ch);
static inline void pcxi_txwinon(struct channel *ch);
static inline void pcxi_memoff(struct channel *ch);

/* - Begin 'specific' do nothing memory functions needed for some cards - */

static inline void dummy_memwinon(struct board_info *b, unsigned int win);
static inline void dummy_memwinoff(struct board_info *b, unsigned int win);
static inline void dummy_globalwinon(struct channel *ch);
static inline void dummy_rxwinon(struct channel *ch);
static inline void dummy_txwinon(struct channel *ch);
static inline void dummy_memoff(struct channel *ch);
static inline void dummy_assertgwinon(struct channel *ch);
static inline void dummy_assertmemoff(struct channel *ch);

/* ------------------- Begin declare functions ----------------------- */

static inline struct channel *verifyChannel(register struct tty_struct *);
static void epca_error(int, char *);
static void epca_mess( char *loglevel, char *fmt, ... );
static int epca_paranoia(int crd, int function);

static void pc_close(struct tty_struct *, struct file *);
static void shutdown(struct channel *, struct tty_struct *);
static void pc_hangup(struct tty_struct *);
static void pc_put_char(struct tty_struct *, unsigned char);
static int pc_write_room(struct tty_struct *);
static int pc_chars_in_buffer(struct tty_struct *);
static void pc_flush_buffer(struct tty_struct *);
static void pc_flush_chars(struct tty_struct *);
static void write_final_offstr( struct tty_struct * tty, long timeout );
static int block_til_ready(struct tty_struct *, struct file *,
                           struct channel *);
int pc_open(struct tty_struct *, struct file *);
static void post_fep_init(unsigned int crd);
static void epcapoll(unsigned long);
static void doevent(int);
static void fepcmd(struct channel *, int, int, int, int, int);
static unsigned termios2digi_h(struct channel *ch, unsigned);
static unsigned termios2digi_i(struct channel *ch, unsigned);
static unsigned termios2digi_c(struct channel *ch, unsigned);
static unsigned termios2digi_misc(struct channel *ch);
static void epcaparam(struct tty_struct *, struct channel *);
static void receive_data(struct channel *);
static void flush_rx_buffer(struct channel *);
static void flush_tx_buffer(struct channel *);
static int pc_ioctl(struct tty_struct *, struct file *,
                    unsigned int, unsigned long);
static void pc_set_termios(struct tty_struct *, struct termios *);
static void pc_stop(struct tty_struct *);
static void pc_start(struct tty_struct *);
static void pc_throttle(struct tty_struct * tty);
static void pc_unthrottle(struct tty_struct *tty);
static void digi_send_break(struct channel *ch, int msec);
static void setup_empty_event(struct tty_struct *tty, struct channel *ch);
static int maxcps_room(struct tty_struct *tty, struct channel *ch,
	int bytes_available);
void epca_setup(char *, int *);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
# if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
static int get_termio(struct tty_struct *, struct termio *);
# endif
static int pc_write(struct tty_struct *, int, const unsigned char *, int);
int pc_init(void);
#else
static int pc_write(struct tty_struct *, int, unsigned char *, int);
long pc_init(long);
#endif

#ifdef ENABLE_PCI
static int init_PCI(int);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,127)
/* For older kernels that don't know about this function, lets redefine it */
long schedule_timeout(long timeout) {
	unsigned long expire;
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = expire = jiffies + timeout;
	schedule();
	current->timeout = 0;
	timeout = expire - jiffies;
	if( timeout < 0 )
		timeout = 0;
	return( timeout );
}
#endif

#ifdef CONFIG_PROC_FS
/* ------------------------------------------------------------------
 *	/proc/epca
 * ------------------------------------------------------------------ */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
#include <linux/proc_fs.h>

static struct digi_struct digi_init =
{
	0,		/* Flags                        */
	100,		/* Max CPS                      */
	50,		/* Max chars in print queue     */
	100,		/* Printer buffer size          */
	4,		/* size of printer on string    */
	4,		/* size of printer off string   */
	"\033[5i",	/* ANSI printer on string ]     */
	"\033[4i",	/* ANSI printer off string ]    */
	"ansi"		/* default terminal type        */
};


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
int epca_read_proc( char *buf, char **start, off_t offset, int len )
#else
int
epca_read_proc( char *buf, char **start, off_t offset, int len, int unused )
#endif
{
	int crd;
	
	len=sprintf( buf, "\nEPCA Linux Driver %s, Digi part Number %s\n",
                     DG_NAME, DG_PART);
#ifdef MODULE
	len+=sprintf( buf+len, "NOTE: This driver is being loaded as "
		      "a kernel module!\n\n");
#endif

	if (num_cards){
		len+=sprintf( buf+len,
			"bd ty                           alt   num base       memory	  conc\n"
			"nm pe desc                      pin ports address    address	  inits\n"
			"-- -- ------------------------- --- ----- ---------- ---------- --------\n"
			);
		for (crd = 0; crd < num_cards; crd++){
			len+=sprintf( buf+len, "%2d %2d %-25s %3s %5d "
				      "0x%-8x 0x%-8x ",
				      crd,
				      boards[crd].type,
				      btypes[boards[crd].type].desc[0] == '\0' ? "FAULTY TYPE!" : btypes[boards[crd].type].desc,
				      boards[crd].altpin?" ON":"OFF",
				      boards[crd].numports,
				      (uint) boards[crd].port,
				      (uint) boards[crd].membase );

			if( boards[crd].type == ISACX ||
			    boards[crd].type == PCICX ||
			    boards[crd].type == PCICX_IBM ||
			    boards[crd].type == PCIEPCJ) {

				len+=sprintf( buf+len, "%2d\n", concinits[crd] );
			} else
				len+=sprintf( buf+len, "%-10s\n", " N/A" );
		}
		
	} else {
		len+=sprintf( buf+len, "There are no boards currently "
			      "installed!\n");
	}
	return len;
}

# if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
struct proc_dir_entry epca_proc_entry = {
	0,			/* low_ino: The inode -- dynamic */
	4, "epca",		/* len of name and name */
	S_IFREG | S_IRUGO,	/* mode */
	1, 0, 0, 		/* nlinks, owner, group */
	0,			/* size - unused */
	NULL,			/* operations -- use default */
	&epca_read_proc,	/* function used to read data */
};
# endif

#endif
#endif /* CONFIG_PROC_FS */

/*
 * Table of functions for each board to handle memory.  Mantaining 
 * parallelism is a *very* good idea here.  The idea is for the 
 * runtime code to blindly call these functions, not knowing/caring    
 * about the underlying hardware.  This stuff should contain no
 * conditionals; if more functionality is needed a different entry
 * should be established.  These calls are the interface calls and 
 * are the only functions that should be accessed.  Anyone caught
 * making direct calls deserves what they get.
 */

static inline void memwinon(struct board_info *b, unsigned int win) {
	(b->memwinon)(b, win);
}

static inline void memwinoff(struct board_info *b, unsigned int win) {
	(b->memwinoff)(b, win);
}

static inline void globalwinon(struct channel *ch) {
	(ch->board->globalwinon)(ch);
}

static inline void rxwinon(struct channel *ch) {
	(ch->board->rxwinon)(ch);
}

static inline void txwinon(struct channel *ch) {
	(ch->board->txwinon)(ch);
}

static inline void memoff(struct channel *ch) {
	(ch->board->memoff)(ch);
}
static inline void assertgwinon(struct channel *ch) {
	(ch->board->assertgwinon)(ch);
}

static inline void assertmemoff(struct channel *ch) {
	(ch->board->assertmemoff)(ch);
}


/*
 * PCXEM windowing is the same as that used in the PCXR 
 * and CX series cards.
 */

static inline void pcxem_memwinon(struct board_info *b, unsigned int win) {
        outb_p(FEPWIN|win, (int)b->port + 1);
}

static inline void pcxem_memwinoff(struct board_info *b, unsigned int win) {
	outb_p(0, (int)b->port + 1);
}

static inline void pcxem_globalwinon(struct channel *ch) {
	outb_p( FEPWIN, (int)ch->board->port + 1);
}

static inline void pcxem_rxwinon(struct channel *ch) {
	outb_p(ch->rxwin, (int)ch->board->port + 1);
}

static inline void pcxem_txwinon(struct channel *ch) {
	outb_p(ch->txwin, (int)ch->board->port + 1);
}

static inline void pcxem_memoff(struct channel *ch) {
	outb_p(0, (int)ch->board->port + 1);
}

/* ----------------- Begin pcxe memory window stuff ------------------ */
static inline void pcxe_memwinon(struct board_info *b, unsigned int win) {
	outb_p(FEPWIN | win, (int)b->port + 1);
}

static inline void pcxe_memwinoff(struct board_info *b, unsigned int win) {
	outb_p(inb((int)b->port) & ~FEPMEM, (int)b->port + 1);
	outb_p(0, (int)b->port + 1);
}

static inline void pcxe_globalwinon(struct channel *ch) {
	outb_p( FEPWIN, (int)ch->board->port + 1);
}

static inline void pcxe_rxwinon(struct channel *ch) {
	outb_p(ch->rxwin, (int)ch->board->port + 1);
}

static inline void pcxe_txwinon(struct channel *ch) {
	outb_p(ch->txwin, (int)ch->board->port + 1);
}

static inline void pcxe_memoff(struct channel *ch) {
	outb_p(0, (int)ch->board->port);
	outb_p(0, (int)ch->board->port + 1);
}

/* ------------- Begin pc64xe and pcxi memory window stuff -------------- */
static inline void pcxi_memwinon(struct board_info *b, unsigned int win) {
	outb_p(inb((int)b->port) | FEPMEM, (int)b->port);
}

static inline void pcxi_memwinoff(struct board_info *b, unsigned int win) {
	outb_p(inb((int)b->port) & ~FEPMEM, (int)b->port);
}

static inline void pcxi_globalwinon(struct channel *ch) {
	outb_p(FEPMEM, (int)ch->board->port);
}

static inline void pcxi_rxwinon(struct channel *ch) {
	outb_p(FEPMEM, (int)ch->board->port);
}

static inline void pcxi_txwinon(struct channel *ch) {
	outb_p(FEPMEM, (int)ch->board->port);
}

static inline void pcxi_memoff(struct channel *ch) {
	outb_p(0, (int)ch->board->port);
}

static inline void pcxi_assertgwinon(struct channel *ch) {
	epcaassert(inb((int)ch->board->port) & FEPMEM, "Global memory off");
}

static inline void pcxi_assertmemoff(struct channel *ch) {
	epcaassert(!(inb((int)ch->board->port) & FEPMEM), "Memory on");
}


/*
 * Not all of the cards need specific memory windowing routines.  Some
 * cards (Such as PCI) needs no windowing routines at all.  We provide
 * these do nothing routines so that the same code base can be used.
 * The driver will ALWAYS call a windowing routine if it thinks it needs
 * to; regardless of the card.  However, dependent on the card the routine
 * may or may not do anything.
 */
static inline void dummy_memwinon(struct board_info *b, unsigned int win) { }
static inline void dummy_memwinoff(struct board_info *b, unsigned int win) { }
static inline void dummy_globalwinon(struct channel *ch) { }
static inline void dummy_rxwinon(struct channel *ch) { }
static inline void dummy_txwinon(struct channel *ch) { }
static inline void dummy_memoff(struct channel *ch) { }
static inline void dummy_assertgwinon(struct channel *ch) { }
static inline void dummy_assertmemoff(struct channel *ch) { }

/*
 * Is this the major device number of a transparent print device?
 */
int istpr( kdev_t dev )
{
	int major,bank;

	major = MAJOR(dev);

	for( bank=0; bank<n_banks; ++bank ) {
			if( majorspr[bank] == major ) {
					return( 1 );
			}
	}
	return( 0 );
}


/*
 * verifyChannel()
 *
 * This routine basically provides a sanity check.  It insures that
 * the channel returned is within the proper range of addresses as
 * well as properly initialized.  If some bogus info gets passed in
 * through tty->driver_data this should catch it.
 */
static inline struct channel *verifyChannel(register struct tty_struct *tty)
{

	int i;

	if (tty) {
		register struct channel *ch =
			(struct channel *)tty->driver_data;

			for( i=0; i < nbdevs; ++i ) {
				if( (ch == digi_channels[i])
				    && (ch->magic == EPCA_MAGIC))
					return ch;
			}
	}

	/* Else return a NULL for invalid */
	return(NULL);
}


/*
 * epca_memaddr()
 */
static int epca_memaddr(int crd)
{
	struct board_info *bd;
	unchar *memaddr;
	
	bd = &boards[crd];
	
	if (bd->membase < 0x100000)
		memaddr = (unchar *)bd->membase;
	else
		memaddr = bd->re_map_membase;

	if ( (btypes[bd->type].dpatype & BUSTYPE) == T_PCIBUS)
		return((int)memaddr);
	else
		return((int)bus_to_virt((unsigned long)memaddr));
}

	

/*
 * epca_mess()
 *
 * Log message via printk directly
 */
static void epca_mess( char *loglevel, char *fmt, ... )
{
	va_list args;

#if 0
	printk("In epca_mess, loglevel:'%s'\n", loglevel);
#endif

	/* Overwrite the loglevel (first initialized in pc_init()) */
	strncpy( epca_mesbuf, loglevel, 3 );

	va_start( args, fmt );
	vsprintf( epca_mesbufpt, fmt, args );
	va_end( args );

	printk( epca_mesbuf );
}


/*
 * epca_error()
 */
static void epca_error(int line, char *msg)
{

	epca_mess(KERN_ERR, "Error: line = %d %s\n",line,msg);
}


/*
 * firstchan()
 *
 * Return line offset (index into digi_channels[]) of the first channel
 * on this card.
 */
static int firstchan( int crd )
{
	int icrd, first;

	/* Get the line offset for the first port on this card */
	first = icrd = 0;
	while( icrd < crd ) {
		first += boards[icrd].numports;
		++icrd;
	}
	return( first );
}

/*
 * pc_close()
 */
static void pc_close(struct tty_struct * tty, struct file * filp)
{

	struct channel *ch;
	unsigned long flags;

    if (tty->driver.subtype == SERIAL_TYPE_INFO) {
		MOD_DEC_USE_COUNT;
		return;
	}
	
	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */

	if ((ch = verifyChannel(tty)) != NULL) {

		save_flags(flags);
		cli();

		if (tty_hung_up_p(filp)) {
            MOD_DEC_USE_COUNT;
            restore_flags(flags);
			return;
		}

		/* Decrement the appropriate open count. */
		if( istpr(tty->device) ) {
			if (--ch->tpr_count < 0) {
				epca_mess( KERN_WARNING,
					"pc_close: negative tpr open count, %d\n", ch->tpr_count );
				ch->tpr_count = 0;
			}
		}
		else {
			if (--ch->tty_count < 0) {
				epca_mess( KERN_WARNING,
					"pc_close: negative tty open count, %d\n", ch->tty_count );
				ch->tty_count = 0;
			}
		}

		/* Check to see if the channel is open more than once */
		if ( ch->tpr_count + ch->tty_count > 0) {
			/*
			 * Return without doing anything.  Someone might
			 * still be using the channel.
			 */

			/* If we are closing the transparent print device,
			 * and sending data to the printer, be sure a
			 * final transparent print off string is sent. */
			if( ch->tpr_count == 0 && ch->tpr && ch->doing_tpr ) {
				write_final_offstr( ch->tpr, 1000 );	/* 10 sec timeout */
			}

			/* On the last tty close, set the tty pointer to null,
			 * since linux will now free up the tty_struct tty
			 * points to.  The same for tpr. */
			if( ch->tty_count == 0 )
				ch->tty = NULL;
			if( ch->tpr_count == 0 )
				ch->tpr = NULL;

#ifdef MODULE
    	MOD_DEC_USE_COUNT;
#endif

			restore_flags(flags);
			return;
		}

		/* Port open only once go ahead with shutdown & reset */

		/*
		 * Let the rest of the driver know the channel is being closed.
		 * This becomes important if an open is attempted before close 
		 * is finished.
		 */

		ch->asyncflags |= ASYNC_CLOSING;
	
		/*
		 * Save the termios structure, since this port may have
		 * separate termios for callout and dialin.
		 */

		if (ch->asyncflags & ASYNC_NORMAL_ACTIVE)
			ch->normal_termios = *tty->termios;

		if (ch->asyncflags & ASYNC_CALLOUT_ACTIVE)
			ch->callout_termios = *tty->termios;

		/* Whether the passed in tty is for a printer or tty, */
		/* since the count is 0, we're closing them both. */
		if( ch->tpr )
			ch->tpr->closing = 1;
		if( ch->tty )
			ch->tty->closing = 1;

		/* If we are sending data to the printer, be sure
		 * a final transparent print off string is sent. */
		if( ch->doing_tpr && (ch->tpr || ch->tty) ) {
			/* 10 sec timeout */
			write_final_offstr( ch->tty ? ch->tty : ch->tpr, 1000 );
		}

		if (ch->asyncflags & ASYNC_INITIALIZED) {
			/* Setup an event to indicate when the transmit
			 * buffer empties */
			setup_empty_event(tty, ch);		
			tty_wait_until_sent(tty, 3000); /* 30 sec timeout */
		}
	
		if (tty->driver.flush_buffer)
		{
			tty->driver.flush_buffer(tty);
		}

		if (tty->ldisc.flush_buffer)
		{
			tty->ldisc.flush_buffer(tty);
		} 

		shutdown(ch, tty);
		if( ch->tpr )
			ch->tpr->closing = 0;
		if( ch->tty )
			ch->tty->closing = 0;
		ch->tpr = ch->tty = NULL;

		if (ch->blocked_open) {
			if (ch->close_delay) {
				current->state = TASK_INTERRUPTIBLE;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
				current->timeout = jiffies + ch->close_delay;
				schedule();
#else
				schedule_timeout(ch->close_delay);
#endif
			}

			wake_up_interruptible(&ch->open_wait);

		}

		ch->asyncflags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_INITIALIZED | 
				    ASYNC_CALLOUT_ACTIVE | ASYNC_CLOSING);
		wake_up_interruptible(&ch->close_wait);

#ifdef MODULE
		MOD_DEC_USE_COUNT;
#endif

		restore_flags(flags);

	}
}


/*
 * shutdown()
 */
static void shutdown(struct channel *ch, struct tty_struct *tty )
{
	unsigned long flags;
	volatile struct board_chan *bc;

	if (!(ch->asyncflags & ASYNC_INITIALIZED)) 
		return;

	save_flags(flags);
	cli();
	globalwinon(ch);

	bc = ch->brdchan;

	/*
	 * In order for an event to be generated on the receipt of data the
	 * idata flag must be set. Since we are shutting down, this is not 
	 * necessary clear this flag.
	 */
	if (bc)
		bc->idata = 0;

	/*
	 * Okay..TX buffer's flushed, at this point. Lets go ahead and
	 * set the board for RESUMETX, so that when we REOPEN the device,
	 * it'll be ready to go, rather than perhaps being set in PAUSETX
	 * state, because of a TCXONC/TCOOFF subioctl call, or being in a
	 * software flow control state at time of port close.
	 */

	fepcmd(ch, RESUMETX, 0, 0, 0, 0);

	/* Recover from any TIOCMBIC's turning DTR or RTS off, on port close */
	ch->modem |= (ch->m_dtr | ch->m_rts);
	
	/*
	 * If we're a modem control device and HUPCL is on, and digiext
	 * flag DIGI_PRINTER" is NOT set, drop RTS & DTR.
	 */
	if(! (ch->digiext.digi_flags & DIGI_PRINTER))
	{
		if ( tty->termios->c_cflag & HUPCL ) {
			ch->omodem &= ~(ch->m_rts | ch->m_dtr);
			fepcmd(ch, SETMODEM, 0, ch->m_dtr | ch->m_rts, 10, 1);
		}
	}

	memoff(ch);

	/*
	 * The channel has officialy been closed.  The next time it is opened
	 * it will have to reinitialized.  Set a flag to indicate this.
	 */

	/* Prevent future Digi programmed interrupts from coming active */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	disable_bh(DIGI_BH);
#endif

	ch->asyncflags &= ~ASYNC_INITIALIZED;
	restore_flags(flags);

}


/*
 * pc_hangup()
 */
static void pc_hangup(struct tty_struct *tty)
{

	struct channel *ch;
	unsigned long flags;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if((ch = verifyChannel(tty)) == NULL) {
		APR(("ch is NULL in pc_hangup()\n"));
		return;
	}
	
	save_flags(flags);
	cli();
	pc_flush_buffer(tty);
	shutdown(ch, tty);

	if( istpr(tty->device) ) {
		ch->tpr = NULL;
		ch->tpr_count = 0;
	}
	else {
		ch->tty = NULL;
		ch->tty_count = 0;
	}

	ch->asyncflags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
	wake_up_interruptible(&ch->open_wait);
	restore_flags(flags);
}


/*
 * pc_write()
 *
 * pc_write is primarily called directly by the kernel routine
 * tty_write (Though it can also be called by put_char) found in
 * tty_io.c.  pc_write is passed a line discipline buffer where 
 * the data to be written out is stored.  The line discipline 
 * implementation itself is done at the kernel level and is not 
 * brought into the driver.  
 *
 *
 * The transmit buffer is a circular buffer.  You can skip the
 * rest of this comment if you are familiar with the concept.
 *
 * Two situations can affect how space in the transmit buffer
 * is calculated.  You can have a situation where the transmit
 * in pointer (tin) head has wrapped around and actually has a 
 * lower address than the transmit out pointer (tout) tail; or
 * the transmit in pointer (tin) head will not be wrapped
 * around yet, and have a higher address than the transmit
 * out pointer (tout) tail.  Obviously space available in
 * the transmit buffer is calculated differently for each case.
 *
 * Example 1:
 *
 * Consider a 10 byte buffer where head is a pointer to the next
 * empty location in the buffer and tail is a pointer to the next 
 * byte to transmit.  In this example head will not have wrapped 
 * around and therefore head > tail.  
 *
 * 0      1      2      3      4      5      6      7      8      9   
 *              tail                               head
 *
 * The above diagram shows that buffer locations 2,3,4,5 and 6 have
 * data to be transmited, while head points at the next empty
 * location.  To calculate how much space is available first we have
 * to determine if the head pointer (tin) has wrapped.  To do this
 * compare the head pointer to the tail pointer,  If head is equal
 * or greater than tail; then it has not wrapped; and the space may
 * be calculated by subtracting tail from head and then subtracting
 * that value from the buffers size.  A one is subtracted from the
 * new value to indicate how much space is available between the 
 * head pointer and end of buffer; as well as the space between the
 * begining of the buffer and the tail.  If the head is not greater
 * or equal to the tail this indicates that the head has wrapped
 * around to the begining of the buffer.  To calculate the space 
 * available in this case simply subtract head from tail.  This new 
 * value minus one represents the space available betwwen the head 
 * and tail pointers.  In this example head (7) is greater than tail (2)
 * and therefore has not wrapped around.  We find the space by first
 * subtracting tail from head (7-2=5).  We then subtract this value
 * from the buffer size of ten and subtract one (10-5-1=4).  The space
 * remaining is 4 bytes. 
 *
 * Example 2:
 *
 * Consider a 10 byte buffer where head is a pointer to the next
 * empty location in the buffer and tail is a pointer to the next 
 * byte to transmit.  In this example head will wrapped around and 
 * therefore head < tail.  
 *
 * 0      1      2      3      4      5      6      7      8      9   
 *              head                               tail
 *
 * The above diagram shows that buffer locations 7,8,9,0 and 1 have
 * data to be transmited, while head points at the next empty
 * location.  To find the space available we compare head to tail.  If
 * head is not equal to, or greater than tail this indicates that head
 * has wrapped around. In this case head (2) is not equal to, or
 * greater than tail (7) and therefore has already wrapped around.  To
 * calculate the available space between the two pointers we subtract
 * head from tail (7-2=5).  We then subtract one from this new value
 * (5-1=4).  We have 5 bytes empty remaining in the buffer.  Unlike the
 * previous example these five bytes are located between the head and
 * tail pointers. 
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
static int pc_write(struct tty_struct * tty, int from_user,
                    const unsigned char *buf, int bytesAvailable)
#else
static int pc_write(struct tty_struct * tty, int from_user,
                    unsigned char *buf, int bytesAvailable)
#endif
{ /* Begin pc_write */

	register unsigned int head, tail;
	register int dataLen;
	register int size;
	register int amountCopied;

	struct channel *ch;
	unsigned long flags;
	int remain;
	volatile struct board_chan *bc;

	int tprlen;
	char *tprpt;

	/* Stop users from hurting themselves on control minor */
	if (tty->driver.subtype == SERIAL_TYPE_INFO)
		return (0) ;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */

	if ((ch = verifyChannel(tty)) == NULL)
		return 0;

	/*
	 * Don't allow data out the printer device after it
	 * is closed and a final printer off string is sent.
	 */
	if( istpr(tty->device) && ch->tpr_count <= 0 && !ch->doing_tpr )
		return( 0 );

	/* Make a pointer to the channel data structure found on the board. */
	bc   = ch->brdchan;
	size = ch->txbufsize;

	if (from_user) {
		save_flags(flags);
		cli();

		globalwinon(ch);

		/*
		 * Anding against size will wrap the pointer back to its
		 * begining  position if it is necessary.  This will only
		 * work if size is a power of 2 which should always be the
		 * case.  Size is determined  by the card's on board FEP/OS.
		 *
		 * head refers to the next empty location in which data
		 * may be stored
		 *
		 * tail refers to the next data byte to be transmitted
		 */
		head = bc->tin & (size - 1);
		tail = bc->tout;
		if (tail != bc->tout) 	/* Consider using a do statement */
			tail = bc->tout;
		tail &= (size - 1);

		/* circular buffer */
		dataLen = ((head >= tail) ?
			   (size - (head - tail) - 1) :
			   (tail - head - 1));

		/*
		 * In this case bytesAvailable has been passed into pc_write
		 * and represents the amount of data that needs to be written.
		 * dataLen represents the amount of space available on the
		 * card.  Whichever value is smaller will be the amount
		 * actually written.  bytesAvailable will then take on
		 * this newly calculated value.
		 */

		bytesAvailable = MIN(dataLen, bytesAvailable);


		/* First we read the data into a temp buffer */

		if (bytesAvailable) {


			/*
			 * copy_from_user() reads data from user memory
			 * and cannot be used in interrupt context as it can
			 * block.
			 *
			 * You must include <asm/segment.h>
			 * I also think (Check hackers guide) that
			 * optimization must be turned ON.  (Which
			 * (sounds strange to me...)
			 */
			if(!verify_area(VERIFY_READ,
					(char*)buf, bytesAvailable)) {
				copy_from_user(ch->tmp_buf,buf,bytesAvailable);
			} else {
				/* Can't do; try again later */
				bytesAvailable = 0; 
			}
		}

		/*
		 * Set buf to this address for the moment.  tmp_buf was
		 * allocated in	post_fep_init.
		 */
		buf = ch->tmp_buf;
		memoff(ch);
		restore_flags(flags);
	}


	/* All data are now local */

	amountCopied = 0;
	save_flags(flags);
	cli();

	globalwinon(ch);

	head = bc->tin & (size - 1);
	tail = bc->tout;

	if (tail != bc->tout)
		tail = bc->tout;
	tail &= (size - 1);

	/* If head >= tail, head has not wrapped around. */ 
	if (head >= tail) {
		/*
		 * remain (much like dataLen above) represents the total
		 * amount of space available on the card for data.  Here
		 * dataLen represents the space existing between the head
		 * pointer and the end of buffer.  This is important because
		 * a memcpy cannot be told to automatically wrap around when
		 * it hits the buffer end.
		 */
		dataLen = size - head;
		remain = size - (head - tail) - 1;
	} else {
		remain = tail - head - 1;
		dataLen = remain;
	}

	/*
	 * Limit printer output to maxcps overall, with bursts allowed
	 * up to bufsize characters.
	 */
	bytesAvailable = maxcps_room(tty, ch, bytesAvailable);

	/* Need to send transparent print ON or OFF string? */
	tprlen = 0;
	tprpt = NULL;
	if( istpr(tty->device) && ch->tpr_count > 0 ) {
		/* We are writing to an open printer device. */
		if( !ch->doing_tpr ) {
			/* Need to send Transparent Print ON string. */
			if( remain > ch->digiext.digi_onlen && bytesAvailable > 0 ) {
				/* And we can send print on and some data, too. */
				tprlen = ch->digiext.digi_onlen;
				tprpt = ch->digiext.digi_onstr;
				ch->doing_tpr = 1;
			}
			else {
				/* Can't send printer on and data.  Send nothing. */
				bytesAvailable = 0;
			}
		}
	}
	else if( ch->doing_tpr ) {
		/* We are writing to the tty device or a closing printer device, */
		/* and we need to send Transparent Print OFF string. */
		if( remain > ch->digiext.digi_onlen
		&& (bytesAvailable > 0 || ch->tpr_count == 0) ) {
			/* And we can send print off and some data, or else
			 * the print device is closing. */
			tprlen = ch->digiext.digi_offlen;
			tprpt = ch->digiext.digi_offstr;
			ch->doing_tpr = 0;
		}
		else {
			/* Can't send printer off and data.  Send nothing. */
			bytesAvailable = 0;
		}
	}

	/*
	 * Check the space on the card.  If we have more data than 
	 * space; reduce the amount of data to fit the space.
	 */
	bytesAvailable = MIN(remain-tprlen, bytesAvailable);
	if( bytesAvailable < 0 )
		bytesAvailable = 0;

	txwinon(ch);
	while (bytesAvailable > 0 || (tprlen > 0 && remain > tprlen)) {
		/*
		 * If head is not wrapped, the below will make sure the first 
		 * data copy fills to the end of card buffer.
		 */
		if( tprlen ) {	/* Transparent print ON or OFF strings */
			dataLen = MIN(tprlen, dataLen);
			memcpy(ch->txptr + head, tprpt, dataLen);
			tprpt += dataLen;
			head += dataLen;
			/*amountCopied += dataLen;	Don't report tpr ON/OFF bytes to user */
			tprlen -= dataLen;

		} else {
			dataLen = MIN(bytesAvailable, dataLen);
			memcpy(ch->txptr + head, buf, dataLen);
			buf += dataLen;
			head += dataLen;
			amountCopied += dataLen;
			bytesAvailable -= dataLen;
		}

		if (head >= size) {

			head = 0;
			dataLen = tail;
		}
	}

	ch->statusflags |= TXBUSY;
	globalwinon(ch);
	bc->tin = head;

	if ((ch->statusflags & LOWWAIT) == 0) {
		ch->statusflags |= LOWWAIT;
		bc->ilow = 1;
	}
	memoff(ch);
	restore_flags(flags);

	/* Update printer buffer empty time. */
	if( istpr(tty->device)
	&& ch->digiext.digi_maxcps > 0 && ch->digiext.digi_bufsize > 0 )
		ch->cps_time += (HZ*amountCopied)/ch->digiext.digi_maxcps;

	/*
	 * This fixes a problem that shows up with some applications
	 * (e.g: "screen") which dribble packets to the card in such
	 * a way that LOW WATER events aren't triggered.  Thus transmits
	 * would hang until some other event is generated on that channel.
	 */
	if( ch->tty )
		setup_empty_event(ch->tty,ch);
	if( ch->tpr )
		setup_empty_event(ch->tpr,ch);

	return(amountCopied);

}


/*
 * pc_put_char()
 *
 */
static void pc_put_char(struct tty_struct *tty, unsigned char c)
{
	pc_write(tty, 0, &c, 1);
	return;
}


/*
 * pc_write_room()
 *
 */
static int pc_write_room(struct tty_struct *tty)
{

	int remain;
	struct channel *ch;
	unsigned long flags;
	unsigned int head, tail;
	volatile struct board_chan *bc;

	remain = 0;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */

	if ((ch = verifyChannel(tty)) != NULL) {
		save_flags(flags);
		cli();
		globalwinon(ch);

		bc   = ch->brdchan;
		head = bc->tin & (ch->txbufsize - 1);
		tail = bc->tout;

		if (tail != bc->tout)
			tail = bc->tout;
		/* Wrap tail if necessary */
		tail &= (ch->txbufsize - 1);

		if ((remain = tail - head - 1) < 0 )
			remain += ch->txbufsize;

		if (remain && (ch->statusflags & LOWWAIT) == 0) {
			ch->statusflags |= LOWWAIT;
			bc->ilow = 1;
		}

		/* Limit printer to maxcps */
		remain = maxcps_room( tty, ch, remain );

		/* Leave room for printer on/off strings */
		if( istpr(tty->device) && ch->tpr_count > 0 ) {
			if( !ch->doing_tpr ) {
				remain -= ch->digiext.digi_onlen;
			}
		}
		else if( ch->doing_tpr ) {
			remain -= ch->digiext.digi_offlen;
		}
		if( remain < 0 )
			remain = 0;

		memoff(ch);
		restore_flags(flags);
	}

	/* Return how much room is left on card */
	return remain;

}


/*
 * pc_chars_in_buffer()
 */
static int pc_chars_in_buffer(struct tty_struct *tty)
{

	int chars;
	unsigned int ctail, head, tail;
	int remain;
	unsigned long flags;
	struct channel *ch;
	volatile struct board_chan *bc;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) == NULL)
		return(0);

	save_flags(flags);
	cli();
	globalwinon(ch);

	bc = ch->brdchan;
	tail = bc->tout;
	head = bc->tin;
	ctail = ch->mailbox->cout;

	if (tail == head && ch->mailbox->cin == ctail && bc->tbusy == 0) {
		chars = 0;
	} else {

		head = bc->tin & (ch->txbufsize - 1);
		tail &= (ch->txbufsize - 1);

		/*
		 * The logic here is basically opposite of the above
		 * pc_write_room here we are finding the amount of bytes
		 * in the buffer filled.  Not the amount of bytes empty.
		 */

		if ((remain = tail - head - 1) < 0 )
			remain += ch->txbufsize;

		chars = (int)(ch->txbufsize - remain);

		/*
		 * Make it possible to wakeup anything waiting for output
		 * in tty_ioctl.c, etc.
		 * 
		 * If not already set.  Setup an event to indicate when the
		 * transmit buffer empties 
		 */
		if (!(ch->statusflags & EMPTYWAIT))
			setup_empty_event(tty,ch);

	}

	memoff(ch);
	restore_flags(flags);

	/* Return number of characters residing on card. */
	return(chars);

}


/*
 * pc_flush_buffer()
 */
static void pc_flush_buffer(struct tty_struct *tty)
{

	int i;
	unsigned long flags;
	struct channel *ch;
	volatile struct board_chan *bc;
	struct tty_struct *tty2;


	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) == NULL)
		return;

	save_flags(flags);
	cli();

	globalwinon(ch);

	bc   = ch->brdchan;
	/* Have FEP move tin pointer; effectively flushing transmit buffer */
	fepcmd(ch, STOUT, (unsigned) bc->tin, 0, 0, 0);

	memoff(ch);
	restore_flags(flags);

	for( i=0; i<2; i++ ) {

		tty2 = (i == 0 ? ch->tty : ch->tpr);

		if( !tty2 )
			continue;

		wake_up_interruptible(&tty2->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
		wake_up_interruptible(&tty2->poll_wait);
#endif
		if ((tty2->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
			tty2->ldisc.write_wakeup)
			(tty2->ldisc.write_wakeup)(tty2);

	}

}


/*
 * pc_flush_chars()
 */
static void pc_flush_chars(struct tty_struct *tty)
{

	struct channel *ch;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) != NULL) {
		unsigned long flags;

		save_flags(flags);
		cli();

		/*
		 * If not already set and the transmitter is busy setup an
		 * event to indicate when the transmit empties.
		 */
		if ((ch->statusflags & TXBUSY) && !(ch->statusflags & EMPTYWAIT)) {
			if( ch->tty )
				setup_empty_event(ch->tty,ch);
			if( ch->tpr )
				setup_empty_event(ch->tpr,ch);
		}

		restore_flags(flags);
	}

}


/*
 * write_final_offstr
 *
 * Write a final transparent off string when closing the transparent
 * print device.
 */
static void write_final_offstr( struct tty_struct * tty, long timeout )
{
	struct channel *ch;
	DECLARE_WAITQUEUE( wait, current );

	if( (ch=verifyChannel(tty)) == NULL)
		return;

	add_wait_queue( &tty->write_wait, &wait );

	while( ch->doing_tpr && timeout ) {
		current->state = TASK_INTERRUPTIBLE;
		if( signal_pending( current ) )
			break;
		pc_write( tty, 0, NULL, 0 );
		if( !ch->doing_tpr )
			break;
		timeout = schedule_timeout( timeout );
	}

	current->state = TASK_RUNNING;
	remove_wait_queue( &tty->write_wait, &wait );
}


/*
 * block_til_ready()
 */
static int block_til_ready(struct tty_struct *tty, 
                           struct file *filp, struct channel *ch)
{

	DECLARE_WAITQUEUE(wait,current);

	int	retval, do_clocal = 0;
	unsigned long flags;


	if (tty_hung_up_p(filp)) {
		if (ch->asyncflags & ASYNC_HUP_NOTIFY)
			retval = -EAGAIN;
		else
			retval = -ERESTARTSYS;	
		return(retval);
	}

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (ch->asyncflags & ASYNC_CLOSING) {
		interruptible_sleep_on(&ch->close_wait);

		if (ch->asyncflags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */

	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (ch->asyncflags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;

		if ((ch->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
		    (ch->asyncflags & ASYNC_SESSION_LOCKOUT) &&
		    (ch->session != current->session))
			return -EBUSY;

		if ((ch->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
		    (ch->asyncflags & ASYNC_PGRP_LOCKOUT) &&
		    (ch->pgrp != current->pgrp))
			return -EBUSY;
 
		ch->asyncflags |= ASYNC_CALLOUT_ACTIVE;

		return 0;
	}

	if (filp->f_flags & O_NONBLOCK) {
		/*
	  	 * If non-blocking mode is set, then make the check up front
	  	 * and then exit.
		 */

		if (ch->asyncflags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;

		ch->asyncflags |= ASYNC_NORMAL_ACTIVE;

		return 0;
	}


	if (ch->asyncflags & ASYNC_CALLOUT_ACTIVE) {
		if (ch->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/* Block waiting for the carrier detect and the line to become free */
	retval = 0;
	add_wait_queue(&ch->open_wait, &wait);
	save_flags(flags);
	cli();

	/* We dec count so that pc_close will know when to free things */
	if (!tty_hung_up_p(filp)) {
		if( istpr(tty->device) )
			--ch->tpr_count;
		else
			--ch->tty_count;
	}

	restore_flags(flags);

	ch->blocked_open++;

	while(1) {

		current->state = TASK_INTERRUPTIBLE;

		if (tty_hung_up_p(filp) ||
		    !(ch->asyncflags & ASYNC_INITIALIZED)) {
			if (ch->asyncflags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
			break;
		}

		if (!(ch->asyncflags & ASYNC_CLOSING) && 
		    !(ch->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
		    (do_clocal || (ch->imodem & ch->dcd)))
			break;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

		if (current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}
#else
 		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
#endif

		/*
		 * Allow someone else to be scheduled.  We will occasionaly go
		 * through this loop until one of the above conditions change.
		 * The below schedule call will allow other processes to enter
		 * and prevent this loop from hogging the cpu.
		 */
		schedule();
	}

	current->state = TASK_RUNNING;
	remove_wait_queue(&ch->open_wait, &wait);
	cli();
	if (!tty_hung_up_p(filp)) {
		if( istpr(tty->device) )
			++ch->tpr_count;
		else
			++ch->tty_count;
	}
	restore_flags(flags);

	ch->blocked_open--;

	if (retval)
		return retval;

	ch->asyncflags |= ASYNC_NORMAL_ACTIVE;

	return 0;

}


/*
 * bankname_to_offset()
 */
static int bankname_to_offset( char bankname )
{
	char *bp;
	bp = epca_banknames;
	while( *bp ) {
		if( *bp == bankname )
			return( NDEV_MAJOR * (bp-epca_banknames) );
		++bp;
	}

	/* On error, return as if we matched the first bank */
	return( 0 );
}


/*
 * name_to_offset()
 *
 * Given a port name (e.g: ttyD15, ttyE200 or cud5), determine the
 * offset (if any) that must be added to the device's minor number
 * to get to the correct entry in the channel arrays (e.g: digi_channels,
 * pc_table, pc_termios, etc.)
 *
 */
static int name_to_offset( char *drivername )
{

	int offset, lookhere;
	char bankname;

	offset = 0;

	/* 
	 * First see if this is one of the drivers that supports
	 * >256 ports e.g: "ttyD*".
	 */
	lookhere = strlen(TTY_NAME)-1;
	if( !strncmp( TTY_NAME, drivername, lookhere ) ) {
		/* Now get the letter designator (e.g: the 'D' in "ttyD*") */
		bankname = drivername[ lookhere ];
		offset = bankname_to_offset( bankname );
	}

	return( offset );
}



/*
 * pc_open()
 */
int pc_open(struct tty_struct *tty, struct file * filp)
{

	struct channel *ch;
	unsigned long flags;
	int line, retval, boardnum;
	volatile struct board_chan *bc;
	volatile unsigned int head;
	int lineoffset;

	/* Nothing "real" happens in open of control device */
	if (tty->driver.subtype == SERIAL_TYPE_INFO) {
		MOD_INC_USE_COUNT;
        return (0) ;
	}

	/* First bank of 256 channels are named "ttyD*", next "ttyE*", etc. */
	lineoffset = name_to_offset( (char *)tty->driver.name );
	line = MINOR(tty->device) - tty->driver.minor_start + lineoffset;


	if (line < 0 || line >= nbdevs) {
		epca_mess( KERN_ERR, "pc_open: line out of range\n");
		tty->driver_data = NULL;
		return(-ENODEV);
	}

	if( !digi_channels[line] ) {
		epca_mess( KERN_ERR, "pc_open found null channel structure "
			   "for line:%d\n", line );

		tty->driver_data = NULL;   /* Mark this device as 'down' */
		return(-ENODEV);
	}

#if 0
	ch = &digi_channels[line];
#endif
	ch = digi_channels[line];
	
	boardnum = ch->boardnum;



	/*
	 * Check status of board configured in system.
	 *
	 * I check to see if the epca_setup routine detected an user error.  
	 * It might be better to put this in pc_init, but for the moment it
	 * goes here.
	 */

	if (invalid_lilo_config) {
		if (setup_error_code & INVALID_BOARD_TYPE)
			epca_mess(KERN_ERR, "pc_open: Invalid board "
				  "type specified in LILO command\n");

		if (setup_error_code & INVALID_NUM_PORTS)
			epca_mess(KERN_ERR, "pc_open: Invalid number "
				  "of ports specified in LILO command\n");

		if (setup_error_code & INVALID_MEM_BASE)
			epca_mess(KERN_ERR, "pc_open: Invalid board memory "
				  "address specified in LILO command\n");

		if (setup_error_code & INVALID_PORT_BASE)
			epca_mess(KERN_ERR, "pc_open: Invalid board port "
				  "address specified in LILO command\n");

		if (setup_error_code & INVALID_BOARD_STATUS)
			epca_mess(KERN_ERR, "pc_open: Invalid board status "
				  "specified in LILO command\n");

		if (setup_error_code & INVALID_ALTPIN)
			epca_mess(KERN_ERR, "pc_open: Invalid board altpin "
				  "specified in LILO command\n");

		tty->driver_data = NULL;   /* Mark this device as 'down' */
		return(-ENODEV);
	}

	if ((boardnum >= num_cards)
	    || (boards[boardnum].status == DISABLED)
	    || (boards[boardnum].initted != 1)) {
		tty->driver_data = NULL;   /* Mark this device as 'down' */
		return(-ENODEV);
	}
	
	if (( bc = ch->brdchan) == 0) {
		tty->driver_data = NULL;
		return(-ENODEV);
	}

	/*
	 * Every time a channel is opened, increment a counter.  This is 
	 * necessary because we do not wish to flush and shutdown the channel
	 * until the last app holding the channel open, closes it.
	 */
	if( istpr(tty->device) )
		++ch->tpr_count;
	else
		++ch->tty_count;
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

	if (ch->tty_count + ch->tpr_count == 1) {
		/*
		 * Set a kernel structures pointer to our local channel 
		 * structure.  This way we can get to it when passed only
		 * a tty struct.
		 */
		tty->driver_data = ch;
	
		/*
		 * If this is the first time the channel has been opened,
		 * initialize the tty->termios struct otherwise let
		 * pc_close handle it.
		 */

		/* Should this be here except for SPLIT termios ? */
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = ch->normal_termios;
		else 
			*tty->termios = ch->callout_termios;

		ch->session = current->session;
		ch->pgrp = current->pgrp;
	
		save_flags(flags);
		cli();
	
		globalwinon(ch);
		ch->statusflags = 0;
	
		/* Save boards current modem status */
		ch->imodem = bc->mstat;
	
		/*
		 * Set receive head and tail ptrs to each other.  This
		 * indicates no data available to read.
		 */
		head = bc->rin;
		bc->rout = head;
	
		/* Set the channel's associated tty structure */
		if( istpr(tty->device) )
			ch->tpr = tty;
		else
			ch->tty = tty;
	
		/*
		 * The below routine generally sets up parity, baud, flow
		 * control issues, etc.... It effect both control flags
		 * and input flags.
		 */
		epcaparam(tty,ch);
	
		ch->asyncflags |= ASYNC_INITIALIZED;
		memoff(ch);
	
		restore_flags(flags);
	
		retval = block_til_ready(tty, filp, ch);
		if (retval) {

#if ALLOW_ALT_FAIL_OPEN
            if(mod_alt_fail_open)
            {
                APR(("dgdm_tty_open failed, "
                           "calling close from open for failed open."));
                pc_close(tty, filp);
	        printk("Finished calling close for failed open.");
            }
#endif

            return retval;
		}
	
		/*
		 * Set this again in case a hangup set it to zero while this 
		 * open() was waiting for the line...
		 */
		if( istpr(tty->device) )
			ch->tpr = tty;
		else
			ch->tty = tty;
	
		save_flags(flags);
		cli();
		globalwinon(ch);
	
		/* Enable Digi Data events */
		bc->idata = 1;
	
		memoff(ch);
		restore_flags(flags);
	}
    else
    {
		tty->driver_data = ch;
		if( istpr(tty->device) )
			ch->tpr = tty;
		else
			ch->tty = tty;
        retval = block_til_ready(tty, filp, ch);
		if (retval) {

#if ALLOW_ALT_FAIL_OPEN
            if(mod_alt_fail_open)
            {
                APR(("dgdm_tty_open failed, "
                           "calling close from open for failed open."));
                pc_close(tty, filp);
            }
            printk("Finished calling close for failed open.");
#endif

            return retval;
		}
    }

	return 0;

}



#if defined(MODULE)
/*
 * init_module()
 */
int init_module()
{
	return( pc_init() );
}


/*
 * cleanup_module()
 */
void cleanup_module()
{

	int               count, crd, i, first;
	struct board_info *bd;
	struct channel    *ch;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
	if (pc_timeron) {
		pc_timeron = 0;
		del_timer(&pc_timer);
	}
#else
	timer_table[DIGI_TIMER].fn = 0;
#endif

	for (crd = 0; crd < num_cards; crd++) {

		bd = &boards[crd];

		if (!bd) {
			epca_mess(KERN_ERR, "cleanup_module failed\n");
			return;
		}

		/* Don't try to release resources that aren't been allocated */
		if( !bd->initted )
			continue;

		ch = card_ptr[crd]; 

		/* Get the line offset for the first port on this card */
#if 0
		first = firstchan( crd );
#endif
		first = bd->firstchan;
		for (count = 0; count < bd->numports; count++) {
			ch = digi_channels[first+count];

			if (ch) {
				ch->asyncflags &= ~ASYNC_INITIALIZED;
				if (ch->tty) {
					/*
					 * I don't think we will ever
					 * get here: if the port has been
					 * closed with pc_close, then ch->tty
					 * is NULL.  If the port is still open,
					 * rmmod will complain that the
					 * resource is busy and exit before
					 * calling cleanup_module()
					 */
					tty_hangup(ch->tty);
				}
				if (ch->tpr) {
					tty_hangup(ch->tpr);
				}

				if( ch->tmp_buf ) {
					kfree(ch->tmp_buf);
					ch->tmp_buf = NULL;
				}
			}
		}
	}

#ifndef CONFIG_DEVFS_FS

        {
                int bank, freeguts;

                for( bank=0; bank<n_banks; ++bank ) {
			freeguts = 0;
                        if (tty_unregister_driver(pc_driver[bank]) ) {
                                epca_mess(KERN_WARNING, "cleanup_module failed"
                                          " to un-register pc driver, "
                                          "bank:%d\n", bank);
                        } else {
                                kfree( (char *)pc_driver[bank]->name );
                                pc_driver[bank]->name = NULL;
				freeguts = 1;
				/* Wait until after unregistering the pr_driver (which refers to these)
                                 * vfree( pc_driver[bank]->table );
                                 * vfree( pc_driver[bank]->termios );
                                 * vfree( pc_driver[bank]->termios_locked );
                                 * kfree( pc_driver[bank] );
                                 * pc_driver[bank] = NULL;
				 */
                        }

                        if (tty_unregister_driver(pr_driver[bank]) ) {
                                epca_mess(KERN_WARNING, "cleanup_module failed"
                                          " to un-register pr driver, "
                                          "bank:%d\n", bank);
                        } else {
                                kfree( (char *)pr_driver[bank]->name );
                                pr_driver[bank]->name = NULL;
                                /* vfree( pr_driver[bank]->table );
                                 * vfree( pr_driver[bank]->termios );
                                 * vfree( pr_driver[bank]->termios_locked );
				 */
                                kfree( pr_driver[bank] );
                                pr_driver[bank] = NULL;
                        }

			if( freeguts ) {
                                vfree( pc_driver[bank]->table );
                                vfree( pc_driver[bank]->termios );
                                vfree( pc_driver[bank]->termios_locked );
                                kfree( pc_driver[bank] );
                                pc_driver[bank] = NULL;
			}
                }
	}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	if (tty_unregister_driver(&pc_callout) ) {
		epca_mess(KERN_WARNING, "cleanup_module failed to "
			  "un-register pc_callout driver\n");
	} else {
		if( pc_callout.termios_locked )
			vfree( pc_callout.termios_locked );
		if( pc_callout.termios )
			vfree( pc_callout.termios );
		if( pc_callout.table )
			vfree( pc_callout.table );
	}
#endif

	if (tty_unregister_driver(&pc_info) )
		epca_mess(KERN_WARNING, "cleanup_module failed to "
			  "un-register pc_info driver\n");

	/* Free up the memory for each channel structures */
	for( i=0; i<nbdevs; ++i ) {
		if( digi_channels[i] ) {
			kfree( digi_channels[i] );
			digi_channels[i] = NULL;
		}
	}
	/* Free up the array of channel structure pointers */
	kfree( digi_channels );
	digi_channels = NULL;

	/* Free up buffers used for the C/X and EPC/X concentrator OSes */
	while( num_conc_images )
	{
		char *tmp_image;

		--num_conc_images;
		tmp_image = conc_images[num_conc_images];
		conc_images[num_conc_images] = NULL;
		kfree( tmp_image );
	}


#ifdef CONFIG_PROC_FS
# if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	proc_unregister( &proc_root, epca_proc_entry.low_ino);
# else
	remove_proc_entry("epca", NULL);
# endif
#endif /* CONFIG_PROC_FS */

	/* Free up the array of tty_driver arrays */
	if( pc_driver )
		kfree( pc_driver );
	pc_driver = NULL;
	if( pr_driver )
		kfree( pr_driver );
	pr_driver = NULL;
}
#endif /* MODULE */


#ifndef ENABLE_PCI
/*
 * epca_removepci()
 */
void epca_removepci(void)
{

	int itmp, pci_offset=0;
	struct board_info **tmpboards;
	
	/* Allocate pointers.. */
	tmpboards = (struct board_info **)
		kmalloc(sizeof(struct board_info *) * MAXBOARDS, GFP_KERNEL);
	
	/* Allocate RAM to pointers.. */
	for(itmp=0;itmp<MAXBOARDS;itmp++) {
		tmpboards[itmp] = (struct board_info *)
			kmalloc(sizeof(struct board_info), GFP_KERNEL);
	}
	
	/* Copy boards structure to tmpboards */
	for(itmp=0;itmp<MAXBOARDS;itmp++) {
		memcpy(tmpboards[itmp], &boards[itmp],
		       sizeof(struct board_info));
	}
	

	/*
	 * Now that we've saved off boards to temporary storage..
	 * Wipe out boards
	 */
	for(itmp=0;itmp<MAXBOARDS;itmp++) {
		memset(&boards[itmp], 0, sizeof(struct board_info));
	}
	
	/* Put it back together, one at a time.. */
	for(itmp=0; itmp < MAXBOARDS; itmp++) {
		if( (btypes[tmpboards[itmp]->type].dpatype & BUSTYPE) ==
		    T_PCIBUS) {
			APR(("This board is PCI! Can't use! "
                             "num_cards-- pci_offset++...\n"));
			pci_offset++;
			num_cards--;
			continue;
		}
		
		memcpy(&boards[itmp-pci_offset], tmpboards[itmp], 
		       sizeof(struct board_info));
	}
	
	/* Okay. Its back together. free up tmpboards */
	for(itmp=0; itmp < MAXBOARDS; itmp++)
		kfree(tmpboards[itmp]);
	kfree(tmpboards);
}
#endif	/* ENABLE_PCI */
	

/*
 * pc_init()
 *
 * pc_init is called by the operating system during boot up prior to
 * any open calls being made.  In the older versions of Linux (Prior
 * to 2.0.0) an entry is made into tty_io.c.  A pointer to the last
 * memory location (from kernel space) used (kmem_start) is passed
 * to pc_init.  It is pc_inits responsibility to modify this value 
 * for any memory that the Digi driver might need and then return
 * this value to the operating system.  For example if the driver
 * wishes to allocate 1K of kernel memory, pc_init would return 
 * (kmem_start + 1024).  This memory (Between kmem_start and kmem_start
 * + 1024) would then be available for use exclusively by the driver.  
 * In this case our driver does not allocate any of this kernel memory.
 */
int pc_init(void)
{

	ulong flags;
	int crd, bank;
	struct board_info *bd;
	unsigned char board_id = 0;
	int itmp;
	struct tty_driver lpc_driver;
	int line, ret = 0;
	ulong save_loops; 

#ifdef ENABLE_PCI
	int pci_boards_found;
#endif

	pollrate = HZ / 100;	/* By what rate do we poll?	*/
				/* EG: (HZ/100) * 64 == 640ms	*/


#ifdef MODULE
	 /* 
	  * Check for no command line options AND no epcaconfig options
	  * This is obviously a situation we can't handle properly, as
	  * we are banking upon the fact that we'll have to initialize
	  * at least 1 board.
	  * 
	  * If we've been loaded as a module, check to see if the user gave
	  * our parameters on the command line.  If so, they are gospel and
	  * should use them.
	  */
    if (mod_numports[0] != 0) {
	 	int i;
	 	
	 	for(i=0;i<MAXCARDS && mod_numports[i]; i++) {
	 		boards[i].status	= ENABLED;
	 		boards[i].type		= (unchar)mod_type[i];
	 		boards[i].altpin	= (unchar)mod_altpin[i];
	 		boards[i].numports	= (ushort)mod_numports[i];
			boards[i].port		= mod_port[i];
			boards[i].membase	= mod_membase[i];
			num_cards++;
			nbdevs+=mod_numports[i];
		}
	 } else {
#endif
		 /*
		  * If epca_setup has not been run by LILO, set
		  * num_cards to defaults; copy board structure
		  * defined by digiConfig into driver's board
		  * structure.  Note : If LILO has run epca_setup
		  * then epca_setup will handle defining num_cards
		  * as well as copying the dataq into the board structure.
		  */
		if(!liloconfig)	{
			nbdevs = NBDEVS;
			num_cards = NUMCARDS;
			memcpy((void*)&boards, (void*)&static_boards,
			       (sizeof(struct board_info) * NUMCARDS));
		}
#ifdef MODULE
	 }
#endif
	
	/* Yay! Now, with PCI statically defined in epcaconfig.h, we can
	 * rely on nbdevs!!!
	 */
	if(nbdevs == 0) {
		APR(("You don't have any boards configured.\n"));
		ret = -EFAULT;
		goto fail1;
	}
	
#if 0
	epca_mess(KERN_INFO, "Supporting %d ports\n",totalports);
#endif
	
	n_banks = ((nbdevs-1)/NDEV_MAJOR) + 1;

	/* Setup message buffer for epca_mess messages */
	sprintf( epca_mesbuf, KERN_INFO"epca: " );
	epca_mesbufpt = epca_mesbuf + strlen( epca_mesbuf );
	
	/* Allocate memory for the array of pointers to the channel structs. */
	if( !(digi_channels =
	      kmalloc(sizeof(struct channel *) * nbdevs , GFP_ATOMIC) )) {
		epca_mess( KERN_ERR, "Couldn't malloc channel pointer space" );
		ret = -ENOMEM;
		goto fail1; 
	}
	memset((void *)digi_channels, 0, sizeof(struct channel *) * nbdevs);

	/*
	 * Now allocate memory for the channel structures themselves and
	 * populate the digi_channel pointer array.
	 */
	for( line = 0; line < nbdevs; ++line ) {
		if( !(digi_channels[line] =
		      (struct channel *)kmalloc(sizeof(struct channel),
						GFP_ATOMIC))  ) {
			epca_mess( KERN_ERR, "Cannot allocate channel "
				   "memory for device %d\n", line );
			ret = -ENOMEM;
			goto fail2;
		} else
			memset((void *)(digi_channels[line]), 0,
			       sizeof(struct channel));
	}

	/*
	 * Allocate memory for the array of pointers to tty_drivers
	 * (one for each bank of NDEV_MAJOR devices)
	 */
	if( !(pc_driver =
	      kmalloc(sizeof(struct tty_driver *) * n_banks , GFP_ATOMIC) )) {
		epca_mess( KERN_ERR, "Couldn't malloc tty_driver array space");
		ret = -ENOMEM;
		goto fail3;
	}
	memset((void *)pc_driver, 0, sizeof(struct tty_driver *) * n_banks);

	/* ... And the same for the corresponding transparent print devices */
	if( !(pr_driver =
      		kmalloc(sizeof(struct tty_driver *) * n_banks , GFP_ATOMIC) )) {
		epca_mess( KERN_ERR, "Couldn't malloc tty_driver printer array space");
		ret = -ENOMEM;
		goto fail3;
	}
	memset((void *)pr_driver, 0, sizeof(struct tty_driver *) * n_banks);

	/*
	 * For all ISA boards, set status as BD_NOFEP, so DPA
	 * won't try to talk directly to them, right after we've
	 * just initialized the driver!
	 */
	for(itmp=0;itmp<num_cards;itmp++)
		boards[itmp].dpastatus = BD_NOFEP;
	
	/*
	 * Now that thats done, lets whip through the boards to at
	 * least verify that at least the type is correct, and that
	 * the board is advertised as having ports...
	 */
	for(itmp=0;itmp<num_cards;itmp++) {
		if(epca_paranoia(itmp, CHECKIFBDVALID)) {
			/* Oops. We were passed the wrong type.  Zero out
			 * the I/O address of the board, and mark it
			 * DISABLED..CWS
			 */
			boards[itmp].status = DISABLED;
			boards[itmp].dpastatus = BD_NOTFOUND;
			boards[itmp].initted = 0;
			boards[itmp].membase = 0;
			boards[itmp].port = 0;
		}
	}
	
	
	/*
	 * Note : If lilo was used to configure the driver and the 
	 * ignore epcaconfig option was choosen (digiepca=2) then 
	 * nbdevs and num_cards will equal 0 at this point.  This is
	 * okay; PCI cards will still be picked up if detected.
	 */
        APR(("%s, Digi International Part Number %s\n", DG_NAME, DG_PART));

#ifdef ENABLE_PCI

	/*
	 * NOTE : This code assumes that the number of ports found in 
	 *        the boards array is correct.  This could be wrong if
	 *        the card in question is PCI (And therefore has no ports 
	 *        entry in the boards structure.)  The rest of the 
	 *        information will be valid for PCI because the beginning
	 *        of pc_init scans for PCI and determines i/o and base
	 *        memory addresses.  I am not sure if it is possible to 
	 *        read the number of ports supported by the card prior to
	 *        it being booted (Since that is the state it is in when 
	 *        pc_init is run).  Because it is not possible to query the
	 *        number of supported ports until after the card has booted;
	 *        we are required to calculate the card_ptrs as the card is
	 *        initialized (Inside post_fep_init).  The negative thing
	 *        about this approach is that digiDload's call to GET_INFO
	 *        will have a bad port value.  (Since this is called prior
	 *        to post_fep_init.)
	 */
  
	pci_boards_found = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	if (pcibios_present())
#else
	if (pci_present())
#endif

	{
		/* num_cards is now entirely specified in epcaconfig.h */
		init_PCI(num_cards);
	} else { 
		epca_mess(KERN_ERR, "No PCI BIOS found\n");
	}

#else /* ENABLE_PCI */
	epca_mess(KERN_CRIT,
		  "No PCI BIOS Found - Can't support Digi PCI devices\n"
		  "If there are any PCI devices specified in your "
		  "configuration,\nthey will be REMOVED from the driver's "
		  "active config.\n");
	epca_removepci();
#endif /* ENABLE_PCI */	
	
	memset(&lpc_driver, 0, sizeof(struct tty_driver));
	lpc_driver.magic = TTY_DRIVER_MAGIC;
#ifdef CONFIG_DEVFS_FS
	lpc_driver.name="dg/epca/%d";
#else
	/* Need to malloc memory for driver.name when used */
# if 0
	lpc_driver.name =TTY_NAME;
# endif
#endif
	lpc_driver.major = DIGI_MAJOR; 
#if 0
	pc_driver.major = 0;
#endif
	lpc_driver.minor_start = 0;
	lpc_driver.num = NDEV_MAJOR;
	lpc_driver.type = TTY_DRIVER_TYPE_SERIAL;
	lpc_driver.subtype = SERIAL_TYPE_NORMAL;
	lpc_driver.init_termios = tty_std_termios;
	lpc_driver.init_termios.c_iflag = 0;
	lpc_driver.init_termios.c_oflag = 0;

	lpc_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | CLOCAL | HUPCL;
	lpc_driver.init_termios.c_lflag = 0;
	lpc_driver.flags = TTY_DRIVER_REAL_RAW;
	lpc_driver.refcount = &pc_refcount;
#if 0
	lpc_driver.table = pc_table;
#endif
	
	/* pc_termios is an array of pointers pointing at termios structs */
	/* The below should get the first pointer */
#if 0
	lpc_driver.termios = pc_termios;
	lpc_driver.termios_locked = pc_termios_locked;
#endif

	/*
	 * Setup entry points for the driver.  These are primarily called by 
	 * the kernel in tty_io.c and n_tty.c
	 */
	lpc_driver.open = pc_open;
	lpc_driver.close = pc_close;
	lpc_driver.write = pc_write;
	lpc_driver.write_room = pc_write_room;
	lpc_driver.flush_buffer = pc_flush_buffer;
	lpc_driver.chars_in_buffer = pc_chars_in_buffer;
	lpc_driver.flush_chars = pc_flush_chars;
	lpc_driver.put_char = pc_put_char;
	lpc_driver.ioctl = pc_ioctl;
	lpc_driver.set_termios = pc_set_termios;
	lpc_driver.stop = pc_stop;
	lpc_driver.start = pc_start;
	lpc_driver.throttle = pc_throttle;
	lpc_driver.unthrottle = pc_unthrottle;
	lpc_driver.hangup = pc_hangup;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	/* Callout device is deprecated. */

	pc_callout = lpc_driver;

	pc_callout.name = "cud";
	pc_callout.major = DIGICU_MAJOR;
# if 0
	pc_callout.major = 0;
# endif
	pc_callout.minor_start = 0;
	pc_callout.init_termios.c_cflag = B9600 | CS8 | CREAD | CLOCAL | HUPCL;
	pc_callout.subtype = SERIAL_TYPE_CALLOUT;
#endif

	pc_info = lpc_driver;
	pc_info.name = "digiCtl";
	pc_info.major = DIGIINFOMAJOR;
	pc_info.minor_start = 0;
	pc_info.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL;
	pc_info.subtype = SERIAL_TYPE_INFO;
	pc_info.table = pc_infotable;
	pc_info.termios = pc_infotermios;
	pc_info.termios_locked = pc_infotermios_locked;
	pc_info.num = 1;

	for( bank=0; bank<n_banks; ++bank ) {
#ifdef CONFIG_DEVFS_FS
	  /* I have no use for this, so I didn't bother doing it.
	   * I think the way it should work would be to allocate
	   * names like "/dev/digi/0/0", "/dev/digi/0/1",
	   * "/dev/digi/1/0", etc, where the first number is the
	   * bank and the second is the device.  The bank number
	   * would be omitted if there was only one.
	   *
	   * NOTE: This should be fixed before release -- FIXME
	   *       JAR 010411
	   */
	  if(bank>0) {
	    epca_mess(KERN_ERR,"Multiple banks with devfs unsupported\n");
	    break;
	  }
#endif

		if( bank >= (sizeof(TTY_BANKNAMES)-1) ) {
			epca_mess( KERN_ERR,
			"No more bank names available.  Ignoring banks "
				   ">= %d\n", bank );
			break;
		} 

		
		/* ... for the corresponding transparent print devices */
		if( !(pc_driver[bank] =
		      kmalloc(sizeof(struct tty_driver), GFP_ATOMIC)) ) {
			epca_mess( KERN_ERR, "Couldn't malloc driver "
				   "structure space, bank:%d", bank);
			ret = -ENOMEM;
			goto fail4;
		}
		memcpy( (void*)(pc_driver[bank]), (void*)(&lpc_driver),
			sizeof(struct tty_driver) );

		/* ... for the corresponding transparent print devices */
		if( !(pr_driver[bank] =
			kmalloc(sizeof(struct tty_driver), GFP_ATOMIC)) ) {
			epca_mess( KERN_ERR, "Couldn't malloc driver "
			"transparent printer structure space, bank:%d", bank);
			ret = -ENOMEM;
			goto fail4;
		}
		memcpy( (void*)(pr_driver[bank]), (void*)(&lpc_driver),
			sizeof(struct tty_driver) );

		/* How many ports in this bank */
		pc_driver[bank]->num = nbdevs - (NDEV_MAJOR*bank);
		if( pc_driver[bank]->num > NDEV_MAJOR )
			pc_driver[bank]->num = NDEV_MAJOR;
		
		pr_driver[bank]->num = pc_driver[bank]->num;


		/* 
		 * Allocate and initialize the driver name to include the
		 * current bank designator, e.g: first ttyD, then ttyE,
		 * then ttyF...
		 */
#ifndef CONFIG_DEVFS_FS
		if( !(pc_driver[bank]->name =
		      kmalloc(sizeof(TTY_NAME), GFP_ATOMIC)) ) {
			epca_mess(KERN_ERR, "Couldn't malloc name space, "
				  "bank:%d", bank);
			ret = -ENOMEM;
			goto fail4;
		}
		
		strcpy( (char *)pc_driver[bank]->name, TTY_NAME );
		*(((char *)pc_driver[bank]->name) + 
		  strlen(pc_driver[bank]->name)-1) = epca_banknames[bank]; 


		if( !(pr_driver[bank]->name = kmalloc(sizeof(TTY_NAME), GFP_ATOMIC)) ) {
			epca_mess(KERN_ERR, "Couldn't malloc printer name space, "
			"bank:%d", bank);
			ret = -ENOMEM;
			goto fail4;
		}
		
		strcpy( (char *)pr_driver[bank]->name, TPR_NAME );
		*(((char *)pr_driver[bank]->name) + 
		strlen(pr_driver[bank]->name)-1) = epca_banknames[bank]; 

#endif

		/*
		 * The first bank's major number is defined in <major.h>;
		 * subsequent banks' major numbers are allocated dynamically
		 */
		if( bank > 0 )
			pc_driver[bank]->major = 0;

		/* Allocate memory for the array of termios structures */
		if( !(pc_driver[bank]->termios = 
		      vmalloc(sizeof(struct termios) * pc_driver[bank]->num))){
			epca_mess( KERN_ERR, "Couldn't malloc termios space" );
			ret = -ENOMEM;
			goto fail4;
		}
		memset((void *)pc_driver[bank]->termios, 0,
		       sizeof(struct termios) * pc_driver[bank]->num );
		pr_driver[bank]->termios = pc_driver[bank]->termios;


		/* Allocate memory for the array of termios_locked structures*/
		if( !(pc_driver[bank]->termios_locked = 
		      vmalloc(sizeof(struct termios) * pc_driver[bank]->num))){
			epca_mess( KERN_ERR, "Couldn't malloc termios_locked "
				   "space" );
			ret = -ENOMEM;
			goto fail4;
		}
		memset((void *)pc_driver[bank]->termios_locked, 0,
		       sizeof(struct termios) * pc_driver[bank]->num );
		pr_driver[bank]->termios_locked = pc_driver[bank]->termios_locked;

		/*
		 * Allocate memory for the array of tty_struct
		 * (use vmalloc because kmalloc cannot address
		 * this much memory
		 */
		if( !(pc_driver[bank]->table = 
		      vmalloc(sizeof(struct tty_struct) *
			      pc_driver[bank]->num) )) {
			epca_mess(KERN_ERR, "Couldn't malloc tty_struct "
				  "space");
			ret = -ENOMEM;
			goto fail4;
		}
		memset((void *)pc_driver[bank]->table, 0,
		       sizeof(struct tty_struct) * pc_driver[bank]->num );

		//pr_driver[bank]->table = pc_driver[bank]->table;
		/* We need a separate bank to insulate the two separate MAJOR numbers */
//#if 0
		if( !(pr_driver[bank]->table = 
		      vmalloc(sizeof(struct tty_struct) *
			      pr_driver[bank]->num) )) {
			epca_mess(KERN_ERR, "Couldn't malloc tty_struct "
				  "space");
			ret = -ENOMEM;
			goto fail4;
		}
		memset((void *)pr_driver[bank]->table, 0,
		       sizeof(struct tty_struct) * pr_driver[bank]->num );
//#endif

		/*
		 * If we set major to 0, tty_register_driver will dynamically
		 * allocate a (positive) major number and return this.
		 * Negative values are true errors.
		 */
		if (tty_register_driver(pc_driver[bank]) < 0 ) {
			epca_mess( KERN_ERR, "Couldn't register "
				   "driver bank:%d", bank );
			ret = -EPERM;
			goto fail5;
		}



#if 0
		epca_mess( KERN_INFO, "Setting majors[%d] to %d "
			   "(n_banks:%d)\n", bank, pc_driver[bank]->major,
			   n_banks );
#endif
		majors[bank] = pc_driver[bank]->major;


		/* Now for the corresponding transparent print devices */
		pr_driver[bank]->major = 0;
		if (tty_register_driver(pr_driver[bank]) < 0 ) {
			epca_mess( KERN_ERR, "Couldn't register "
			"transparent printer bank:%d", bank );
			ret = -EPERM;
			goto fail5;
		}
		majorspr[bank] = pr_driver[bank]->major;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	/* Only bank 0 (up to first 256 units) have equiv "cu" devices. */
	pc_callout.num = pc_driver[0]->num;
	/* Allocate memory for the array of termios structures */
	if( !(pc_callout.termios = 
	      vmalloc(sizeof(struct termios) * pc_callout.num) )) {
		epca_mess( KERN_ERR, "Couldn't malloc termios space" );
		ret = -ENOMEM;
		goto fail5;
	}
	memset((void *)pc_callout.termios, 0,
	       sizeof(struct termios) * pc_callout.num );
	/* Allocate memory for the array of termios_locked structures */
	if( !(pc_callout.termios_locked = 
	      vmalloc(sizeof(struct termios) * pc_callout.num) ) ){
		epca_mess( KERN_ERR, "Couldn't malloc callout "
			   "termios_locked space" );
		ret = -ENOMEM;
		goto fail5A;
	}
	memset((void *)pc_callout.termios_locked, 0,
	       sizeof(struct termios) * pc_callout.num );
	/* Allocate memory for the table */
	if( !(pc_callout.table = 
	      vmalloc(sizeof(struct tty_struct) * pc_callout.num) )) {
		epca_mess(KERN_ERR,"Couldn't malloc callout tty_struct space");
		ret = -ENOMEM;
		goto fail5B;
	}
	memset((void *)pc_callout.table, 0,
	       sizeof(struct tty_struct) * pc_callout.num );
	if (tty_register_driver(&pc_callout) < 0 ) {
		epca_mess( KERN_ERR, "Couldn't register callout device");
		ret = -EPERM;
		goto fail6;
	}
#endif

	if (tty_register_driver(&pc_info) < 0 ) {
		epca_mess( KERN_ERR, "Couldn't register info device");
		ret = -EPERM;
		goto fail7;
	}

	/*
	 * loops_per_sec hasn't been set at this point :-(, so fake it out... 
	 * I set it, so that I can use the __delay() function.
	 *
	 * The wonderful people doing the Linux kernel decided that they
	 * wanted to change this declaration within a STABLE REV KERNEL.
	 * Isn't that nice.   we'll IFDEF around it.  now, anything 2.2.18pre
	 * and up will compile correctly.
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,17)
	save_loops = loops_per_sec;
	loops_per_sec = 13L * 500000L;
#else
	save_loops = loops_per_jiffy;
	loops_per_jiffy = (13L * 500000L) * HZ;
#endif

	save_flags(flags);
	cli();

	for (crd = 0; crd < num_cards; crd++) {
		/*
		 * This is where the appropriate memory handlers for the
		 * hardware is set.  Everything at runtime blindly jumps
		 * through these vectors.
		 */

		/* defined in epcaconfig.h */
		bd = &boards[crd];

		switch (bd->type) {
		case PCXR:
		case PCXEM:
			bd->memwinon     = pcxem_memwinon ;
			bd->memwinoff    = pcxem_memwinoff ;
			bd->globalwinon  = pcxem_globalwinon ;
			bd->txwinon      = pcxem_txwinon ;
			bd->rxwinon      = pcxem_rxwinon ;
			bd->memoff       = pcxem_memoff ;
			bd->assertgwinon = dummy_assertgwinon;
			bd->assertmemoff = dummy_assertmemoff;
			break;

		case PCICX:
		case PCICX_IBM:
		case PCIXEM:
		case PCIXRJ:
		case PCIXR:
		case PCI920_2:
		case PCI920_4:
		case PCI920_8:
		case PCIXR_422:
		case PCIXR_IBM:
		case PCIEPCJ:
			bd->memwinon     = dummy_memwinon;
			bd->memwinoff    = dummy_memwinoff;
			bd->globalwinon  = dummy_globalwinon;
			bd->txwinon      = dummy_txwinon;
			bd->rxwinon      = dummy_rxwinon;
			bd->memoff       = dummy_memoff;
			bd->assertgwinon = dummy_assertgwinon;
			bd->assertmemoff = dummy_assertmemoff;
			break;

		case ISACX:
		case PCXE:
			bd->memwinon     = pcxe_memwinon;
			bd->memwinoff    = pcxe_memwinoff;
			bd->globalwinon  = pcxe_globalwinon;
			bd->txwinon      = pcxe_txwinon;
			bd->rxwinon      = pcxe_rxwinon;
			bd->memoff       = pcxe_memoff;
			bd->assertgwinon = dummy_assertgwinon;
			bd->assertmemoff = dummy_assertmemoff;
			break;

		case PCXI:
			bd->memwinon     = pcxi_memwinon;
			bd->memwinoff    = pcxi_memwinoff;
			bd->globalwinon  = pcxi_globalwinon;
			bd->txwinon      = pcxi_txwinon;
			bd->rxwinon      = pcxi_rxwinon;
			bd->memoff       = pcxi_memoff;
			bd->assertgwinon = pcxi_assertgwinon;
			bd->assertmemoff = pcxi_assertmemoff;
			break;

		default:
			break;
		}


		/*
		 * Lets do *ANOTHER* switch, on bd->type, this time setting
		 * appropriate pagesizes and windowing stuff. CWS
		 */
		switch(bd->type) {
			
		case PCXE:
			/* 8k Boards */
			bd->winsize = 0x2000; /* 8k Page */
			bd->winbits = 13;     /* 11 bits */
			break;

		case PCXR:
		case PCXEM:
		case ISACX:
			/* 32k Boards */
			bd->winsize = 0x8000; /* 32k Page */
			bd->winbits = 15;     /* 15 bits  */
			break;
			
		case PCXI:
			/* 64k Boards */
			bd->winsize = 0x16000; 	/* 64k Page */
			bd->winbits = 16;	/* 16 bits  */
			break;
			
#ifdef ENABLE_PCI
		case PCICX:
		case PCICX_IBM:
		case PCIXEM:
		case PCIXRJ:
		case PCIXR:
		case PCI920_2:
		case PCI920_4:
		case PCI920_8:
		case PCIXR_422:
		case PCIXR_IBM:
		case PCIEPCJ:
			/* PCI boards, NO PAGE */
			bd->winsize = PCI_WIN_SIZE;	/* UNLIMITED PAGE */
			bd->winbits = PCI_WIN_SHIFT;
			break;
#endif /* ENABLE_PCI */

		}
				
		
		/*
		 * Some cards need a memory segment to be defined for use in 
		 * transmit and receive windowing operations.  These boards
		 * are listed in the below switch.  In the case of the XI the
		 * amount of memory on the board is variable so the memory_seg
		 * is also variable.  This code determines what they segment 
		 * should be.
		 */
		switch (bd->type) {

		case PCXE:
			bd->memory_seg = 0xf000;
			break;

		case PCXI:
			board_id = inb((int)bd->port);
			if ((board_id & 0x1) == 0x1) {

				/* Is it a 64K board */
				if ((board_id & 0x30) == 0) 
					bd->memory_seg = 0xf000;

				/* Is it a 128K board */
				if ((board_id & 0x30) == 0x10) 
					bd->memory_seg = 0xe000;

				/* Is is a 256K board */	
				if ((board_id & 0x30) == 0x20) 
					bd->memory_seg = 0xc000;

				/* Is it a 512K board */
				if ((board_id & 0x30) == 0x30) 
					bd->memory_seg = 0x8000;

			} else {
				epca_mess(KERN_ERR, "Board at 0x%x doesn't "
					  "appear to be an XI\n",
					  (int)bd->port);
			}
			break;

		}
	}

	restore_flags(flags);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,17)
	loops_per_sec = save_loops;  /* reset it to what it should be */
#else
	loops_per_jiffy = save_loops;
#endif

	/* Start up the poller to check for events on all enabled boards */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
	pc_timer.function = (void *)epcapoll;
	pc_timer.expires = jiffies + pollrate;
	pc_timeron = 1;
	add_timer(&pc_timer);
#else
	timer_table[DIGI_TIMER].fn = (void *)epcapoll;
	timer_table[DIGI_TIMER].expires = 0;
#endif

#ifdef CONFIG_PROC_FS

	/* Register /proc/epca */

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	if(!create_proc_info_entry("epca", 0, NULL, epca_read_proc)) {
		APR(("register /proc/epca failed\n"));
	}
# elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	/* proc_register() has dynamic allocation code in 2.2 kernels */
	proc_register( &proc_root, &epca_proc_entry);
# else
	proc_register_dynamic( &proc_root, &epca_proc_entry);
# endif
#endif /* CONFIG_PROC_FS */


	/*
	 * Initialize KME waitqueues...
	 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        for(itmp=0;itmp<num_cards;itmp++)
                init_waitqueue_head(&boards[itmp].kme_wait);
#else
	for(itmp=0;itmp<num_cards;itmp++)	
		boards[itmp].kme_wait = 0;
#endif
	
	return(0);


	/*
	 * On any sort of failure, de-allocate and unregister any previous
	 * previous pieces that succeeded.  Note: the calls for later fails
	 * intentionally falls through to the code for earlier ones.
	 * ("goto" has its uses!)
	 */
 fail7:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0) 
	if( tty_unregister_driver(&pc_callout) ) {
		epca_mess( KERN_ERR,
			   "pc_init: Cannot unregister callout device\n" );
	}
 fail6:
#endif
	bank = n_banks;	// This will be decremented at fail5
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	if( pc_callout.termios_locked )
		vfree( pc_callout.termios_locked );
 fail5B:
	if( pc_callout.table )
		vfree( pc_callout.table );
 fail5A:
	if( pc_callout.termios )
		vfree( pc_callout.termios );
#endif
 fail5:
	while( --bank >= 0 ) {
		if( tty_unregister_driver(pc_driver[bank]) ) {
			epca_mess( KERN_ERR, "pc_init: Cannot unregister "
				   "driver bank:%d\n", bank );
		}
	}
 fail4:
	for( bank=0; bank<n_banks; ++bank ) {
		if( pc_driver[bank] )
			kfree( pc_driver[bank] );
		if( pc_driver[bank]->name )
			kfree( (void*)(pc_driver[bank]->name) );
		if( pc_driver[bank]->termios )
			vfree( pc_driver[bank]->termios );
		if( pc_driver[bank]->termios_locked )
			vfree( pc_driver[bank]->termios_locked );
		if( pc_driver[bank]->table )
			vfree( pc_driver[bank]->table );
	}

 fail3:
	if( pc_driver )
		kfree( pc_driver );
 fail2:
	for( line = 0; line < nbdevs; ++line ) {
		if( digi_channels[line] )
			kfree( digi_channels[line] );
	}
 fail1:
	if( digi_channels )
		kfree( digi_channels );
	return( ret );
}


/*
 * card_conc()
 */
static void card_conc( int crd )
{
	struct board_info *bd;
	volatile int_16 addr;
	int_16 offset, bsize;
	volatile DL_CX *req;
	volatile DLH Concs_hdr;
	int i;

	volatile unchar *memaddr;

	char *which_conc = NULL;

	bd = &boards[crd];

	memaddr = (unsigned char *)epca_memaddr(crd);

	memwinon( bd, 0 );
	/* Is card ready to accept a concentrator download? */
	if( (addr = (*(int_16 *)(memaddr + DOWNREQ))) == 0 ) {
		memwinoff( bd, 0 );
		return;
	}

	/* reload concentrator */

	req = (DL_CX *)(memaddr + addr);
	/* Which concentrator image to use? */
	which_conc = NULL;
	for( i=0; i<MAX_CONC_IMAGES; i++ ) {
		if( conc_images[i] != NULL ) {
			memcpy( (DLH *) &Concs_hdr, conc_images[i], sizeof(DLH) );
			if( Concs_hdr.lrev <= req->hdr.lrev
			    && req->hdr.lrev <= Concs_hdr.hrev ) {
				which_conc = conc_images[i];
				break;	// Got a suitable image
			}
		}
	}

	bsize = 0;
	/* Figure out size to copy over. */

	if( which_conc == NULL ) {
		/* Bummer -- skip the download */
		epca_mess( KERN_INFO, "Failed to find suitable CX/EPCX "
			   "concentrator image\n" );
	} else if( ((offset = 1024 * req->hdr.seq) < Concs_hdr.size) ) {
		req->hdr.seg = (64 * req->hdr.seq) + Concs_hdr.seg;
		if ((bsize = (Concs_hdr.size - offset)) > 1024) 
			bsize = 1024;

		/* Move the concentrator image into req->data */
		memcpy ((void *)&(req->data),
			(void *)(which_conc + offset), bsize);
    
		if( (memcmp ((void *)&(req->data),
			     (void *)(which_conc + offset), bsize)) != 0) {
			epca_mess(KERN_INFO, "memcmp in card_conc failed\n");
		}

		req->hdr.size = bsize;
		
	} else {
		/* No more to download, mark Download complete. */
		req->hdr.seg = Concs_hdr.seg;
		req->hdr.size = 0;
		concinits[crd]=concinits[crd]++;
	}

	/* Done with this round, mark done. */
	*(u16 *)(memaddr + DOWNREQ) = 0;
	memwinoff( bd, 0 );
}


/*
 * epca_paranoia()
 *
 * This function exists for the sole purpose of sanity checking
 * It also *ASSUMES* that any memory it needs to be acting on
 * is ENABLED and paged properly, already.
 *
 * This serves BASIC SANITY CHECKING ONLY. Since the board
 * may or may not be running yet by the time this function
 * gets called, we can't check if it has a memory address
 * mapped yet or not... The below check at least weeds out
 * mishaps before they have a chance to blow up. CWS
 */
static int epca_paranoia(int crd, int function)
{

	struct board_info *bd = NULL;
	unchar *memaddr = NULL;
	
	if ((crd < 0) || (crd >= num_cards) || (num_cards == 0))
		return (-1);
	                                                        
	bd = &boards[crd];
	
	memaddr = (unsigned char *)epca_memaddr(crd);

	switch(function) {
	case CHECKFORFEP:
	{
		int offset;
			
		/* ISA C/X's put their FEP check at a different
		 * address than every other card digi makes.
		 * Hrm. Interesting.
		 */
			
		if(bd->type == ISACX) 
			offset = 0xc00;
		else
			offset = 0xd20;
			
		if( (*(char*)(memaddr+offset) == 'O') &&
		    (*(char*)(memaddr+offset+1) == 'S') )
			return(0);
		else
			return(-1);
	}
		
	case CHECKIFBDVALID:
		if(bd->type >= EPCA_NUM_TYPES)
			return(-1);
		/* If the description's null.. */
			
		if(btypes[bd->type].desc[0] == '\0')
			return(-1);
		
		if(!bd->numports) return(-1);		
		
		return(0);
		
	default:
		epca_mess(KERN_INFO, "Unimplemented epca_paranoia() "
			  "check %d\n",function);
		return(-1);
	}

	return(-1);
}
			
	
	

/*
 * post_fep_init()
 */
static void post_fep_init(unsigned int crd)
{

	int i, bank;
	unchar *memaddr;
	volatile struct global_data *gd;
	struct board_info *bd;
	volatile struct board_chan *bc;
	struct channel *ch; 
	int shrinkmem = 0, lowwater ; 
	ushort numports = 0;
	int abort=0;
	int first;

	/*
	 * This call is made by the user via. the ioctl call DIGI_INIT.
	 * It is resposible for setting up all the card specific stuff.
	 */
	bd = &boards[crd];
	bd->firstchan = firstchan(crd);
	bd->status = DISABLED;
	bd->dpastatus = BD_NOFEP;
	
	/*
	 * If this is a PCI board, get the port info.  Remember PCI cards
	 * do not have entries into the epcaconfig.h file, so we can't get 
	 * the number of ports from it.  Unfortunetly, this means that anyone
	 * doing a DIGI_GETINFO before the board has booted will get an
	 * invalid number of ports returned (It should return 0).  Calls to
	 * DIGI_GETINFO after DIGI_INIT has been called will return the
	 * proper values. 
	 */

	/* If the board in question is PCI */
	if ( (btypes[bd->type].dpatype & BUSTYPE) == T_PCIBUS) {

		/*
		 * If board reports I/O address and memory address of 0x0,
		 * we *KNOW* the board should NOT have post_fep_init ran
		 * on it.
		 */
		if(bd->membase == 0 && bd->port == 0) {
			epca_mess(KERN_INFO, "  %s "
				  "I/O = 0x%lx Mem = 0x%lx:Board not found!\n",
				  btypes[bd->type].desc,
				  (long)bd->port, (long)bd->membase );
#if 0
			epca_mess(KERN_INFO, "Board %d not found!\n");
#endif
			return;
		}
		
	}

	/* Set memaddr.. */
	
	memaddr = (unsigned char *)epca_memaddr(crd);

	memwinon(bd, 0);
		
	if(epca_paranoia(crd, CHECKFORFEP)) {
		bd->status = DISABLED;
		bd->dpastatus = BD_NOIOPORT;
		bd->initted = 0;
		return;
	}
	
	/* If we got past this point, we *KNOW* FEP/OS is RUNNING... */
	if(bd->type == PCXE)
		numports = *(char*)(memaddr+XEPORTS);
	else
		numports = (unsigned short)*(unsigned char *)
			(memaddr + XEMPORTS);
	
	
	memwinoff(bd, 0);
		
	/*
	 * If the board is enabled, but the number of ports configured
	 * doesn't match what FEP reports...
	 */
	
	if(numports != bd->numports)
		APR(("Warning: Board %d configured for %d ports, but "
                     "running board reports %d ports!\n", crd,
                     bd->numports, numports));

	/*
	 * If not enough memory has been allocated to handle all the ports on
	 * this adapter, ignore it altogether.  totalports can be specified
	 * at driver load time
	 */
	if( bd->firstchan + bd->numports > nbdevs ) {
		epca_mess( KERN_ERR,
			   "board #%d (ports %d-%d) disabled; configured "
			   "for %d ports total.\n", crd, bd->firstchan,
			   bd->firstchan+bd->numports, nbdevs );
		return;
	}
	
	if( numports <= 0 ) {
		/*
		 * If can't find ports, disable the card.  This could happen,
		 * for * example: on a PCI Xem if the external PORTS module is
		 * disconnected.
		 */
		sprintf(mesg, "   %s I/O = 0x%lx, "
			"Mem = 0x%lx: No ports found!\n", 
			btypes[bd->type].desc, (long)bd->port,
			(long)bd->membase );
		bd->status = DISABLED;
		bd->dpastatus = BD_NOFEP;
		bd->initted = 0;
	} else {
		/* Get the line offset for the first port on this card */
#if 0
		first = firstchan( crd );
#endif
		first = bd->firstchan;
		card_ptr[crd] = digi_channels[first];

		ch = card_ptr[crd];

		/*
		 * The below assignment will set bc to point at the BEGINING
		 * of the cards channel structures.  For 1 card there will
		 * be between 8 and 64 of these structures.
		 */
		bc = (volatile struct board_chan *)((ulong)memaddr+CHANSTRUCT);
	
		/*
		 * The below assignment will set gd to point at the BEGINING of
		 * global memory address 0xc00.  The first data in that global
		 * memory actually starts at address 0xc1a.  The command in 
		 * pointer begins at 0xd10.
		 */
		gd = (volatile struct global_data *)((ulong)memaddr + GLOBAL);
	
		/*
		 * XEPORTS (address 0xc22) points at the number of channels the
		 * card supports. (For 64XE, XI, XEM, and XR use 0xc02)
		 */
		memwinon(bd, 0);
	
		if ((bd->type == PCXE) &&
		    (*(ushort *)((ulong)memaddr + XEPORTS) < 3))
			shrinkmem = 1;
		if (bd->type < PCIXEM)
			request_region((int)bd->port,4, btypes[bd->type].desc);

		/* Get the value of MEMKB and store it in bd->memkb */
		memcpy(&bd->memkb, memaddr+MEMKB, 2);
		
		/*
		 * ch is the main driver's channels structure, bc is 
		 * the card's channel structure.
		 */

		abort=0;

		/* For every port on the card do ..... */
		for (i = 0; !abort && (i < bd->numports); i++, bc++) {
			ch = digi_channels[first+i];

			ch->channelindex = first+i;

			ch->brdchan        = bc;
			ch->mailbox        = gd; 
			ch->board          = &boards[crd];

			ch->doing_tpr = 0;	/* Not in transparent print mode yet */
			/* Initialize transparent print and other settings */
			ch->digiext = digi_init;

			/* It's possible for a process to abort without
			 * properly cleaning up after itself.  When running
			 * the driver as a module, this may makes it look
			 * like the driver is busy and prevent unloading.
			 * Setting the count to 0 on init should allow
			 * digiDload (or similar) to clear this condition.
			 */
			ch->tty_count = ch->tpr_count = 0;

			/*
			 * Different boards use different bitmaps
			 * for their control signals
			 */
			switch (bd->type) {
		
			case PCXR:
			case PCXEM:
			case PCIXEM:
			case PCIXRJ:
			case PCIXR:
			case PCI920_2:
			case PCI920_4:
			case PCI920_8:
			case PCIXR_422:
			case PCIXR_IBM:
			case ISACX:
			case PCICX:
			case PCICX_IBM:
			case PCIEPCJ:
				ch->m_rts = 0x02 ;
				ch->m_dcd = 0x80 ; 
				ch->m_dsr = 0x20 ;
				ch->m_cts = 0x10 ;
				ch->m_ri  = 0x40 ;
				ch->m_dtr = 0x01 ;
				break;
		
			case PCXE:
			case PCXI:
				ch->m_rts = 0x02 ;
				ch->m_dcd = 0x08 ; 
				ch->m_dsr = 0x10 ;
				ch->m_cts = 0x20 ;
				ch->m_ri  = 0x40 ;
				ch->m_dtr = 0x80 ;
				break;
			
			}
		
			if (boards[crd].altpin) {
				ch->dsr = ch->m_dcd;
				ch->dcd = ch->m_dsr;
				ch->digiext.digi_flags |= DIGI_ALTPIN;
			} else 	{ 
				ch->dcd = ch->m_dcd;
				ch->dsr = ch->m_dsr;
			}
			
			ch->boardnum   = crd;
			ch->channelnum = i;
			ch->magic      = EPCA_MAGIC;
			ch->tty        = 0;
		
			if (shrinkmem) {
				fepcmd(ch, SETBUFFER, 32, 0, 0, 0);
				shrinkmem = 0;
			}
		
			switch (bd->type) {
			case PCICX:
			case PCICX_IBM:
			case PCIXEM:
			case PCIXRJ:
			case PCIXR:
			case PCI920_2:
			case PCI920_4:
			case PCI920_8:
			case PCIXR_422:
			case PCIXR_IBM:
			case PCIEPCJ:
				/* Cover all the 2MEG cards */
				ch->txptr = memaddr +
					(((bc->tseg) << 4) & 0x1fffff);
				ch->rxptr = memaddr +
					(((bc->rseg) << 4) & 0x1fffff);
				ch->txwin = FEPWIN | ((bc->tseg) >> 11);
				ch->rxwin = FEPWIN | ((bc->rseg) >> 11);
				break;
			case ISACX:
			case PCXR:
			case PCXEM:
				/* Cover all the 32K windowed cards */
				/* Mask equal to window size - 1 */
				ch->txptr = memaddr + (((bc->tseg) << 4) &
						       0x7fff);
				ch->rxptr = memaddr + (((bc->rseg) << 4) &
						       0x7fff);
				ch->txwin = FEPWIN | ((bc->tseg) >> 11);
				ch->rxwin = FEPWIN | ((bc->rseg) >> 11);
				break;
			case PCXE:
				ch->txptr = memaddr + 
					(((bc->tseg - bd->memory_seg) << 4) &
					 0x1fff);
				ch->txwin = FEPWIN |
					((bc->tseg - bd->memory_seg) >> 9);
				ch->rxptr = memaddr +
					(((bc->rseg - bd->memory_seg) << 4) &
					 0x1fff);
				ch->rxwin = FEPWIN |
					((bc->rseg - bd->memory_seg) >>9 );
				break;
		
			case PCXI:
				ch->txptr = memaddr +
					((bc->tseg - bd->memory_seg) << 4);
				ch->rxptr = memaddr +
					((bc->rseg - bd->memory_seg) << 4);
				ch->txwin = ch->rxwin = 0;
				break;
		
			} /* End switch bd->type */

			ch->cps_time = 0;	/* printer buffer is empty */

		
			ch->txbufhead = 0;
			ch->txbufsize = bc->tmax + 1;
			
			ch->rxbufhead = 0;
			ch->rxbufsize = bc->rmax + 1;
		
			if(bd->type == PCI920_2 || bd->type == PCI920_4
			   || bd->type == PCI920_8 || bd->type == PCIXR_422
			   || bd->type == PCXR || bd->type == PCIXR
			   || bd->type == PCIXR_IBM) {
				lowwater = (ch->txbufsize *5) / 8;
			} else {
				lowwater = ch->txbufsize
					>= 2000 ? 1024 : (ch->txbufsize / 2);
			}
		
			/* Set transmitter low water mark */
			fepcmd(ch, STXLWATER, lowwater, 0, 10, 0);
		
			/* Set receiver low water mark */
			fepcmd(ch, SRXLWATER, (ch->rxbufsize / 4), 0, 10, 0);
		
			/* Set receiver high water mark */
			fepcmd(ch, SRXHWATER, (3 * ch->rxbufsize / 4), 0,10,0);
		
			bc->edelay = 100;
			bc->idata = 1;
			
			ch->startc  = bc->startc;
			ch->stopc   = bc->stopc;
			ch->startca = bc->startca;
			ch->stopca  = bc->stopca;
			
			ch->fepcflag = 0;
			ch->fepiflag = 0;
			ch->fepoflag = 0;
			ch->fepstartc = 0;
			ch->fepstopc = 0;
			ch->fepstartca = 0;
			ch->fepstopca = 0;
			
			ch->close_delay = 50;
			ch->tty_count = ch->tpr_count = 0;
			ch->blocked_open = 0;
			ch->callout_termios = pc_callout.init_termios;

			/*
			 *  For any adapter which supports the operation,
			 *  set the default communications mode to 232
			 */
			if(bd->type == PCIXR_IBM) {
				fepcmd(ch, SCOMMODE, MODE_232, 0, 0, 0);
			}

			/* Assuming all banks start the same... */
			bank = 0;
			ch->normal_termios = pc_driver[bank]->init_termios;

			/*
			 * Set RTS and DTR high by default (though the lines
			 * are physically set high only when the port is
			 * opened).
			 */
			ch->modem |= (ch->m_dtr | ch->m_rts);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
			init_waitqueue_head(&ch->open_wait);
               		init_waitqueue_head(&ch->close_wait);
#else
			ch->open_wait = 0;
			ch->close_wait = 0;
#endif

			/* Running digiDload multiple times could leak
			 * memory; free up old blocks before allocating new
			 */
			if( ch->tmp_buf ) {
				kfree( ch->tmp_buf );
			}
			ch->tmp_buf = kmalloc(ch->txbufsize,GFP_KERNEL);
			if (!(ch->tmp_buf)) {
				epca_mess(KERN_ERR,
					  "post_fep_init: kmalloc failed for "
					  "port 0x%x\n",i);
				abort = 1;	/* Forget the other ports! */
			} else
				memset((void *)ch->tmp_buf,0,ch->txbufsize);
		}
			
		memwinoff(bd, 0);
		if( abort ) {
			sprintf(mesg, "  %s I/O = 0x%lx, "
				"Mem = 0x%lx: kmalloc failed!  Disabling "
				"card\n", btypes[bd->type].desc,
				(long)bd->port, (long)bd->membase );
		
			bd->status = DISABLED;
			bd->dpastatus = BD_NOFEP;
			bd->initted = 0;

		} else {
			sprintf(mesg, "  %s I/O = 0x%lx,"
				" Mem = 0x%lx Ports = %d\n", 
				btypes[bd->type].desc, (long)bd->port,
				(long)bd->membase, bd->numports);
			
			bd->status = ENABLED;
			bd->dpastatus = BD_RUNNING;
			bd->initted = 1;
		}
	}
	epca_mess( KERN_INFO, "%s", mesg );
}


/*
 * epcapoll()
 */
static void epcapoll(unsigned long ignored)
{

	unsigned long flags;
	int crd;
	volatile unsigned int head, tail;
	struct channel *ch;
	struct board_info *bd;
	u16 addr;
	volatile unchar *memaddr;

	/*
	 * This routine is called upon every timer interrupt.  Even though
	 * the Digi series cards are capable of generating interupts this 
	 * method of non-looping polling is more efficient.  This routine
	 * checks for card generated events (Such as receive data, are transmit
	 * buffer empty) and acts on those events.
	 */
	
	save_flags(flags);
	cli();

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
	pc_timeron = 0;
	/* Not that it should matter with ints turned off */
	del_timer(&pc_timer); 
#endif


	for (crd = 0; crd < num_cards; crd++) {

		bd = &boards[crd];
		ch = card_ptr[crd];

		/*
		 * If this is a CX, check if it's asking for a concentator
		 * image download
		 */
		if( bd->initted == 1 && (bd->type == PCICX
					 || bd->type == PCICX_IBM
					 || bd->type == ISACX
					 || bd->type == PCIEPCJ)) {
			memaddr = (unsigned char *)epca_memaddr(crd);
			memwinon( bd, 0 );
			if( (addr = *(u16 *)(memaddr + DOWNREQ)) != 0 )
				card_conc( crd );
			memwinoff( bd, 0 );
		}

		if ((bd->status == DISABLED) || digi_poller_inhibited)
			continue; /* Begin loop next interation */

		/*
		 * assertmemoff is not needed here; indeed it is an empty
		 * subroutine. It is being kept because future boards may
		 * need this as well as some legacy boards.
		 */

		assertmemoff(ch);

		globalwinon(ch);

		/*
		 * In this case head and tail actually refer to the event
		 * queue not the transmit or receive queue.
		 */
		head = ch->mailbox->ein;
		tail = ch->mailbox->eout;
		
		/* If head isn't equal to tail we have an event */
		if (head != tail)
			doevent(crd);
	
		memoff(ch);

	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
	pc_timer.function = epcapoll;
	pc_timer.expires = jiffies + pollrate;
	pc_timeron = 1;
	add_timer(&pc_timer);
#else
	timer_table[DIGI_TIMER].fn = (void *)epcapoll;
	timer_table[DIGI_TIMER].expires = jiffies + pollrate;
	timer_active |= 1 << DIGI_TIMER;
#endif

	restore_flags(flags);

}


/*
 * doevent()
 *
 * This routine is called by epcapoll when an event is detected 
 * in the event queue.  This routine responds to those events.
 */
static void doevent(int crd)
{

	int i;
	volatile unchar *eventbuf;
	struct channel *ch, *chan0;
	static struct tty_struct *tty;
	volatile struct board_info *bd;
	volatile struct board_chan *bc;
	register volatile unsigned int tail, head;
	register int event, channel;
	register int mstat, lstat;
	volatile int offset;
	ulong  lstatusflags;


	bd = &boards[crd];

	chan0 = digi_channels[bd->firstchan];
	assertgwinon(chan0);

	while ((tail = chan0->mailbox->eout) != (head = chan0->mailbox->ein)) {

		assertgwinon(chan0);

		offset = tail + ISTART;
		
		eventbuf = (volatile unchar *)epca_memaddr(crd);
		eventbuf += offset;
		
		/* Get the channel the event occured on */
		channel = eventbuf[0];

		/* Get the actual event code that occured */
		event = eventbuf[1];


		/*
		 * The two assignments below get the current modem status
		 * (mstat) and the previous modem status (lstat).  These
		 * are useful becuase an event could signal a change in
		 * modem signals itself.
		 */
		mstat = eventbuf[2];
		lstat = eventbuf[3];

		ch = digi_channels[bd->firstchan+channel];

		if ((unsigned)channel >= bd->numports || !ch) { 
			if (channel >= bd->numports)
				ch = chan0;
			bc = ch->brdchan;
			goto next;
		}

		if ((bc = ch->brdchan) == NULL)
			goto next;

		if (event & DATA_IND) {
			receive_data(ch);
			assertgwinon(ch);
		}

		if (event & MODEMCHG_IND) {

			/* A modem signal change has been indicated */
			ch->imodem = mstat;

			if (ch->asyncflags & ASYNC_CHECK_CD) {
				if (mstat & ch->dcd) {
					/* We now see dcd */
					wake_up_interruptible(&ch->open_wait);
				} else if(!((ch->asyncflags & 
					     ASYNC_CALLOUT_ACTIVE) &&
					    (ch->asyncflags &
					     ASYNC_CALLOUT_NOHUP))) {
					if(ch->tty && ch->tty->driver_data) {
						tty_hangup ( ch->tty );
						ch->asyncflags &=
							~(ASYNC_CALLOUT_ACTIVE|
							  ASYNC_NORMAL_ACTIVE);
					}
					if(ch->tpr && ch->tpr->driver_data) {
						tty_hangup ( ch->tpr );
						ch->asyncflags &=
							~(ASYNC_CALLOUT_ACTIVE|
							  ASYNC_NORMAL_ACTIVE);
					}
				}
			}
		}

		/* Breaks only go to the tty port, not the printer port. */
		if (ch->tty && (event & BREAK_IND)) {
			/* A break has been indicated */
			ch->tty->flip.count++;
			*ch->tty->flip.flag_buf_ptr++ = TTY_BREAK;
			*ch->tty->flip.char_buf_ptr++ = 0;
			tty_schedule_flip(ch->tty); 
		}

		/* Wait until after processing both tpr and tty devices
		 * before setting ch->statusflags.
		 */
		lstatusflags = ch->statusflags;
		for( i=0; i<2; i++ ) {
 
			tty = (i == 0 ? ch->tty : ch->tpr);

			if( !tty )
				continue;

			if (event & LOWTX_IND) {
				if (ch->statusflags & LOWWAIT) {
					lstatusflags &= ~LOWWAIT;
					if ((tty->flags &
						 (1 << TTY_DO_WRITE_WAKEUP)) &&
						tty->ldisc.write_wakeup)
						{
						(tty->ldisc.write_wakeup)(tty);
						}
					wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
					wake_up_interruptible(&tty->poll_wait);
#endif
				}
			}

			if (event & EMPTYTX_IND) { 

				/* This event is gen'd by setup_empty_event */
				lstatusflags &= ~TXBUSY;
				if (ch->statusflags & EMPTYWAIT) {
					lstatusflags &= ~EMPTYWAIT;
					if ((tty->flags &
						 (1 << TTY_DO_WRITE_WAKEUP)) &&
						tty->ldisc.write_wakeup)
						{
						(tty->ldisc.write_wakeup)(tty);
						}
					wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
					wake_up_interruptible(&tty->poll_wait);
#endif
				}
			}
		}
		ch->statusflags = lstatusflags;

	next:
		globalwinon(ch);

		if (!bc)
			epca_mess(KERN_ERR, "bc == NULL in doevent!\n");
		else 
			bc->idata = 1;

		chan0->mailbox->eout = (tail + 4) & (IMAX - ISTART - 4);
		globalwinon(chan0);
	}
}


/*
 * fepcmd()
 *
 * This is the routine in which commands may be passed to the card.
 */
static void fepcmd(struct channel *ch, int cmd, int word_or_byte,
                   int byte2, int ncmds, int bytecmd)
{
	unchar *memaddr;
	unsigned int head, cmdTail, cmdStart, cmdMax;
	long count;
	int n;
	
#if 0
	if (ch->board->status == DISABLED) {
		return;
	}
#endif

	assertgwinon(ch);

	/* Remember head (As well as max) is just an offset not a base addr */
	head = ch->mailbox->cin;

	/* cmdStart is a base address */
	cmdStart = ch->mailbox->cstart;

	/*
	 * We do the addition below because we do not want a max pointer 
	 * relative to cmdStart.  We want a max pointer that points at the 
	 * physical end of the command queue.
	 */
	cmdMax = (cmdStart + 4 + (ch->mailbox->cmax));

	/* 
	 * The below command is necessary because newer kernels (2.1.x and
	 * up) do not have a 1:1 virtual to physical mapping.  The below
	 * call adjust for that.
	 */

	memaddr = (unsigned char *)epca_memaddr(ch->boardnum);

	if (head >= (cmdMax - cmdStart) || (head & 03)) {
		epca_mess(KERN_ERR, "fepcmd() - Command out of range!, "
			  "cmd=0x%x, head=0x%x, cmdMax=0x%x, cmdStart=0x%x\n",
			  cmd, head, cmdMax, cmdStart);
		return;
	}

	if (bytecmd) {
		*(volatile unchar *)(memaddr + head + cmdStart + 0) =
			(unchar)cmd;

		*(volatile unchar *)(memaddr + head + cmdStart + 1) =
			(unchar)ch->channelnum;

		/* Below word_or_byte is bits to set */
		*(volatile unchar *)(memaddr + head + cmdStart + 2) =
			(unchar)word_or_byte;

		/* Below byte2 is bits to reset */
		*(volatile unchar *)(memaddr + head + cmdStart + 3) =
			(unchar)byte2;
	} else {
		*(volatile unchar *)(memaddr + head + cmdStart + 0) =
			(unchar)cmd;
		*(volatile unchar *)(memaddr + head + cmdStart + 1) =
			(unchar)ch->channelnum;
		*(volatile ushort*)(memaddr + head + cmdStart + 2) =
			(ushort)word_or_byte;
	}

	head = (head + 4) & (cmdMax - cmdStart - 4);
	ch->mailbox->cin = head;

	count = FEPTIMEOUT;

	for (;;) {

		count--;
		if (count == 0) {
			epca_mess(KERN_ERR, "Fep not responding "
				  "in fepcmd()\n");
			return;
		}

		head = ch->mailbox->cin;
		cmdTail = ch->mailbox->cout;

		n = (head - cmdTail) & (cmdMax - cmdStart - 4);

		/*
		 * Basically this will break when the FEP acknowledges the 
		 * command by incrementing cmdTail (Making it equal to head).
		 */
		if (n <= ncmds * (sizeof(short) * 4))
			break;
	}
}


/*********************************************************************
 * Digi products use fields in their channels structures that are very
 * similar to the c_cflag and c_iflag fields typically found in UNIX
 * termios structures.  The below three routines allow mappings 
 * between these hardware "flags" and their respective Linux flags.
 */

/*
 * termios2digi_h()
 */
static unsigned termios2digi_h(struct channel *ch, unsigned cflag)
{

	unsigned res = 0;

	if (cflag & CRTSCTS) {
		res |= ( (ch->m_cts) | (ch->m_rts));
	} else {
		if(ch->digiext.digi_flags & CTSPACE)
			res |= ch->m_cts;
		if(ch->digiext.digi_flags & RTSPACE)
			res |= ch->m_rts;
	}

	if (ch->digiext.digi_flags & DTRPACE)
		res |= ch->m_dtr;

	if (ch->digiext.digi_flags & DSRPACE)
		res |= ch->dsr;

	if (ch->digiext.digi_flags & DCDPACE)
		res |= ch->dcd;

	return res;

}


/*
 * termios2digi_i()
 */
static unsigned termios2digi_i(struct channel *ch, unsigned iflag)
{

	unsigned res = iflag & (IGNBRK | BRKINT | IGNPAR | PARMRK | 
	                        INPCK | ISTRIP|IXON|IXANY|IXOFF);
	
	if (ch->digiext.digi_flags & DIGI_AIXON)
		res |= IAIXON;
	if (ch->digiext.digi_flags & DIGI_ALTPIN)
		res |= IALTPIN;
	return res;

}


/*
 * termios2digi_c()
 */
static unsigned termios2digi_c(struct channel *ch, unsigned cflag)
{

	unsigned res = 0;
	unsigned tcflag;
	unsigned baudpart;

	if (cflag & CBAUDEX || (ch->digiext.digi_flags & DIGI_FAST) ) {
		/*
		 * HUPCL bit is used by FEP to indicate fast baud
		 * table is to be used.
		 */
		res |= FEP_HUPCL;
	} else {
		ch->digiext.digi_flags &= ~DIGI_FAST; 
	}
		
	/*
	 * CBAUD has bit position 0x1000 set these days to indicate Linux
	 * baud rate remap.  Digi hardware can't handle the bit assignment.
	 * (We use a different bit assignment for high speed.).  Clear this
	 * bit out while grabbing the parts of "cflag" we want.
	 */
	res |= cflag & ((CBAUD ^ CBAUDEX) | PARODD | PARENB | CSTOPB | CSIZE);



	/*
	 * This gets a little confusing.  The Digi cards have their own
	 * representation of c_cflags controling baud rate.  For the most
	 * part this is identical to the Linux implementation.  However;
	 * Digi supports one rate (76800) that Linux doesn't.  This means 
	 * that the c_cflag entry that would normally mean 76800 for Digi
	 * actually means 115200 under Linux.  Without the below mapping,
	 * a stty 115200 would only drive the board at 76800.  Since 
	 * the rate 230400 is also found after 76800, the same problem
	 * affects us when we choose a rate of 230400.  Without the below
	 * modificiation stty 230400 would actually give us 115200.
	 *
	 * There are two additional differences.  The Linux value for CLOCAL
	 * (0x800; 0004000) has no meaning to the Digi hardware.  Also in 
	 * later releases of Linux; the CBAUD define has CBAUDEX (0x1000;
	 * 0010000) ored into it (CBAUD = 0x100f as opposed to 0xf). CBAUDEX
	 * should be checked for and screened out prior to termios2digi_c 
	 * returning.  Since CLOCAL isn't used by the board this can be
	 * ignored as long as the returned value is used only by Digi hardware.
	 *
	 *
	 * Note: there are two ways of accessing the extended baud rates:
	 * 	1. Using stty or other normal Linux calls to set baud rates to
	 *	   the predefined extended baud rate values (from termbits.h):
	 *			#define  B57600  0010001
	 *			#define  B115200 0010002
	 *			#define  B230400 0010003
	 *			#define  B460800 0010004
	 *
	 *	These are the values that need remapping to match the digi
	 *      firmware table
	 *
	 *	2. Using "ditty fastbaud" to set the DIGI_FAST bit, then
	 *	   directly use one of the lower baud rates.  E.g: 
	 *	   "ditty fastbaud 110" sets DIGI_FAST and invokes a
	 *	   cflag of B110 (0000003).  This method does not
	 *         require (or want) the remapping.
	 */

	/*
	 * Don't even consider baudrate remapping if fastbaud is set.
	 * CBAUDEX is set for rates above 38400
	 */
	if ((cflag & CBAUDEX) && !(ch->digiext.digi_flags & DIGI_FAST) ) {
		/*
		 * The below code is trying to guarantee that only baud rates
		 * 115200, 230400, 460800 are remapped.  We use exclusive or
		 * because the various baud rates share common bit positions
		 * and therefore can't be tested for easily.
		 */
		tcflag = (cflag & CBAUD) | CBAUDEX;
		baudpart = 0;

		/* Map high speed requests to index into FEP's baud table */
		switch( tcflag ) {
#ifdef B57600
		case B57600 :
			baudpart = 1;
			break;
#endif
#ifdef B76800
		case B76800 :
			baudpart = 2;
			break;
#endif
#ifdef B115200
		case B115200 :
			baudpart = 3;
			break;
#endif
#ifdef B230400
		case B230400 :
			
			baudpart = 9;
			break;
#endif
#ifdef B460800
		case B460800 :
			baudpart = 11;
			break;
#endif
#ifdef B921600
		case B921600 :
			baudpart = 12;
			break;
#endif
		default:
			epca_mess( KERN_ERR, "termios2digi_c (bd:%d:ch%d), "
				   "bad baudrate request:%xx\n",
				   ch->boardnum, ch->channelnum, tcflag );
		}

		if( baudpart ) {
			/* Clear old baud bits, and reset them to baudpart */
			res = (res & ~(CBAUD | CBAUDEX)) | baudpart;
		}
	}
	return res;
}


/*
 * termios2digi_misc()
 *
 */
static unsigned termios2digi_misc(struct channel *ch)
{
	
	/* If we are turning *ON* printer. CWS */
	if(ch->digiext.digi_flags & DIGI_PRINTER) {
	
		if(ch->printer) {
			/* we're already a printer.. CWS */
			return(0);
		}
	
		ch->printer = 1;
		MOD_INC_USE_COUNT;
		return(0);
	} else {
		/* Okay. we must not want printer. CWS */
		
		if(!ch->printer) {
			/* its already off...No need to do anything...CWS */
			return(0);
		}
		
		/* Must be on..Lets go ahead and turn it off...CWS */
		ch->printer = 0;
        MOD_DEC_USE_COUNT;
	}
	
	return(0);
}		


/*
 * epcaparam()
 */
static void epcaparam(struct tty_struct *tty, struct channel *ch )
{

	unsigned long flags;
	unsigned int cmdHead;
	struct termios *ts;
	volatile struct board_chan *bc;
	unsigned hflow, cflag, iflag;
	
	bc = ch->brdchan;
	epcaassert(bc !=0, "bc out of range");

	assertgwinon(ch);

	ts = tty->termios;

	if ((ts->c_cflag & CBAUD) == 0) {
		cmdHead = bc->rin;
		bc->rout = cmdHead;
		cmdHead = bc->tin;

		/*
		 * Flush current transmit buffer by setting cmdTail pointer
		 * (tout) to cmdHead pointer (tin).  Hopefully the transmit
		 * buffer is empty.
		 */

		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, STOUT, (unsigned) cmdHead, 0, 0, 0);
		memoff(ch);
		restore_flags(flags);
		
		/* 
		 * Now, also toggle DTR down, wait a second, and toggle it
		 * back up but ONLY do it, if DTR was PREVIOUSLY HIGH. (We
		 * don't want to falsely lower it and raise it if the user
		 * didn't have it high in the first place!!
		 */
		if(ch->omodem & ch->m_dtr) {
			save_flags(flags);
			cli();
			globalwinon(ch);
			ch->omodem &= ~ch->m_dtr;
			fepcmd(ch, SETMODEM, 0, ch->m_dtr, 10, 1);
			memoff(ch);
			restore_flags(flags);
			
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ);   /* Delay 1 second */
			current->state = TASK_RUNNING;
			
			save_flags(flags);
			cli();
			globalwinon(ch);
			ch->omodem |= ch->m_dtr;
			fepcmd(ch, SETMODEM, ch->m_dtr, 0, 10, 1);
			memoff(ch);
			restore_flags(flags);
		}
	} else {
		/* Propagate the change to the card. */

		cflag = termios2digi_c(ch, ts->c_cflag);

		if (cflag != ch->fepcflag) {
			ch->fepcflag = cflag;
			/* Set baud rate, char size, stop bits, parity */
			save_flags(flags);
			cli();
			globalwinon(ch);
			fepcmd(ch, SETCTRLFLAGS, (unsigned) cflag, 0, 0, 0);
			memoff(ch);
			restore_flags(flags);
		}

		/*
		 * If the user has not forced CLOCAL and if the device is not
		 * a CALLOUT device (Which is always CLOCAL) we set flags
		 * such that the driver will wait on carrier detect.
		 */
		if ((ts->c_cflag & CLOCAL) ||
		    (tty->driver.subtype == SERIAL_TYPE_CALLOUT)) {
			ch->asyncflags &= ~ASYNC_CHECK_CD;
		} else {
			ch->asyncflags |= ASYNC_CHECK_CD;
		}
	}

	iflag = termios2digi_i(ch, ts->c_iflag);

	/* Check input mode flags */
	if (iflag != ch->fepiflag) {
		ch->fepiflag = iflag;
		/*
		 * Command sets channels iflag structure on the board. Such
		 * things as input soft flow control, handling of parity
		 * errors, and break handling are all set here.
		 */
		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, SETIFLAGS, (unsigned int) ch->fepiflag, 0, 0, 0);
		memoff(ch);
		restore_flags(flags);
	}

	/*
	 * Set the board mint value for this channel.  This will cause hardware
	 * events to be generated each time the DCD signal (Described in mint) 
	 * changes.	
	 */
	bc->mint = ch->dcd;

	if ((ts->c_cflag & CLOCAL) || (ch->digiext.digi_flags & DIGI_FORCEDCD))
		bc->mint = 0;

	ch->imodem = bc->mstat;

	hflow = termios2digi_h(ch, ts->c_cflag);

	if (hflow != ch->hflow) {
		ch->hflow = hflow;
		/*
		 * Hard flow control has been selected but the board is not
		 * using it.  Activate hard flow control now.
		 */
		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, SETHFLOW, hflow, 0xff, 0, 1);
		memoff(ch);
		restore_flags(flags);
	}
	

	/*
	 * This gets a little complex here. This states:
	 * If the signals we want (modem) don't match whats already set
	 * (omodem) - then go ahead and make the requested changes and
	 * save them back into what we THINK is set - so that they
	 * match now. CWS
	 */
	if( (ch->omodem & ch->m_dtr) != (ch->modem & ch->m_dtr) ||
	    (ch->omodem & ch->m_rts) != (ch->modem & ch->m_rts) ) {
		ch->omodem = ch->modem;
		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, SETMODEM, 0, ((ch->m_dtr)|(ch->m_rts)), 0, 1);
		fepcmd(ch, SETMODEM, ch->modem, 0, 0, 1);
		memoff(ch);
		restore_flags(flags);
	}

	if (ch->startc != ch->fepstartc || ch->stopc != ch->fepstopc) {
		ch->fepstartc = ch->startc;
		ch->fepstopc = ch->stopc;
		/*
		 * The XON / XOFF characters have changed; propogate these
		 * changes to the card.	
		 */
		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, SONOFFC, ch->fepstartc, ch->fepstopc, 0, 1);
		memoff(ch);
		restore_flags(flags);
	}

	if (ch->startca != ch->fepstartca || ch->stopca != ch->fepstopca) {
		ch->fepstartca = ch->startca;
		ch->fepstopca = ch->stopca;

		/*
		 * Similar to the above, this time the auxilarly XON / XOFF 
		 * characters have changed; propogate these changes to the
		 * card.
		 */
		save_flags(flags);
		cli();
		globalwinon(ch);
		fepcmd(ch, SAUXONOFFC, ch->fepstartca, ch->fepstopca, 0, 1);
		memoff(ch);
		restore_flags(flags);
	}

	/* Deal with ditty printer - CWS */
	termios2digi_misc(ch);
}


/*
 * receive_data()
 *
 * Called by doevent when a receive data event has taken place.
 */
static void receive_data(struct channel *ch)
{

	unchar *rptr;
	struct termios *ts = 0;
	struct tty_struct *tty;
	volatile struct board_chan *bc;
	register int dataToRead, wrapgap, bytesAvailable;
	register unsigned int tail, head;
	unsigned int wrapmask;
	int rc;


	globalwinon(ch);

	if (ch->statusflags & RXSTOPPED) 
		return;

	/* All received data goes only to the tty port, not the printer port. */
	tty = ch->tty;
	if (tty)
		ts = tty->termios;

	bc = ch->brdchan;

	if (!bc) {
		epca_mess(KERN_ERR, "bc is NULL in receive_data!\n");
		return;
	}

	wrapmask = ch->rxbufsize - 1;

	/*
	 * Get the head and tail pointers to the receiver queue.  Wrap the 
	 * head pointer if it has reached the end of the buffer.
	 */
	head = bc->rin;
	head &= wrapmask;
	tail = bc->rout & wrapmask;

	bytesAvailable = (head - tail) & wrapmask;

	if (bytesAvailable == 0)
		return;

	/* If CREAD bit is off or device not open, set TX tail to head */
	if (!tty || !ts || !(ts->c_cflag & CREAD)) {
		bc->rout = head;
		return;
	}

	if (tty->flip.count == TTY_FLIPBUF_SIZE) 
		return;

	if (bc->orun) {
		bc->orun = 0;
		epca_mess(KERN_WARNING, "overrun! DigiBoard device "
			"major = %d, minor = %d\n", MAJOR(tty->device), MINOR(tty->device));
	}

	rxwinon(ch);
	rptr = tty->flip.char_buf_ptr;
	rc = tty->flip.count;

	while (bytesAvailable > 0) {
		wrapgap = (head >= tail) ? head - tail : ch->rxbufsize - tail;

		/*
		 * Even if head has wrapped around only report the amount of
		 * data to be equal to the size - tail.  Remember memcpy can't
		 * automaticly wrap around the receive buffer.
		 */
		dataToRead = ((wrapgap < bytesAvailable) ?
			      wrapgap : bytesAvailable);

		/* Make sure we don't overflow the buffer */
		if ((rc + dataToRead) > TTY_FLIPBUF_SIZE)
			dataToRead = TTY_FLIPBUF_SIZE - rc;

		if (dataToRead == 0)
			break;

		/*
		 * Move data read from our card into the line disciplines
		 * buffer for translation if necessary.
		 */
		if ((memcpy(rptr, ch->rxptr + tail, dataToRead)) != rptr)
			epca_mess(KERN_ERR, "receive_data : memcpy failed\n");
			
		rc   += dataToRead;

		/*
		 * Need to set flag to TTY_NORMAL for all these bytes,
		 * otherwise prior conditions like TTY_BREAK persist.
		 */
		memset(tty->flip.flag_buf_ptr, TTY_NORMAL, dataToRead);
		tty->flip.flag_buf_ptr += dataToRead;

		rptr += dataToRead;
		tail = (tail + dataToRead) & wrapmask;
		bytesAvailable -= dataToRead;
	}

	tty->flip.count = rc;
	tty->flip.char_buf_ptr = rptr;
	globalwinon(ch);
	bc->rout = tail;

	/* Must be called with global data */
	tty_schedule_flip(ch->tty); 
	return;
}


/*
 * flush_rx_buffer()
 */
static void flush_rx_buffer(struct channel *ch)
{

	/*
	 * Make the rout pointer equal rin
	 * NO DATA ON BOARD NOW! GUARATEED! CWS
	 */
	ch->brdchan->rout = ch->brdchan->rin;
}


/*
 * flush_tx_buffer()
 */
static void flush_tx_buffer(struct channel *ch)
{

	/*
	 * Make the tout pointer equal tin
	 * NO DATA IN BOARD TX QUEUE! GUARANTEED!
	 */
	ch->brdchan->tout = ch->brdchan->tin;
}
                                                

/*
 * pc_ioctl()
 */
static int pc_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	digiflow_t dflow;
	int retval, error, print_buf_cnt;
	unsigned long flags;
	unsigned int mflag, mstat;
	unsigned char startc, stopc;
	volatile struct board_chan *bc;
	struct channel *ch = (struct channel *) tty->driver_data;
	unsigned int conc_size = 0;
	static char *conc_pt = NULL; 

#if 0
# define IOCTL_DEBUG
#endif

#ifdef IOCTL_DEBUG
	epca_mess( KERN_INFO, "pc_ioctl top\n" );
#endif
	/* The control device has it's own set of commands */
	if (tty->driver.subtype == SERIAL_TYPE_INFO) {

		switch (cmd) {

		case DIGI_DISABLE:
		{ 
			int brd;
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_DISABLE\n" );
#endif
			getUser(brd, (unsigned int *)arg);

			if ((brd < 0) || (brd >= num_cards) ||(num_cards == 0))
				return (-ENODEV);
				
			boards[brd].status = DISABLED;
			boards[brd].dpastatus = BD_NOFEP;
			boards[brd].memkb = 0;
			boards[brd].initted = 0;
			break;
		}

#ifdef MODULE
		case DIGI_MODDEC:
			epca_mess(KERN_INFO, "pc_ioctl DIGI_MODDEC - "
				  "Decreasing Module Usage Count By 1\n");
			MOD_DEC_USE_COUNT;
			break;
#endif


		/* FIXME -- seperate function  JAR 010412 */
		case DIGI_KME:
		{
			int itmp, jtmp;
			unchar *memaddr = NULL;
			rw_t kme;
			rw_t *mp = NULL;
			int brd = 0;
			struct board_info *bd;

			/* This ioctl takes an argument of type 'rw_t'
			 * and uses it to interact with the KME struct
			 * located on the digiboard itself.
			 */
#ifdef IOCTL_DEBUG
			epca_mess(KERN_INFO, "pc_ioctl DIGI_KME - "
				  "0x%x\n",DIGI_KME);
#endif
			if ((error = verify_area(VERIFY_READ,
						 (char*)arg, sizeof(kme))))
				return(error);
				
				
			copy_from_user(&kme, (char*)arg, sizeof(kme));
				
			if(kme.rw_size > 128)
				kme.rw_size = 128;
			
			brd = kme.rw_board;
			
#if IOCTL_DEBUG
			epca_mess(KERN_INFO, "DIGI_KME: %s asked for "
				  "board %d\n", current->comm, brd);
#endif
				
			/* Sanity Checking... */
				
			if ((brd < 0) || (brd >= num_cards) ||(num_cards == 0))
				return (-ENODEV);
				
			bd = &boards[brd];
				
			if(bd->status == DISABLED || bd->initted == 0) {
#if 0
				boards[brd].dpatype = 0;
#endif
				return(-ENODEV);
			}
				
			/* Calculate the base address of the card.. */
			if (bd->membase < 0x100000)
				memaddr = (unchar *)bd->membase;
			else 
				memaddr = bd->re_map_membase;
				
			/* If it needs remapping, do so... */
			memaddr = (unsigned char *)epca_memaddr(brd);
				
			/* If the concentrator number is 0... */
			if(kme.rw_conc == 0 && kme.rw_addr < 0x100000) {
				int page = 0;
				int addr = kme.rw_addr;
				int size = kme.rw_size;
				caddr_t data = (caddr_t) kme.rw_data;
					
				/* Turn on card memory.. */
					
				while((itmp = size)) {

					/* Find out the page we need */
						
					page += addr >> bd->winbits;
					addr &= (bd->winsize - 1);
						
					if((addr + itmp) > bd->winsize)
						itmp = bd->winsize - addr;
							
					/* Turn it on! */
					memwinon(bd, page);

					switch(kme.rw_req) {
					case RW_READ:
					{
						register caddr_t cp1 = (char *)memaddr + addr;
						register caddr_t cp2 = kme.rw_data;
						
#ifdef KME_IOCTL_DEBUG
						epca_mess(KERN_INFO, "RW_READ CARDMEM - page=%d rw_addr=0x%x  rw_size=%d\n", 
							  page, kme.rw_addr, kme.rw_size);
#endif
						
						for(jtmp = 0;jtmp < itmp; jtmp++)
							*cp2++ = *cp1++;
					}
					
					break;
					
					case RW_WRITE:
					{
						register caddr_t cp1 = memaddr + addr;
						register caddr_t cp2 = data;
						
#ifdef KME_IOCTL_DEBUG						
						epca_mess(KERN_INFO, "RW_WRITE CARDMEM - page=%d rw_addr=0x%x  rw_size=%d\n",
							  page, kme.rw_addr, kme.rw_size);
#endif						
						
						for (jtmp = 0; jtmp < itmp; jtmp++)
							*cp1++ = *cp2++;
					}																
					break;
					}	
					
					addr += itmp;
					data += itmp;
					size -= itmp;
					
				}
					
				memwinoff(bd, page);
			} else {
				
				/*
				 * Read/Write memory in a REMOTE CONCENTRATOR..
				 * There is only 1 buffer, so do mutual
				 * exclusion to make sure only one KME
				 * request is pending...
				 */
					
				mp = (rw_t *) (memaddr + KMEMEM);
					
				while(kmebusy != 0) {
					kmebusy = 2;
					interruptible_sleep_on(&bd->kme_wait);
				}
					
				kmebusy = 1;
					
				/* Assert card's memory.. */
					
				memwinon(bd, 0);
					
				/* Copy KME request to the board.. */
					
				mp->rw_board = kme.rw_board;
				mp->rw_conc  = kme.rw_conc;
				mp->rw_reserved = kme.rw_reserved;
				memcpy(&mp->rw_addr, &kme.rw_addr, sizeof(int));
				memcpy(&mp->rw_size, &kme.rw_size, sizeof(short));
					
				if(kme.rw_req == RW_WRITE) {
					register caddr_t cp1 = (caddr_t) mp->rw_data;
					register caddr_t cp2 = (caddr_t) kme.rw_data;
						
#ifdef KME_IOCTL_DEBUG
					epca_mess(KERN_INFO, "RW_WRITE CONCMEM - rw_addr=0x%x  rw_size=%d\n",
						  kme.rw_addr, kme.rw_size);
#endif			
					
					for (jtmp = 0; jtmp < (int) kme.rw_size; jtmp++)
						*cp1++ = *cp2++;
				}
			
					
				/* EXECUTE REQUEST */
					
				mp->rw_req = kme.rw_req;
						
				/*
				 * Wait for the board to process the
				 * request, but limit the wait to 2 secs
				 */
					 
				for(itmp = jiffies + (2 * HZ); mp->rw_req;) {
					if(jiffies >= itmp) {
						error = ENXIO;
						/* Set request back to 0.. */
						mp->rw_req = 0;
						goto endkme;
					}
					
					schedule_timeout(HZ/10);

				}
					 
				/*
				 * Since this portion of code is looksee
				 * ported from the HPUX EMUX code, i'm
				 * leaving OUT a portion of that code where
				 * the HP/UX code actually puts the process
				 * to sleep for some reason
				 */
				if(mp->rw_size < kme.rw_size) 
					memcpy(&kme.rw_size, &mp->rw_size,
					       sizeof(short));
						
				/* Copy the READ data back to the source buffer... */
					
				if(kme.rw_req == RW_READ) {
					register caddr_t cp1 = (caddr_t) mp->rw_data;
					register caddr_t cp2 = (caddr_t) kme.rw_data;
	
#ifdef KME_IOCTL_DEBUG
					epca_mess(KERN_INFO, "RW_READ CONCMEM - rw_addr=0x%x  rw_size=%d\n",
						  kme.rw_addr, kme.rw_size);
#endif			
	
					for (jtmp = 0; jtmp < (int) kme.rw_size; jtmp++)
						*cp2++ = *cp1++;
				}
					
					
				/*
				 * Common exit point for code sharing the
				 * kme buffer. Before exiting, always wake
				 * another process waiting for the buffer
				 */
					      
			endkme:
			                
				memwinoff(bd, 0);
				if (kmebusy != 1)
			                wake_up_interruptible(&bd->kme_wait);
				kmebusy = 0;
				if(error == ENXIO) return(-EINVAL);
			}
			
			/* Copy the whole (Possibly Modified) mess */
			/* back out to user space...		   */
				
			if(!error) { 
				copy_to_user((char *)arg, &kme, sizeof(kme));
				return(0);
			}
		} /* FIXME -- seperate function  JAR 010412  END */			

		case DIGI_GETDD:
		{
			/* This returns the total number of boards
			 * in the system, as well as driver version
			 * and has space for a reserved entry
			 */
			
			struct digi_dinfo ddi;

#ifdef IOCTL_DEBUG				
			epca_mess(KERN_INFO, "pc_ioctl DIGI_GETDD - "
				  "ioctl 0x%x\n", DIGI_GETDD);
#endif				
			ddi.dinfo_nboards = num_cards;

#if 0
			sprintf(ddi.dinfo_version, "%s", DG_PART);
#else
                        /* 14 is the cvs offset that is the version number */
                        strncpy(ddi.dinfo_version, (epca_version + 14), 16);
#endif
								
			copy_to_user((char *)arg, &ddi, sizeof (ddi));
			break;
		}
			
		case DIGI_GETBD:
		{
			/* Similar to DIGI_GETINFO - Except we're remaining
			 * compliant with DPA, in this one.
			 */

			int brd;
			struct digi_info di;
#ifdef IOCTL_DEBUG
			epca_mess(KERN_INFO, "pc_ioctl DIGI_GETBD - "
				  "ioctl 0x%x\n", DIGI_GETBD);
#endif
			getUser(brd, (unsigned int *)arg);
			if((error = verify_area(VERIFY_WRITE,
						(char*)arg, sizeof(di)))) {
				epca_mess(KERN_ERR, "DIGI_GETBD : verify area"
					  " size 0x%x failed\n",sizeof(di));
				return(error);
			}

			if ((brd < 0) || (brd >= num_cards) ||(num_cards == 0))
				return (-ENODEV);
                                
#ifdef IOCTL_DEBUG
			epca_mess(KERN_INFO, "pc_ioctl DIGI_GETBD - %s "
				  "asking for board %d\n",current->comm, brd);
#endif
                                      	
			memset(&di, 0, sizeof(di));
				
			di.info_bdnum = brd;
			di.info_bdstate = boards[brd].dpastatus;
			di.info_bdtype = btypes[boards[brd].type].dpatype;
			di.info_nports = boards[brd].numports;
			di.info_ioport = boards[brd].port;
			di.info_physaddr = boards[brd].membase;
			di.info_physsize = (boards[brd].memkb * 1024);
				
			di.firstchan = boards[brd].firstchan;
				
			copy_to_user((char *)arg, &di, sizeof (di));
			break;
		}

		case DIGI_GETINFO:
		{

			struct digi_info di ;
			int brd;
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_GETINFO\n" );
#endif

			getUser(brd, (unsigned int *)arg);

			if ((error = verify_area(VERIFY_WRITE,
						 (char*)arg, sizeof(di)))) {
				epca_mess(KERN_ERR, "DIGI_GETINFO : verify "
					  "area size 0x%x failed\n",
					  sizeof(di));
				return(error);
			}

			if ((brd < 0) || (brd >= num_cards) ||(num_cards == 0))
				return (-ENODEV);

			memset(&di, 0, sizeof(di));

			di.info_bdnum = brd ; 
			di.info_bdstate = boards[brd].status;
			di.info_bdtype = boards[brd].type ;
			di.info_nports = boards[brd].numports ;
			di.info_ioport = boards[brd].port ;
			di.info_physaddr = boards[brd].membase ;
			di.firstchan = boards[brd].firstchan ;

			copy_to_user((char *)arg, &di, sizeof (di));
			break;

		}

		case DIGI_POLLER:
		{
			int brd = arg & 0xff000000 >> 16 ; 
			unsigned char state = arg & 0xff ; 
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_POLLER\n" );
#endif

			if ((brd < 0) || (brd >= num_cards)) {
				epca_mess(KERN_ERR, "DIGI POLLER : brd "
					  "not valid!\n");
				return (-ENODEV);
			}

			digi_poller_inhibited = state ;
			break ; 

		}

		case DIGI_INIT:
		{
			/*
			 * This call is made by the apps to complete the
			 * initialization of the board(s).  This routine
			 * is responsible for setting the card to its
			 * initial state and setting the driver's control
			 * fields to the suitable settings for the card
			 * in question.
			 */
			
			int crd ; 
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_INIT\n" );
#endif
			for (crd = 0; crd < num_cards; crd++) 
				post_fep_init (crd);

			break ; 

		}

		/* User passes us a bank number, we overwrite it with
		 * the major number for that bank of minor numbers.
		 */
		case DIGI_GETMAJOR:
		case DIGI_GETPRMAJOR:
		{
			int majbank;
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_GETMAJOR/DIGI_PRGETMAJOR\n" );
#endif
			if ((error =
			     verify_area(VERIFY_READ, (char*)arg,
					 sizeof(majbank))))
				return(error);

			copy_from_user(&majbank, (char*)arg, sizeof(majbank));
			if( majbank < 0 ||  majbank > n_banks ) {
				epca_mess(KERN_ERR, "DIGI GETMAJOR/GETPRMAJOR : invalid "
					  "major bank number:%d!\n",
					  majbank );
				return (-ENODEV);
			}
					
			if( cmd == (DIGI_GETMAJOR) )
				copy_to_user((char *)arg, &majors[majbank], sizeof (majbank));
			else
				copy_to_user((char *)arg, &majorspr[majbank], sizeof (majbank));
			break;
		}


		/* Prepare for getting a copy of the C/X concentrator binary */
		case DIGI_CXCON_INIT:
		case DIGI_CXCON_NEXT:
		case DIGI_EPCXCON_INIT:
		case DIGI_EPCXCON_NEXT:
		{
			char *iocnam;

			switch( cmd ) {
			case DIGI_CXCON_INIT: iocnam = "DIGI_CXCON_INIT"; break;
			case DIGI_CXCON_NEXT: iocnam = "DIGI_CXCON_NEXT"; break;
			case DIGI_EPCXCON_INIT: iocnam = "DIGI_EPCXCON_INIT"; break;
			case DIGI_EPCXCON_NEXT: iocnam = "DIGI_EPCXCON_NEXT"; break;
			default: iocnam = "Unknown (FATAL ERROR)"; break;
			}
			epca_mess( KERN_INFO, "obsolete ioctl %s called\n", iocnam );
			break;
		}

		case DIGI_CON_INIT:
		{
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_CON_INIT\n" );
#endif
			if ((error = verify_area(VERIFY_READ, (char*)arg,
						 sizeof(conc_size))))
				return(error);

			if( num_conc_images >= MAX_CONC_IMAGES ) {
				error = -EINVAL;
				return(error);
			}

			copy_from_user(&conc_size, (char*)arg, sizeof(conc_size));

			/*
			 *  By convention, an initialization size of _0_ indicates
			 *  that we should reset by throwing away all current images.
			 */
			if( !conc_size ) {
				while( num_conc_images )
				{
					char *tmp_image;

					--num_conc_images;
					tmp_image = conc_images[num_conc_images];
					conc_images[num_conc_images] = NULL;
					kfree( tmp_image );
				}
				conc_pt = NULL;
			} else {
				conc_pt = kmalloc(conc_size,GFP_KERNEL);
				
				if (!conc_pt) {
					epca_mess(KERN_ERR,
						  "digi_con_init: kmalloc failed "
						  "for concentrator download "
						  "image\n");
					return (-ENODEV);
				}

				memset((void *)conc_pt,0,conc_size);
				conc_images[num_conc_images] = conc_pt;
				num_conc_images++;
			}
			break;
		}

		case DIGI_CON_NEXT:
		{
			CONC_UPLOAD conc_chunk;
#ifdef IOCTL_DEBUG
			epca_mess( KERN_INFO, "pc_ioctl DIGI_CON_NEXT\n" );
#endif
			if ((error = verify_area(VERIFY_READ, (char*)arg,
						 sizeof(CONC_UPLOAD))))
				return(error);

			copy_from_user(&conc_chunk, (char*)arg,
				       sizeof(conc_chunk));
			if( conc_pt != NULL ) {
				memcpy( conc_pt, conc_chunk.data,
					conc_chunk.size );
				conc_pt += conc_chunk.size;
			} else
				epca_mess( KERN_ERR, "digi_con_next: "
					   "NULL pointer\n" );
			break;
		}

		default:
			return -ENOIOCTLCMD;

		}
		return (0) ;
	}

	if (ch) {
		bc = ch->brdchan;
	} else {
		epca_mess(KERN_ERR, "ch is NULL in pc_ioctl!\n");
		return(-EINVAL);
	}
#ifdef IOCTL_DEBUG
	epca_mess( KERN_INFO, "pc_ioctl non-control bd#:%d, ch:%d\n",
		   ch->boardnum, ch->channelnum );
#endif

	save_flags(flags);

	switch (cmd) {

	case DIGI_SPOLL:
	{
		/*
		 * Sets the poller rate implicitly, rather than
		 * using the driver's usual polling interval (20ms)
		 *
		 * If arg is less than 1, its zero or lower, which is
		 * invalid. If its at 100, thats 1000ms (1s) - at
		 * least on an Intel
		 */
		if(!arg)
			return(pollrate);
					
		if(arg < 1 || arg > (HZ/100) * 100) {
			epca_mess(KERN_CRIT, "DIGI_SPOLL: Invalid "
				  "PollRate - Thrown Out\n");
			return(-1);
		}
			
		pollrate = arg;
		return(pollrate);
	}
		

	case TCGETS:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCGETS\n" );
#endif
		retval = verify_area(VERIFY_WRITE, (void *)arg,
				     sizeof(struct termios));
			
		if (retval)
			return(retval);

		copy_to_user((struct termios *)arg, 
			     tty->termios, sizeof(struct termios));
		return(0);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0)
	case TCGETA:
# ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCGETA\n" );
# endif
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
                if(tty->ldisc.ioctl) {
                        int retval;
                        retval = (tty->ldisc.ioctl) (tty, file, cmd, arg);
                        return(retval);
                }
                return(-ENOIOCTLCMD);
# else
		return get_termio(tty, (struct termio *)arg);
# endif
#endif

	case TCSBRK:	/* SVID version: non-zero arg --> no break */
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCSBRK\n" );
#endif

		retval = tty_check_change(tty);
		if (retval)
			return retval;

		/*
		 * Setup an event to indicate when the transmit 
		 * buffer empties
		 */
		save_flags(flags);
		cli();
		globalwinon(ch);
		setup_empty_event(tty,ch);
		memoff(ch);
		restore_flags(flags);
		tty_wait_until_sent(tty, 0);
		if (!arg)
			digi_send_break(ch, 250 );    /* 250 millisecond */
		return 0;

	case TCSBRKP:	/* support for POSIX tcsendbreak() */
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCSBRKP\n" );
#endif

		retval = tty_check_change(tty);
		if (retval)
			return retval;
		
		/* Setup event to indicate when the transmit buffer empties */

#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCSBRKP bd:%d/ch:%d calling "
			   "setup_empty event\n", ch->boardnum,ch->channelnum);
#endif
		setup_empty_event(tty,ch);		
		tty_wait_until_sent(tty, 0);

		/* Default is 250 msec */
		digi_send_break(ch, arg ? arg*(HZ/10) : 250);
		return 0;


	case TIOCGSOFTCAR:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCGSOFTCAR\n" );
#endif
		error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));
		if (error)
			return error;

		putUser(C_CLOCAL(tty) ? 1 : 0,
			(unsigned long *) arg);
		return 0;


	case TIOCSSOFTCAR:
	{
		unsigned int value;
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCSSOFTCAR\n" );
#endif
		getUser(value, (unsigned int *)arg);
		tty->termios->c_cflag =
			((tty->termios->c_cflag & ~CLOCAL) |
			 (value ? CLOCAL : 0));
		return 0;
	}


	case TIOCMODG:
	case TIOCMGET:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCMODG/TIOCMGET\n" );
#endif
		mflag = 0;
			
		/*
		 * We always wait a default of 2ms to at least give
		 * FEP/OS a chance to keep up, if an input signal
		 * transition... And then, if there's been a
		 * timeout set as well, add that to the total wait
		 * time, before actually displaying the signal states
		 */
		cli();
		globalwinon(ch);

		mstat = bc->mstat;
			
		memoff(ch);
		restore_flags(flags);
			
		if (ch->omodem & ch->m_dtr)
			mflag |= TIOCM_DTR;

		if (ch->omodem & ch->m_rts)
			mflag |= TIOCM_RTS;

		if (mstat & ch->m_cts)
			mflag |= TIOCM_CTS;

		if (mstat & ch->dsr)
			mflag |= TIOCM_DSR;

		if (mstat & ch->m_ri)
			mflag |= TIOCM_RI;

		if (mstat & ch->dcd)
			mflag |= TIOCM_CD;

		error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));

		if (error)
			return error;

		putUser(mflag, (unsigned long *) arg);

		break;

	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMODS:
	case TIOCMSET:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCMBIS/TIOCMBIC/"
			   "TIOCMODS/TIOCMSET\n" );
#endif

		getUser(mstat, (unsigned int *)arg);

		mflag = 0;
		if (mstat & TIOCM_DTR)
			mflag |= ch->m_dtr;

		if (mstat & TIOCM_RTS)
			mflag |= ch->m_rts;

		switch (cmd) {

		case TIOCMODS:
		case TIOCMSET:
			ch->modem = mflag;
			break;

		case TIOCMBIS:
			ch->modem |= mflag;
			break;

		case TIOCMBIC:
			ch->modem &= ~mflag;
			break;

		}

		cli();
		globalwinon(ch);
		epcaparam(tty,ch);
		memoff(ch);
		restore_flags(flags);
		break;


	case TIOCSDTR:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCSDTR\n" );
#endif
		ch->omodem |= ch->m_dtr;
		cli();
		globalwinon(ch);
		fepcmd(ch, SETMODEM, ch->m_dtr, 0, 10, 1);
		memoff(ch);
		restore_flags(flags);
		break;


	case TIOCCDTR:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TIOCCDTR\n" );
#endif
		ch->omodem &= ~ch->m_dtr;
		cli();
		globalwinon(ch);
		fepcmd(ch, SETMODEM, 0, ch->m_dtr, 10, 1);
		memoff(ch);
		restore_flags(flags);
		break;


	case DIGI_GETA:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl DIGI_GETA\n" );
#endif
		if ((error = verify_area(VERIFY_WRITE, (char*)arg,
					 sizeof(digi_t)))) {
			epca_mess(KERN_ERR, "Digi GETA failed\n");
			return(error);
		}

		copy_to_user((char*)arg, &ch->digiext, sizeof(digi_t));
		break;


	case TCSETSW:
	case TCSETAW:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl TCSETSW/TCSETAW\n" );
#endif
		/*
		 * Here we just setup an event to indicate when
		 * the transmit buffer empties.  The real work is
		 * done by n_tty_ioctl() when it calls set_termios()
		 * and hence tty_wait_until_sent().  The latter needs
		 * a low watermark event to wake it up from deep
		 * slumber -- otherwise ports appear to hang until
		 * another event is triggered on that channel.
		 */
		setup_empty_event(tty,ch);		
		return -ENOIOCTLCMD;


	case DIGI_SETAW:
	case DIGI_SETAF:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl DIGI_SETAW/AF\n" );
#endif
		if ((cmd) == (DIGI_SETAW)) {
			/* event to indicate when transmit buffer empties */
			setup_empty_event(tty,ch);		
			tty_wait_until_sent(tty, 0);
		} else {
			if (tty->ldisc.flush_buffer)
				tty->ldisc.flush_buffer(tty);
		}
		/* Fall Thru */


	case DIGI_SETA:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl DIGI_SETA\n" );
#endif
		if ((error = verify_area(VERIFY_READ, (char*)arg,
					 sizeof(digi_t))))
			return(error);

		/* calc number of characters in print buffer */
		print_buf_cnt = (ch->digiext.digi_maxcps*(ch->cps_time-jiffies))/HZ;

		copy_from_user(&ch->digiext, (char*)arg, sizeof(digi_t));

		/* be sure we have sane values */
		if (ch->digiext.digi_maxcps > 10000)
			ch->digiext.digi_maxcps = 10000;
		if (ch->digiext.digi_maxchar > ch->digiext.digi_bufsize)
			ch->digiext.digi_maxchar = ch->digiext.digi_bufsize;
		if (ch->digiext.digi_onlen > DIGI_PLEN)
			ch->digiext.digi_onlen = DIGI_PLEN;
		if (ch->digiext.digi_offlen > DIGI_PLEN)
			ch->digiext.digi_offlen = DIGI_PLEN;

		/* reset ch->cps_time based on new values */
		if( ch->digiext.digi_maxcps > 0 && ch->digiext.digi_bufsize > 0 ) {
			if( print_buf_cnt > ch->digiext.digi_bufsize )
				print_buf_cnt = ch->digiext.digi_bufsize;
			ch->cps_time = jiffies + (HZ*print_buf_cnt)/ch->digiext.digi_maxcps;
		}
		else {
			ch->cps_time = 0;
		}

		cli();
		globalwinon(ch);

		if (ch->digiext.digi_flags & DIGI_ALTPIN) {
			/* Swaps DCD, DSR on concentrator */
			ch->fepiflag |= IALTPIN;
			ch->dcd = ch->m_dsr;
			ch->dsr = ch->m_dcd;
		} else {
			/* Do not swap DCD, DSR on concentrator */
			ch->fepiflag &= ~IALTPIN;
			ch->dcd = ch->m_dcd;
			ch->dsr = ch->m_dsr;
		}
		fepcmd(ch, SETIFLAGS, (unsigned int) ch->fepiflag, 0, 0, 0);
		

		/*
		 * The below routine generally sets up parity, baud, flow
		 * control issues, etc.... It effect both control flags
		 * and input flags.
		 */

		/*
		 * The third argument signals to epcaparam that this is a
		 * Digi-extended call
		 */
		epcaparam(tty,ch);
		memoff(ch);
		restore_flags(flags);
		break;


	case DIGI_GETFLOW:
	case DIGI_GETAFLOW:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl DIGI_GETFLOW/GETAFLOW\n" );
#endif
		cli();	
		globalwinon(ch);
		if ((cmd) == (DIGI_GETFLOW)) {
			dflow.startc = bc->startc;
			dflow.stopc = bc->stopc;
		} else {
			dflow.startc = bc->startca;
			dflow.stopc = bc->stopca;
		}
		memoff(ch);
		restore_flags(flags);

		if ((error = verify_area(VERIFY_WRITE, (char*)arg,
					 sizeof(dflow))))
			return(error);

		copy_to_user((char*)arg, &dflow, sizeof(dflow));
		break;


	case DIGI_SETAFLOW:
	case DIGI_SETFLOW:
#ifdef IOCTL_DEBUG
		epca_mess( KERN_INFO, "pc_ioctl DIGI_SETFLOW/SETAFLOW\n" );
#endif
		if ((cmd) == (DIGI_SETFLOW)) {
			startc = ch->startc;
			stopc = ch->stopc;
		} else {
			startc = ch->startca;
			stopc = ch->stopca;
		}

		if ((error = verify_area(VERIFY_READ, (char*)arg,
					 sizeof(dflow))))
			return(error);

		copy_from_user(&dflow, (char*)arg, sizeof(dflow));

		if (dflow.startc != startc || dflow.stopc != stopc) {
			cli();
			globalwinon(ch);

			if ((cmd) == (DIGI_SETFLOW)) {
				ch->fepstartc = ch->startc = dflow.startc;
				ch->fepstopc = ch->stopc = dflow.stopc;
				fepcmd(ch, SONOFFC, ch->fepstartc,
				       ch->fepstopc, 0, 1);
			} else {
				ch->fepstartca = ch->startca = dflow.startc;
				ch->fepstopca  = ch->stopca = dflow.stopc;
				fepcmd(ch, SAUXONOFFC, ch->fepstartca,
				       ch->fepstopca, 0, 1);
			}

			if(ch->statusflags & TXSTOPPED)
				pc_start(tty);

			memoff(ch);
			restore_flags(flags);

		}
		break;
		

	case TCFLSH:
		switch(arg) {
		case TCIFLUSH:
			flush_rx_buffer(ch);
			break;
				
		case TCIOFLUSH:
			flush_rx_buffer(ch);
			/* fall through */
					
		case TCOFLUSH:
			/*
			 * Alright. Flush the TX queue on the board...
			 * We'll break after we do this, so the n_tty_ioctl()
			 * code can take care of flushing out the tty buffers
			 */
			flush_tx_buffer(ch);
			break;
				
		default:
			return(-EINVAL);
		}				                 				
			
		/* Okay. We've done OUR part. Now, lets let the
		 *  TTY subsystem deal with the rest... CWS
		 */
		return(-ENOIOCTLCMD);
		break;
			

		/*
		 * This ioctl call should be executed, and then passed off
		 * to the n_tty_ioctl() subsystem, by returning -ENOIOCTLCMD
		 */
	case TCXONC:
		switch(arg) {
		case TCOOFF:
			/* ditty stopout */
				
			/* Stop TX *NOW*! */
			save_flags(flags);
			cli();
			globalwinon(ch);
			fepcmd(ch, PAUSETX, 0, 0, 0, 0);
			memoff(ch);
			restore_flags(flags);
			return(0);
			break;
			 
		case TCOON:
			/* ditty startout */
			save_flags(flags);
			cli();
			globalwinon(ch);

			/* Let TX resume.. */
			fepcmd(ch, RESUMETX, 0, 0, 0, 0);
			 		
			memoff(ch);
			restore_flags(flags);
			return(0);
			break;
			 	
		case TCIOFF:
			/* ditty stopin */
			break;
			 	
		case TCION:
			/* ditty startin */
			break;
			 	
		default: return(-EINVAL);
		}
		
		return(-ENOIOCTLCMD);
			 	 	
			
		/*
		 * These are implemented via n_tty_ioctl() and its calls into
		 * our other entry points -- no need to take action here.
		 */
	case TCSETA:
	case TCSETS:
	case TCSETSF:
	case FIONREAD:
		return -ENOIOCTLCMD;
	default:
#if 0
		epca_mess( KERN_INFO, "pc_ioctl IOCTL not implemented: "
			   "%xx\n", cmd );
#endif
		return -ENOIOCTLCMD;

	}

	return 0;

}



/*
 * pc_set_termios()
 */
static void pc_set_termios(struct tty_struct *tty, struct termios *old_termios)
{

	struct channel *ch;
	unsigned long flags;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */

	if ((ch = verifyChannel(tty)) != NULL) {

		save_flags(flags);
		cli();
		globalwinon(ch);
		epcaparam(tty, ch);
		memoff(ch);

		if ((old_termios->c_cflag & CRTSCTS) &&
		    ((tty->termios->c_cflag & CRTSCTS) == 0))
			tty->hw_stopped = 0;

		if (!(old_termios->c_cflag & CLOCAL) &&
		    (tty->termios->c_cflag & CLOCAL))
			wake_up_interruptible(&ch->open_wait);

		restore_flags(flags);

	}
}


/***************************************************************************
 * pc_stop and pc_start provide software flow control to the 
 * routine and the pc_ioctl routine.
 */

/*
 * pc_stop()
 */
static void pc_stop(struct tty_struct *tty)
{

	struct channel *ch;
	unsigned long flags;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) != NULL) {

		save_flags(flags); 
		cli();

		if ((ch->statusflags & TXSTOPPED) == 0) {

			globalwinon(ch);

			/* STOP transmitting now !! */
			fepcmd(ch, PAUSETX, 0, 0, 0, 0);

			ch->statusflags |= TXSTOPPED;
			memoff(ch);
		}

		restore_flags(flags);
	}
}


/*
 * pc_start()
 */
static void pc_start(struct tty_struct *tty)
{

	struct channel *ch;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) != NULL) {

		unsigned long flags;

		save_flags(flags);
		cli();

		/*
		 * Just in case output was resumed because of a change
		 * in Digi-flow
		 */
		if (ch->statusflags & TXSTOPPED) {

			volatile struct board_chan *bc;

			globalwinon(ch);
			bc = ch->brdchan;
			if (ch->statusflags & LOWWAIT)
				bc->ilow = 1;

			/* Okay, you can start transmitting again... */
			fepcmd(ch, RESUMETX, 0, 0, 0, 0);

			ch->statusflags &= ~TXSTOPPED;
			memoff(ch);
		}
		restore_flags(flags);
	}
}


/*******************************************************************
 * The below routines pc_throttle and pc_unthrottle are used 
 * to slow (And resume) the receipt of data into the kernels
 * receive buffers.  The exact occurence of this depends on the
 * size of the kernels receive buffer and what the 'watermarks'
 * are set to for that buffer.  See the n_ttys.c file for more
 * details. 
 */

/*
 * pc_throttle()
 */
static void pc_throttle(struct tty_struct * tty)
{

	struct channel *ch;
	unsigned long flags;

	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) != NULL) {

		save_flags(flags);
		cli();

		if ((ch->statusflags & RXSTOPPED) == 0)	{
			globalwinon(ch);

#if 0 /* Let the FEP handle this on its own */
			fepcmd(ch, PAUSERX, 0, 0, 0, 0);
#endif
			ch->statusflags |= RXSTOPPED;
			memoff(ch);
		}
		restore_flags(flags);
	}
}


/*
 * pc_unthrottle()
 */
static void pc_unthrottle(struct tty_struct *tty)
{

	struct channel *ch;
	unsigned long flags;
	volatile struct board_chan *bc;


	/*
	 * verifyChannel returns the channel from the tty struct
	 * if it is valid.  This serves as a sanity check.
	 */
	if ((ch = verifyChannel(tty)) != NULL) {

		/*
		 * Just in case output was resumed because of a change
		 * in Digi-flow
		 */
		save_flags(flags);
		cli();

		if (ch->statusflags & RXSTOPPED) {

			globalwinon(ch);
			bc = ch->brdchan;

#if 0 /* Let the FEP handle this on its own */
			fepcmd(ch, RESUMERX, 0, 0, 0, 0);
#endif

			ch->statusflags &= ~RXSTOPPED;
			memoff(ch);
		}
		restore_flags(flags);
	}
}


/*
 * digi_send_break()
 */
void digi_send_break(struct channel *ch, int msec)
{

	unsigned long flags;
	save_flags(flags);
	cli();
	globalwinon(ch);

	/*
	 * Maybe I should send an infinite break here, schedule() for
	 * msec amount of time, and then stop the break.  This way,
	 * the user can't screw up the FEP by causing digi_send_break()
	 * to be called (i.e. via an ioctl()) more than once in msec amount 
	 * of time.  Try this for now...
	 */

	/*
	 * msec is number of milleseconds requested, but FEPOS measures
	 * time in 10 msec TICS, (page 47 of FEP/OS 5.X Interface
	 * Specification), hence divide msec by 10
	 */
	fepcmd(ch, SENDBREAK, msec/10, 0, 10, 0);
	memoff(ch);

	restore_flags(flags);

}


/*
 * setup_empty_event()
 */
static void setup_empty_event(struct tty_struct *tty, struct channel *ch)
{

	volatile struct board_chan *bc = ch->brdchan;

#if 0
	unsigned long int flags;

	save_flags(flags);
	cli();
#endif

	globalwinon(ch);
	ch->statusflags |= EMPTYWAIT;
	
	/*
	 * When set the iempty flag request a event to be generated when the 
	 * transmit buffer is empty (If there is no BREAK in progress).
	 */
	bc->iempty = 1;
	memoff(ch);

#if 0
	restore_flags(flags);
#endif

}


/*
 * maxcps_room
 *
 * Reduces bytes_available to the max number of characters
 * that can be sent currently given the maxcps value, and
 * returns the new bytes_available.  This only affects printer
 * output.
 */
static int maxcps_room(struct tty_struct *tty, struct channel *ch,
	int bytes_available)
{

	int cps_limit;
	unsigned long current_time;
	unsigned long buffer_time;


	/*
	 * Limit printer output to maxcps overall, with bursts allowed
	 * up to bufsize characters.
	 */
	if( istpr(tty->device)
	&& ch->digiext.digi_maxcps > 0 && ch->digiext.digi_bufsize > 0 ) {

		current_time = jiffies;
		buffer_time = current_time
				+ (HZ*ch->digiext.digi_bufsize)/ch->digiext.digi_maxcps;

		if( ch->cps_time < current_time ) {
			/* buffer is empty */
			ch->cps_time = current_time;		/* reset cps_time */
			cps_limit = ch->digiext.digi_bufsize;
		}
		else if( ch->cps_time < buffer_time ) {
			/* still room in the buffer */
			cps_limit
				= ((buffer_time-ch->cps_time)*ch->digiext.digi_maxcps)/HZ;
		}
		else {
			/* no room in the buffer */
			cps_limit = 0;
		}

		bytes_available = MIN(cps_limit, bytes_available);

	}

	return( bytes_available );

}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,0,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
/*
 * get_termio()
 */
static int get_termio(struct tty_struct * tty, struct termio * termio)
{

	int error;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	struct termio temp_termio;
#endif

	error = verify_area(VERIFY_WRITE, termio, sizeof (struct termio));
	if (error)
		return error;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	trans_to_termio(tty->termios, &temp_termio);
	copy_to_user(termio, &temp_termio, sizeof (struct termio));
#else
	kernel_termios_to_user_termio(termio, tty->termios);
#endif

	return 0;
}
#endif


/*
 * epca_setup()
 */
void epca_setup(char *str, int *ints)
{

	struct	 board_info board;
	int      index, loop, last;
	char     *temp, *t2;
	unsigned len;

	/*
	 * If this routine looks a little strange it is because it is
	 * only called if a LILO append command is given to boot the
	 * kernel with parameters.  In this way, we can provide the
	 * user a method of changing his board configuration without
	 * rebuilding the kernel.
	 */
	if (!liloconfig) 
		liloconfig = 1; 

	memset(&board, 0, sizeof(board));

	/* Assume the data is int first, later we can change it */
	/* I think that array position 0 of ints holds the number of args */
	for (last = 0, index = 1; index <= ints[0]; index++) {

		switch(index) {

		case 1:
			board.status = ints[index];
				
			/*
			 * We check for 2 (As opposed to 1; because 2 is a
			 * flag instructing the driver to ignore epcaconfig.)
			 * For this reason we check for 2.
			 */

			if (board.status == 2) {
				/*
				 * Begin ignore epcaconfig as well as lilo
				 * cmd line
				 */
				nbdevs = 0;
				num_cards = 0;
				return;
			}
	
			if (board.status > 2)
			{
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "board status 0x%x\n", board.status);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_BOARD_STATUS;
				return;
			}
			last = index;
			break;

		case 2:
			board.type = ints[index];
			if (board.type >= PCIXEM) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "board type 0x%x\n", board.type);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_BOARD_TYPE;
				return;
			}
			last = index;
			break;

		case 3:
			board.altpin = ints[index];
			if (board.altpin > 1) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "board altpin 0x%x\n", board.altpin);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_ALTPIN;
				return;
			}
			last = index;
			break;

		case 4:
			board.numports = ints[index];
			if ((board.numports < 2) || (board.numports > 256)) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "board numports 0x%x\n",
					  board.numports);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_NUM_PORTS;
				return;
			}
			nbdevs += board.numports;
			last = index;
			break;

		case 5:
			board.port = ints[index];
			if (board.port <= 0) {
				epca_mess(KERN_ERR, "epca_setup: Invalid io "
					  "port 0x%x\n",
					  (unsigned int)board.port);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_PORT_BASE;
				return;
			}
			last = index;
			break;

		case 6:
			board.membase = ints[index];
			if (board.membase <= 0)	{
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "memory base 0x%x\n",
					  (unsigned int)board.membase);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_MEM_BASE;
				return;
			}
			last = index;
			break;

		default:
			epca_mess(KERN_ERR, "epca_setup: Too many integer "
				  "parms\n");
			return;

		}
	}

	while (str && *str) {

		/* find the next comma or terminator */
		temp = str;

		/* While string is not null, and a comma hasn't been found */
		while (*temp && (*temp != ','))
			temp++;

		if (!*temp)
			temp = NULL;
		else
			*temp++ = 0;

		/* Set index to the number of args + 1 */
		index = last + 1;

		switch(index) {
		case 1:
			len = strlen(str);
			if (strncmp("Override", str, len) == 0) {
				return;
			} else {
				if (strncmp("Disable", str, len) == 0) {
					board.status = DISABLED;
					board.dpastatus = BD_NOFEP;
				} else if (strncmp("Enable", str, len) == 0) {
					board.status = ENABLED;
					board.dpastatus = BD_NOFEP;
				} else {
					epca_mess(KERN_ERR, "epca_setup: "
						  "Invalid status %s\n", str);
					invalid_lilo_config = 1;
					setup_error_code|=INVALID_BOARD_STATUS;
					return;
				}
			}
			last = index;
			break;

		case 2:

			for(loop = 0; loop < EPCA_NUM_TYPES; loop++)
				if (!strcmp(btypes[loop].desc, str))
					break;


			/*
			 * If the index incremented above refers to a
			 * legitamate board type set it here. 
			 */
			if (index < EPCA_NUM_TYPES) {
				board.type = loop;
			} else {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "board type: %s\n", str);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_BOARD_TYPE;
				return;
			}
			last = index;
			break;

		case 3:
			len = strlen(str);
			if (strncmp("Disable", str, len) == 0) {
				board.altpin = 0;
			} else {
				if (strncmp("Enable", str, len) == 0) {
					board.altpin = 1;
				} else {
					epca_mess(KERN_ERR, "epca_setup: "
						  "Invalid altpin %s\n", str);
					invalid_lilo_config = 1;
					setup_error_code |= INVALID_ALTPIN;
					return;
				}
			}
			last = index;
			break;

		case 4:
			t2 = str;
			while (isdigit(*t2))
				t2++;

			if (*t2) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "port count %s\n", str);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_NUM_PORTS;
				return;
			}

			/*
			 * There is not a man page for simple_strtoul but
			 * the code can be found in vprintk.c.  The first
			 * argument is the string to translate (To an
			 * unsigned long obviously),  the second argument
			 * can be the address of any character variable or
			 * a NULL.  If a variable is given, the end pointer
			 * of the string will be stored in that variable; if
			 * a NULL is given the the end pointer will not be
			 * returned.  The last argument is the base to use.
			 * If a 0 is indicated, the routine will attempt to
			 * determine the proper base by looking at the values
			 * prefix (A '0' for octal, a 'x' for hex, etc ... 
			 * If a value is given it will use that value as the
			 * base. 
			 */
			board.numports = simple_strtoul(str, NULL, 0);
			nbdevs += board.numports;
			last = index;
			break;

		case 5:
			t2 = str;
			while (isxdigit(*t2))
				t2++;

			if (*t2) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "i/o address %s\n", str);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_PORT_BASE;
				return;
			}

			board.port = simple_strtoul(str, NULL, 16);
			last = index;
			break;

		case 6:
			t2 = str;
			while (isxdigit(*t2))
				t2++;

			if (*t2) {
				epca_mess(KERN_ERR, "epca_setup: Invalid "
					  "memory base %s\n",str);
				invalid_lilo_config = 1;
				setup_error_code |= INVALID_MEM_BASE;
				return;
			}

			board.membase = simple_strtoul(str, NULL, 16);
			last = index;
			break;

		default:
			epca_mess(KERN_ERR, "epca_setup: Too many "
				  "string parms\n");
			return;
		}
		str = temp;

	}


	if (last < 6) {
		epca_mess(KERN_ERR, "epca_setup: Insufficient "
			  "parms specified\n");
		return;
	}
 
	/* I should REALLY validate the stuff here */

	/* Copies our local copy of board into boards */
	memcpy((void *)&boards[num_cards],(void *)&board, sizeof(board));

	epca_mess(KERN_INFO, "epca_setup: Added board %i, %s %i "
		  "ports at 0x%4.4X base 0x%6.6X\n", 
		  num_cards, btypes[board.type].desc, board.numports,
		  (int)board.port, (unsigned int) board.membase);

	num_cards++;
}


#ifdef ENABLE_PCI

/*
 * get_PCI_configuration()
 */
int get_PCI_configuration( unsigned char bus, unsigned char device_fn,
			   u32 *base_addr0, u32 *base_addr1,
			   u32 *base_addr2, u32 *base_addr3,
			   u32 *base_addr4, u32 *base_addr5)
{

	int	error;
	
	error = pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_0,
	                                  base_addr0);
	error |= pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_1,
					   base_addr1);
	error |= pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_2,
					   base_addr2);
	error |= pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_3,
					   base_addr3);
	error |= pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_4,
					   base_addr4);
	error |= pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_5,
					   base_addr5);

	/*
	 * NOTE - The code below mask out either the 2 or 4 bits dependent
	 * on the space being addressed. (base_addr value reflecting io
	 * space, have their first 2 bits mask out, while base_addr value
	 * reflecting mem space, have their first 4 bits mask out.)  These
	 * bits are flag bits and should always	be 0 when used as an address.
	 */

	if ((*base_addr0) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr0) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr0) &= PCI_BASE_ADDRESS_MEM_MASK;

	if ((*base_addr1) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr1) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr1) &= PCI_BASE_ADDRESS_MEM_MASK;

	if ((*base_addr2) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr2) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr2) &= PCI_BASE_ADDRESS_MEM_MASK;

	if ((*base_addr3) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr3) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr3) &= PCI_BASE_ADDRESS_MEM_MASK;

	if ((*base_addr4) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr4) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr4) &= PCI_BASE_ADDRESS_MEM_MASK;

	if ((*base_addr5) & PCI_BASE_ADDRESS_SPACE_IO) /* Is this an io addr?*/
		(*base_addr5) &= PCI_BASE_ADDRESS_IO_MASK;
	else
		(*base_addr5) &= PCI_BASE_ADDRESS_MEM_MASK;

	if (error) {
		epca_mess(KERN_ERR, "get_PCI_configuration error: "
			  "board not initializing due to error\n");
		return(0);
	}
	return(1);
}


/*
 * init_one_PCI()
 *
 * Attempt to initialize a PCI board.
 * Return 1 (number of boards initialized) if successful, 0 otheriwise.
 */

static int init_one_PCI( unsigned bd_type, unsigned int bd_num,
			 unsigned char bus, unsigned char device_fn )
{

	unsigned int base_addr0, base_addr1, base_addr2,
                base_addr3, base_addr4, base_addr5,
		xbase_addr;

	base_addr0 = base_addr1 = base_addr2 = 0;
	base_addr3 = base_addr4 = base_addr5 = 0;

	if (get_PCI_configuration(bus, device_fn, &base_addr0, &base_addr1,
				  &base_addr2, &base_addr3,
				  &base_addr4, &base_addr5)) {
		pci_boards[bd_num]->status   = ENABLED;
		pci_boards[bd_num]->type     = bd_type;
		pci_boards[bd_num]->numports = 0x0;

		/* Most cards use base_addr0, but some use base_addr2 */
		if( bd_type == PCIXRJ || bd_type == PCIEPCJ )
			xbase_addr = base_addr2;
		else
			xbase_addr = base_addr0;

		pci_boards[bd_num]->port =
			(unsigned long)(xbase_addr + PCI_IO_OFFSET);

		pci_boards[bd_num]->membase  = (unsigned long)xbase_addr;

		if (xbase_addr >= 0x100000) {
			/*
			 * Standard physical addresses are valid to the kernel
			 * as long as they aren't above RAM. Higher addresses
			 * (such as are typical of a PCI system) need to be
			 * mapped in with the ioremap command.  For boards
			 * using such high addresses the driver will store
			 * both addresses.  One address (the physical) will
			 * be held for the app's use (and mmap) and the other
			 * (the	ioremapped address) will be used in the kernel.
			 */
			pci_boards[bd_num]->re_map_port =
				ioremap((xbase_addr + PCI_IO_OFFSET),0x200000);
			pci_boards[bd_num]->re_map_membase =
				ioremap(xbase_addr, 0x200000);
		}

		/*
		 * Special initialization for non-PLX boards
		 */
		if( bd_type != PCIXRJ && bd_type != PCIEPCJ) {
			pcibios_write_config_byte(bus, device_fn, 0x40, 0);  
			pcibios_write_config_byte(bus, device_fn, 0x46, 0);  

			/* Limit burst length to 2 doubleword transactions*/
			pcibios_write_config_byte(bus, device_fn, 0x42, 1);  
		}

		return( 1 );	/* Successfully initialized 1 board */
	}

	return( 0 );
}


/*
 * init_PCI()
 */
static int init_PCI(int boards_found)
{

	struct board_info **tmpboards;
	
	unsigned char	bus, device_fn;
	unsigned int itmp;
	unsigned int ktmp;
	unsigned int pci_count = 0; /* num successfully initialized adapters */
	unsigned int dev_index = 0; /* Index into device array */
	unsigned int dev_id = 0;    /* Device id (e.g: PCI_DEVICE_XEM) */
	unsigned int board_index = 0; /* Support multiple boards of each
				       * device type */

/*
 *  Change this to be 1 to enable PCI probing debugging
 */
#if 0
	#define PROBE_DBG(x) epca_mess x
#else
	#define PROBE_DBG(x)
#endif


	struct {
		unsigned pci_id;
		unsigned bd_type;
	}  pci_dev[] =  {	/* { PCI_id, epca_driver_board_type } */
		{ PCI_DEVICE_XEM, PCIXEM },
		{ PCI_DEVICE_XR, PCIXR },
		{ PCI_DEVICE_CX, PCICX },
		{ PCI_DEVICE_XRJ, PCIXRJ }, /* Jupiter board PLX9060 chip*/
		{ PCI_DEVICE_920_2, PCI920_2 }, /* XR-Plus 920 K, 4 port */
		{ PCI_DEVICE_920_4, PCI920_4 }, /* XR-Plus 920 K, 4 port */
		{ PCI_DEVICE_920_8, PCI920_8 }, /* XR-Plus 920 K, 8 port */
		{ PCI_DEVICE_XR_422, PCIXR_422 }, /* XR-422 */
		{ PCI_DEVICE_EPCJ, PCIEPCJ },
		{ PCI_DEVICE_XR_IBM, PCIXR_IBM }, /* IBM 8-Port Async Adapter */
		{ PCI_DEVICE_CX_IBM, PCICX_IBM }, /* IBM 128-Port Async Adapter */
		{ 0, 0 }
	};


	/*
	 *  Prepare the data structures which will hold all
	 *  information about detected PCI adapters.
	 */

	pci_boards = (struct board_info**)kmalloc(sizeof(struct board_info*) *
						  MAXBOARDS, GFP_KERNEL);
	for(itmp=0;itmp<MAXBOARDS;itmp++) 
		pci_boards[itmp] = kmalloc(sizeof(struct board_info),
					   GFP_KERNEL);
	for(itmp=0;itmp<MAXBOARDS;itmp++)
		memset(pci_boards[itmp], 0, sizeof(struct board_info));


	/*
	 *  Run through each PCI device ID.  For each one in turn, detect
	 *  all adapters of that type, up to the maximum number of boards.
	 *  Detect as many as we can... we'll later only use those boards
	 *  we were expecting to see.
	 *
	 *  The "while" loop below is a mini state machine.  Each time
	 *  through the loop is a check to see if there is a currently
	 *  unprobed adapter with a given PCI device ID.  If there are no
	 *  more adapters with the current ID, we can move to the next
	 *  ID, and prepare to probe the FIRST adapter with that ID.
	 *  If we DID detect a new adapter of the appropriate type,
	 *  be sure to record the information, then prepare to probe
	 *  the NEXT adapter with the SAME ID.
	 */
	board_index = dev_index = pci_count = 0 ;

	while( pci_count < MAXBOARDS &&
	       (dev_id = pci_dev[dev_index].pci_id) != 0 ) {

		PROBE_DBG(( KERN_INFO, "Probing Device ID 0x%04x (%d)\n",
		            dev_id, board_index ));

		if ( !pcibios_find_device(PCI_VENDOR_DIGI, dev_id,
		                          board_index, &bus, &device_fn) ) {
			/* Prepare to detect another adapter with this PCI ID */
			board_index++;

			/* Setup the detected adapter and update running total */
			pci_count += init_one_PCI( pci_dev[dev_index].bd_type,
			                           pci_count, bus, device_fn );

			PROBE_DBG(( KERN_INFO, "                  Found.  Tot so far: %d\n",
			            pci_count ));
		} else {
			dev_index++;
			board_index = 0;
		}
	}

	PROBE_DBG(( KERN_INFO, "Detected %d PCI adapters total.\n", pci_count ));


	/*
	 * Put together a temporary board structure with the data from
	 * the original boards structure..
	 */
	
	/* Allocate proper pointer space */
	tmpboards = (struct board_info **)kmalloc(sizeof(struct board_info *) *
						  MAXBOARDS, GFP_KERNEL);

	/* Clear it */
	memset(tmpboards, 0, sizeof(struct board_info *) * MAXBOARDS);


	PROBE_DBG(( KERN_INFO, "\nPre PCI Probe Board Structures:\n\n" ));
	

	/* Fill those pointers up with memory, and put data in 'em! */
	for(itmp=0; itmp < MAXBOARDS; itmp++) {
		tmpboards[itmp] = (struct board_info *)
			kmalloc(sizeof(struct board_info), GFP_KERNEL);
		memcpy(tmpboards[itmp], &boards[itmp],
		       sizeof(struct board_info));
		PROBE_DBG(( KERN_INFO, "    (%d)  type: 0x%x  "
		                       "stat: 0x%x  membase: 0x%08x\n",
		            itmp, boards[itmp].type, boards[itmp].status,
		            boards[itmp].membase ));
	}

	/* clear out boards structure - Don't worry - We'll refill it! */
	for(itmp=0; itmp < MAXBOARDS; itmp++) 
		memset(&boards[itmp], 0, sizeof(struct board_info));
	
	/*
	 * Check tmpboards[0] to find out if there isn't ANYTHING in it
	 * If there isn't, fail hard, at this point. This is a CRITICAL
	 * ERROR.
	 */
	if(tmpboards[0]->numports == 0) 
		return(-1);
	
	/* Okay. there's at least 1! */
	for(itmp=0;itmp<MAXBOARDS;itmp++) {
	
		/*
		 * look at tmpboards[itmp] - if it doesn't exist, go on
		 * to next test condition...
		 */
		if(tmpboards[itmp]->numports == 0) 
			continue;
		
		/* If the board's DPA type shows PCI... */
		if( (btypes[tmpboards[itmp]->type].dpatype & BUSTYPE) !=
		    T_PCIBUS ) {
			continue;
		} else {			
			/* Okay. we've found a PCI board */
			for(ktmp=0;ktmp<MAXBOARDS+1;ktmp++) {
				
				/*
				 * Cheap hack to allow us to detect when we've
				 * been through all boards..
				 */
				if(ktmp == MAXBOARDS) {
					tmpboards[itmp]->used = -1;
					break;
				}
				
				/*
				 * Does this PCI board match that of what
				 * we're looking for? If not, continue...
				 */
				if(pci_boards[ktmp]->type !=
				   tmpboards[itmp]->type)
					continue;
					
				/*
				 * It does! Is the board already used? If so,
				 * continue..
				 */
				if(pci_boards[ktmp]->used)
					continue;
				
				/*
				 * Its not used!! Lets use it for the board we
				 * want, and mark as such!
				 */
				pci_boards[ktmp]->used = 1;
				
				/*
				 * since the numports structure is HELD in
				 * the right place when we get here, we need
				 * to save it before memcpy..
				 */
				/* Do the copy.. */

				PROBE_DBG(( KERN_INFO, "          PCI #%d --> Board #%d\n",
				            ktmp, itmp ));
				
				tmpboards[itmp]->status =
					pci_boards[ktmp]->status;
				tmpboards[itmp]->dpastatus =
					pci_boards[ktmp]->dpastatus;
				tmpboards[itmp]->pdev =
					pci_boards[ktmp]->pdev;
				tmpboards[itmp]->port =
					pci_boards[ktmp]->port;
				tmpboards[itmp]->membase =
					pci_boards[ktmp]->membase;
				tmpboards[itmp]->re_map_port =
					pci_boards[ktmp]->re_map_port;
				tmpboards[itmp]->re_map_membase =
					pci_boards[ktmp]->re_map_membase;
				

#if 0	/* These should BE here already! */
				tmpboards[itmp]->type =
					pci_boards[ktmp].type;
				tmpboards[itmp]->dpatype =
					pci_boards[ktmp].dpatype;
				tmpboards[itmp]->numports =
					pci_boards[ktmp].numports;
				tmpboards[itmp]->altpin =
					pci_boards[ktmp].altpin;
#endif
				break;
			}
		}
	}
	
	/* Okay. Lets put the boards structure back together, piece by piece */
	
	/*
	 * Check for something bad, like tmpboards[0] not holding anything..
	 * If that happens, something REALLY BAD happened, and we are going
	 * to fail miserably..
	 */
	if(!tmpboards[0]->numports) {
		APR(("We broke BADLY! tmpboards[0] doesn't have "
                     "anything in it!!\n"));
		return(-1);
	}


	PROBE_DBG(( KERN_INFO, "\nPost PCI Probe Board Structures:\n\n" ));

	
	for(itmp=0; itmp < MAXBOARDS; itmp++) {
		memcpy(&boards[itmp], tmpboards[itmp],
		       sizeof(struct board_info));
		if(boards[itmp].used == -1) {
			boards[itmp].used = 0;
			boards[itmp].status = DISABLED;
			boards[itmp].port = 0x0;
			boards[itmp].membase = 0;
		}
		PROBE_DBG(( KERN_INFO, "    (%d)  type: 0x%x  "
		                       "stat: 0x%x  membase: 0x%08x\n",
		            itmp, boards[itmp].type, boards[itmp].status,
		            boards[itmp].membase ));
	}

	/* Okay. Get rid of tmpboards */
	for(itmp=0; itmp < MAXBOARDS; itmp++) 
		kfree(tmpboards[itmp]);
	kfree(tmpboards);
	
	/* Don't need pci_boards either. get rid of it too! */
	for(itmp=0; itmp < MAXBOARDS; itmp++)
		kfree(pci_boards[itmp]);
	kfree(pci_boards);
	
	return( pci_count );
}
#endif
